/* Файл scormAPI.js */

function WindowObjectFinder(objectName) {
    this.objectName = objectName;
    this.findAPITries = 0;
    this.objectHandle = null;
    this.noObjectFound = false;

    /**
     * Функция поиска объекта с названием objectName в текущем фрейме (window) и во всех
     * родительских (parent)
     * Параметры:  объект window
     * Возвращает: объект с названием objectName, если таковой был найден, в противном случае - null
     */
    this.findObject = function(win) {
        while (!win[this.objectName]) {
            this.findAPITries++;
            if (this.findAPITries > 100) {
                return null;
            }
            if (win.parent == null || win.parent == win) {
                break;
            }
            win = win.parent
        }
        return win[this.objectName];
    };

    /**
     * Функция возвращает искомый объект, если таковой был найден в текущем фрейме (window),
     * в каком либо из его родительских (parent), в окне opener, или его
     * родительских. В противном случае возвращается null.
     */
    this.getObject = function() {
        if (this.objectHandle == null && !this.noObjectFound) {
            for (var curWnd = window; curWnd && typeof(curWnd) != "undefined"; curWnd = curWnd.opener) {
                var theObject = this.findObject(curWnd);
                if (theObject) {
                    this.objectHandle = theObject;
                    break;
                }
            }

            if (this.objectHandle == null) {
                this.noObjectFound = true;
                throw new Error("NoObjectFound")
            }
        }
        return this.objectHandle;
    };
}

//конструктор для объекта, предоставляющего интерфейс к SCORM-API
function _SCORM(initializerWindow) {
    // Переменная для отладки. Во время отладки устанавливается в true для показа
    // отладочной информации, в финальном варианте должна быть false
    this._debug = false;

    // локальные переменные
    this.terminated = false;
    this.initializerWindow = initializerWindow;
    this.scormAPIFinfer = new WindowObjectFinder("API_1484_11");

    /**
     * Функция возвращает SCORM API, если таковой был найден в текущем фрейме (window),
     * в каком либо из его родительских (parent), в окне opener, или его
     * родительских. В противном случае возвращается null.
     */
    this.getAPI = function() {
        try {
            return this.scormAPIFinfer.getObject();
        }
        catch (e) {
            if(this._debug && e.message == "NoObjectFound") {
                alert("Не удалось найти реализацию SCORM API.");
            }
        }
        return null;
    };


    /*******************************************************************************
     * Данная функция вызывается для инициализации сессии. Функция должна вызываться
     * до вызовов getDataValue, setDataValue, или terminate
     * Возвращает: "true", если процесс инициализации прошел успешно, в противном
     *             случае - "false".
     *******************************************************************************/
    this.initialize = function() {
        var result = "false";
        var api = this.getAPI();
        if (api != null) {
            result = api.Initialize("");
            if (result != "true") {
                this.displayErrorInfo(this.getLastErrorCode());
            }
        }
        return result;
    };

    /*******************************************************************************
     * Данная функция вызывается для завершения сессии. После ее вызова нельзя
     * вызывать функции initialize, getDataValue или setDataValue
     * Возвращает: "true", если прошло успешно, в противном случае - "false".
     *******************************************************************************/
    this.terminate = function() {
        var result = "false";
        var api = this.getAPI();
        if (api != null) {
            if (!this.terminated) {
                result = api.Terminate("");
                if (result != "true") {
                    this.displayErrorInfo(this.getLastErrorCode());
                } else {
                    this.terminated = true;
                }
            }
        }
        return result;
    };

    /*******************************************************************************
     * Функция запрашивает у Системы значение модели данных. Может вызываться после
     * вызова initialize, но до вызова terminate
     * Параметры:  название элемента модели данных (напр. "cmi.learner_id")
     * Возвращает: значение запрашиваемых данных
     *******************************************************************************/
    this.getDataValue = function(name) {
        var result = "";
        var api = this.getAPI();
        if (!this.terminated && api != null) {
            result = api.GetValue(name);
            var errCode = this.getLastErrorCode();
            if (errCode != "0") {
                this.displayErrorInfo(errCode);
            }
        }
        return result;
    };

    /*******************************************************************************
     * Функция сохраняет значение в модели данных, определенное значением name.
     * Может вызываться после вызова initialize, но до вызова terminate.
     * Параметры:  @name - название элемента модели данных
     *              @value - сохраняемое значение
     * Возвращает: true - если успешно, false - при ошибке.
     *******************************************************************************/
    this.setDataValue = function(name, value) {
        var result = "false";
        var api = this.getAPI();
        if (!this.terminated && api != null) {
            result = api.SetValue(name, value);
            if (result != "true") {
                this.displayErrorInfo(this.getLastErrorCode());
            }
        }
        return result;
    };

    /*******************************************************************************
     * Донная функция дает команду Системе на сохранение последних изменений.
     * Может вызываться после вызова initialize, но до вызова terminate.
     *******************************************************************************/
    this.commitData = function() {
        var result = "false";
        var api = this.getAPI();
        if (!this.terminated && api != null) {
            result = api.Commit("");
            if (result != "true") {
                this.displayErrorInfo(this.getLastErrorCode());
            }
        }
        return result;
    };

    /*******************************************************************************
     * Функция возвращает код последней ошибки.
     * Возвращает: строку с кодом ошибки (возвращает "0" - если не было ошибки).
     *******************************************************************************/
    this.getLastErrorCode = function() {
        var api = this.getAPI();
        return ( api == null) ? ("") : ( api.GetLastError() );
    };

    /*******************************************************************************
     * Вспомогательная функция для показа информации об ошибке (код ошибки, описание
     * и диагностическую информацию)
     * Параметры:  errCode - код ошибки
     *******************************************************************************/
    this.displayErrorInfo = function(errCode) {
        var errString, errDiagnostic;
        if (this._debug) {
            var api = this.getAPI();
            if (api != null) {
                errString = api.GetErrorString(errCode);
                errDiagnostic = api.GetDiagnostic("");
            }
            alert("Ошибка SCORM RTE: " + errCode + " - " + errString + "\n\n" + errDiagnostic);
        }
    };
}

_SCORM.getInstance = function() {
    var objectFinder = new WindowObjectFinder("SCORM");
    try {
        var obj = objectFinder.getObject();
        if(obj) {
            return obj;
        }
    }
    catch (e) {
    }
    return new _SCORM(window);
};

var SCORM = _SCORM.getInstance();
var SCORM_INITIALIZED = false;
var fOnLoad = function() {
    if(window == SCORM.initializerWindow)
        SCORM_INITIALIZED = SCORM.initialize()=="true";
};
window.addEventListener("load", fOnLoad, false);

var fOnBeforeUnload = function() {
    window.removeEventListener("beforeunload", fOnBeforeUnload, false);
    window.removeEventListener("unload", fOnUnload, false);
    //window.removeEventListener("pagehide", fOnUnload, false);
    if(window == SCORM.initializerWindow){
        SCORM.terminate();
	}
};

var fOnUnload = function() {
    window.removeEventListener("beforeunload", fOnBeforeUnload, false);
    window.removeEventListener("unload", fOnUnload, false);
    //window.removeEventListener("pagehide", fOnUnload, false);
    if(window == SCORM.initializerWindow){
        SCORM.terminate();
	}
};
window.addEventListener("unload", fOnUnload, false);
//window.addEventListener("pagehide", fOnUnload, false);
/*!
* @license PreloadJS
* Visit http://createjs.com/ for documentation, updates and examples.
*
* Copyright (c) 2011-2013 gskinner.com, inc.
*
* Distributed under the terms of the MIT license.
* http://www.opensource.org/licenses/mit-license.html
*
* This notice shall be included in all copies or substantial portions of the Software.
*/
this.createjs=this.createjs||{},function(){"use strict";var a=createjs.PreloadJS=createjs.PreloadJS||{};a.version="0.4.1",a.buildDate="Thu, 12 Dec 2013 23:33:38 GMT"}(),this.createjs=this.createjs||{},function(){"use strict";var a=function(a,b,c){this.initialize(a,b,c)},b=a.prototype;b.type=null,b.target=null,b.currentTarget=null,b.eventPhase=0,b.bubbles=!1,b.cancelable=!1,b.timeStamp=0,b.defaultPrevented=!1,b.propagationStopped=!1,b.immediatePropagationStopped=!1,b.removed=!1,b.initialize=function(a,b,c){this.type=a,this.bubbles=b,this.cancelable=c,this.timeStamp=(new Date).getTime()},b.preventDefault=function(){this.defaultPrevented=!0},b.stopPropagation=function(){this.propagationStopped=!0},b.stopImmediatePropagation=function(){this.immediatePropagationStopped=this.propagationStopped=!0},b.remove=function(){this.removed=!0},b.clone=function(){return new a(this.type,this.bubbles,this.cancelable)},b.toString=function(){return"[Event (type="+this.type+")]"},createjs.Event=a}(),this.createjs=this.createjs||{},function(){"use strict";var a=function(){},b=a.prototype;a.initialize=function(a){a.addEventListener=b.addEventListener,a.on=b.on,a.removeEventListener=a.off=b.removeEventListener,a.removeAllEventListeners=b.removeAllEventListeners,a.hasEventListener=b.hasEventListener,a.dispatchEvent=b.dispatchEvent,a._dispatchEvent=b._dispatchEvent,a.willTrigger=b.willTrigger},b._listeners=null,b._captureListeners=null,b.initialize=function(){},b.addEventListener=function(a,b,c){var d;d=c?this._captureListeners=this._captureListeners||{}:this._listeners=this._listeners||{};var e=d[a];return e&&this.removeEventListener(a,b,c),e=d[a],e?e.push(b):d[a]=[b],b},b.on=function(a,b,c,d,e,f){return b.handleEvent&&(c=c||b,b=b.handleEvent),c=c||this,this.addEventListener(a,function(a){b.call(c,a,e),d&&a.remove()},f)},b.removeEventListener=function(a,b,c){var d=c?this._captureListeners:this._listeners;if(d){var e=d[a];if(e)for(var f=0,g=e.length;g>f;f++)if(e[f]==b){1==g?delete d[a]:e.splice(f,1);break}}},b.off=b.removeEventListener,b.removeAllEventListeners=function(a){a?(this._listeners&&delete this._listeners[a],this._captureListeners&&delete this._captureListeners[a]):this._listeners=this._captureListeners=null},b.dispatchEvent=function(a,b){if("string"==typeof a){var c=this._listeners;if(!c||!c[a])return!1;a=new createjs.Event(a)}if(a.target=b||this,a.bubbles&&this.parent){for(var d=this,e=[d];d.parent;)e.push(d=d.parent);var f,g=e.length;for(f=g-1;f>=0&&!a.propagationStopped;f--)e[f]._dispatchEvent(a,1+(0==f));for(f=1;g>f&&!a.propagationStopped;f++)e[f]._dispatchEvent(a,3)}else this._dispatchEvent(a,2);return a.defaultPrevented},b.hasEventListener=function(a){var b=this._listeners,c=this._captureListeners;return!!(b&&b[a]||c&&c[a])},b.willTrigger=function(a){for(var b=this;b;){if(b.hasEventListener(a))return!0;b=b.parent}return!1},b.toString=function(){return"[EventDispatcher]"},b._dispatchEvent=function(a,b){var c,d=1==b?this._captureListeners:this._listeners;if(a&&d){var e=d[a.type];if(!e||!(c=e.length))return;a.currentTarget=this,a.eventPhase=b,a.removed=!1,e=e.slice();for(var f=0;c>f&&!a.immediatePropagationStopped;f++){var g=e[f];g.handleEvent?g.handleEvent(a):g(a),a.removed&&(this.off(a.type,g,1==b),a.removed=!1)}}},createjs.EventDispatcher=a}(),this.createjs=this.createjs||{},function(){"use strict";createjs.indexOf=function(a,b){for(var c=0,d=a.length;d>c;c++)if(b===a[c])return c;return-1}}(),this.createjs=this.createjs||{},function(){"use strict";createjs.proxy=function(a,b){var c=Array.prototype.slice.call(arguments,2);return function(){return a.apply(b,Array.prototype.slice.call(arguments,0).concat(c))}}}(),this.createjs=this.createjs||{},function(){"use strict";var a=function(){this.init()};a.prototype=new createjs.EventDispatcher;var b=a.prototype,c=a;c.FILE_PATTERN=/^(?:(\w+:)\/{2}(\w+(?:\.\w+)*\/?)|(.{0,2}\/{1}))?([/.]*?(?:[^?]+)?\/)?((?:[^/?]+)\.(\w+))(?:\?(\S+)?)?$/,c.PATH_PATTERN=/^(?:(\w+:)\/{2})|(.{0,2}\/{1})?([/.]*?(?:[^?]+)?\/?)?$/,b.loaded=!1,b.canceled=!1,b.progress=0,b._item=null,b.getItem=function(){return this._item},b.init=function(){},b.load=function(){},b.close=function(){},b._sendLoadStart=function(){this._isCanceled()||this.dispatchEvent("loadstart")},b._sendProgress=function(a){if(!this._isCanceled()){var b=null;"number"==typeof a?(this.progress=a,b=new createjs.Event("progress"),b.loaded=this.progress,b.total=1):(b=a,this.progress=a.loaded/a.total,(isNaN(this.progress)||1/0==this.progress)&&(this.progress=0)),b.progress=this.progress,this.hasEventListener("progress")&&this.dispatchEvent(b)}},b._sendComplete=function(){this._isCanceled()||this.dispatchEvent("complete")},b._sendError=function(a){!this._isCanceled()&&this.hasEventListener("error")&&(null==a&&(a=new createjs.Event("error")),this.dispatchEvent(a))},b._isCanceled=function(){return null==window.createjs||this.canceled?!0:!1},b._parseURI=function(a){return a?a.match(c.FILE_PATTERN):null},b._parsePath=function(a){return a?a.match(c.PATH_PATTERN):null},b._formatQueryString=function(a,b){if(null==a)throw new Error("You must specify data.");var c=[];for(var d in a)c.push(d+"="+escape(a[d]));return b&&(c=c.concat(b)),c.join("&")},b.buildPath=function(a,b){if(null==b)return a;var c=[],d=a.indexOf("?");if(-1!=d){var e=a.slice(d+1);c=c.concat(e.split("&"))}return-1!=d?a.slice(0,d)+"?"+this._formatQueryString(b,c):a+"?"+this._formatQueryString(b,c)},b._isCrossDomain=function(a){var b=document.createElement("a");b.href=a.src;var c=document.createElement("a");c.href=location.href;var d=""!=b.hostname&&(b.port!=c.port||b.protocol!=c.protocol||b.hostname!=c.hostname);return d},b._isLocal=function(a){var b=document.createElement("a");return b.href=a.src,""==b.hostname&&"file:"==b.protocol},b.toString=function(){return"[PreloadJS AbstractLoader]"},createjs.AbstractLoader=a}(),this.createjs=this.createjs||{},function(){"use strict";var a=function(a,b,c){this.init(a,b,c)},b=a.prototype=new createjs.AbstractLoader,c=a;c.loadTimeout=8e3,c.LOAD_TIMEOUT=0,c.BINARY="binary",c.CSS="css",c.IMAGE="image",c.JAVASCRIPT="javascript",c.JSON="json",c.JSONP="jsonp",c.MANIFEST="manifest",c.SOUND="sound",c.SVG="svg",c.TEXT="text",c.XML="xml",c.POST="POST",c.GET="GET",b._basePath=null,b._crossOrigin="",b.useXHR=!0,b.stopOnError=!1,b.maintainScriptOrder=!0,b.next=null,b._typeCallbacks=null,b._extensionCallbacks=null,b._loadStartWasDispatched=!1,b._maxConnections=1,b._currentlyLoadingScript=null,b._currentLoads=null,b._loadQueue=null,b._loadQueueBackup=null,b._loadItemsById=null,b._loadItemsBySrc=null,b._loadedResults=null,b._loadedRawResults=null,b._numItems=0,b._numItemsLoaded=0,b._scriptOrder=null,b._loadedScripts=null,b.init=function(a,b,c){this._numItems=this._numItemsLoaded=0,this._paused=!1,this._loadStartWasDispatched=!1,this._currentLoads=[],this._loadQueue=[],this._loadQueueBackup=[],this._scriptOrder=[],this._loadedScripts=[],this._loadItemsById={},this._loadItemsBySrc={},this._loadedResults={},this._loadedRawResults={},this._typeCallbacks={},this._extensionCallbacks={},this._basePath=b,this.setUseXHR(a),this._crossOrigin=c===!0?"Anonymous":c===!1||null==c?"":c},b.setUseXHR=function(a){return this.useXHR=0!=a&&null!=window.XMLHttpRequest,this.useXHR},b.removeAll=function(){this.remove()},b.remove=function(a){var b=null;if(!a||a instanceof Array){if(a)b=a;else if(arguments.length>0)return}else b=[a];var c=!1;if(b){for(;b.length;){var d=b.pop(),e=this.getResult(d);for(f=this._loadQueue.length-1;f>=0;f--)if(g=this._loadQueue[f].getItem(),g.id==d||g.src==d){this._loadQueue.splice(f,1)[0].cancel();break}for(f=this._loadQueueBackup.length-1;f>=0;f--)if(g=this._loadQueueBackup[f].getItem(),g.id==d||g.src==d){this._loadQueueBackup.splice(f,1)[0].cancel();break}if(e)delete this._loadItemsById[e.id],delete this._loadItemsBySrc[e.src],this._disposeItem(e);else for(var f=this._currentLoads.length-1;f>=0;f--){var g=this._currentLoads[f].getItem();if(g.id==d||g.src==d){this._currentLoads.splice(f,1)[0].cancel(),c=!0;break}}}c&&this._loadNext()}else{this.close();for(var h in this._loadItemsById)this._disposeItem(this._loadItemsById[h]);this.init(this.useXHR)}},b.reset=function(){this.close();for(var a in this._loadItemsById)this._disposeItem(this._loadItemsById[a]);for(var b=[],c=0,d=this._loadQueueBackup.length;d>c;c++)b.push(this._loadQueueBackup[c].getItem());this.loadManifest(b,!1)},c.isBinary=function(a){switch(a){case createjs.LoadQueue.IMAGE:case createjs.LoadQueue.BINARY:return!0;default:return!1}},c.isText=function(a){switch(a){case createjs.LoadQueue.TEXT:case createjs.LoadQueue.JSON:case createjs.LoadQueue.MANIFEST:case createjs.LoadQueue.XML:case createjs.LoadQueue.HTML:case createjs.LoadQueue.CSS:case createjs.LoadQueue.SVG:case createjs.LoadQueue.JAVASCRIPT:return!0;default:return!1}},b.installPlugin=function(a){if(null!=a&&null!=a.getPreloadHandlers){var b=a.getPreloadHandlers();if(b.scope=a,null!=b.types)for(var c=0,d=b.types.length;d>c;c++)this._typeCallbacks[b.types[c]]=b;if(null!=b.extensions)for(c=0,d=b.extensions.length;d>c;c++)this._extensionCallbacks[b.extensions[c]]=b}},b.setMaxConnections=function(a){this._maxConnections=a,!this._paused&&this._loadQueue.length>0&&this._loadNext()},b.loadFile=function(a,b,c){if(null==a){var d=new createjs.Event("error");return d.text="PRELOAD_NO_FILE",this._sendError(d),void 0}this._addItem(a,null,c),b!==!1?this.setPaused(!1):this.setPaused(!0)},b.loadManifest=function(a,b,d){var e=null,f=null;if(a instanceof Array){if(0==a.length){var g=new createjs.Event("error");return g.text="PRELOAD_MANIFEST_EMPTY",this._sendError(g),void 0}e=a}else if("string"==typeof a)e=[{src:a,type:c.MANIFEST}];else{if("object"!=typeof a){var g=new createjs.Event("error");return g.text="PRELOAD_MANIFEST_NULL",this._sendError(g),void 0}if(void 0!==a.src){if(null==a.type)a.type=c.MANIFEST;else if(a.type!=c.MANIFEST){var g=new createjs.Event("error");g.text="PRELOAD_MANIFEST_ERROR",this._sendError(g)}e=[a]}else void 0!==a.manifest&&(e=a.manifest,f=a.path)}for(var h=0,i=e.length;i>h;h++)this._addItem(e[h],f,d);b!==!1?this.setPaused(!1):this.setPaused(!0)},b.load=function(){this.setPaused(!1)},b.getItem=function(a){return this._loadItemsById[a]||this._loadItemsBySrc[a]},b.getResult=function(a,b){var c=this._loadItemsById[a]||this._loadItemsBySrc[a];if(null==c)return null;var d=c.id;return b&&this._loadedRawResults[d]?this._loadedRawResults[d]:this._loadedResults[d]},b.setPaused=function(a){this._paused=a,this._paused||this._loadNext()},b.close=function(){for(;this._currentLoads.length;)this._currentLoads.pop().cancel();this._scriptOrder.length=0,this._loadedScripts.length=0,this.loadStartWasDispatched=!1},b._addItem=function(a,b,c){var d=this._createLoadItem(a,b,c);if(null!=d){var e=this._createLoader(d);null!=e&&(this._loadQueue.push(e),this._loadQueueBackup.push(e),this._numItems++,this._updateProgress(),this.maintainScriptOrder&&d.type==createjs.LoadQueue.JAVASCRIPT&&e instanceof createjs.XHRLoader&&(this._scriptOrder.push(d),this._loadedScripts.push(null)))}},b._createLoadItem=function(a,b,c){var d=null;switch(typeof a){case"string":d={src:a};break;case"object":d=window.HTMLAudioElement&&a instanceof window.HTMLAudioElement?{tag:a,src:d.tag.src,type:createjs.LoadQueue.SOUND}:a;break;default:return null}var e=this._parseURI(d.src);null!=e&&(d.ext=e[6]),null==d.type&&(d.type=this._getTypeByExtension(d.ext));var f="",g=c||this._basePath,h=d.src;if(e&&null==e[1]&&null==e[3])if(b){f=b;var i=this._parsePath(b);h=b+h,null!=g&&i&&null==i[1]&&null==i[2]&&(f=g+f)}else null!=g&&(f=g);if(d.src=f+d.src,d.path=f,(d.type==createjs.LoadQueue.JSON||d.type==createjs.LoadQueue.MANIFEST)&&(d._loadAsJSONP=null!=d.callback),d.type==createjs.LoadQueue.JSONP&&null==d.callback)throw new Error("callback is required for loading JSONP requests.");(void 0===d.tag||null===d.tag)&&(d.tag=this._createTag(d)),(void 0===d.id||null===d.id||""===d.id)&&(d.id=h);var j=this._typeCallbacks[d.type]||this._extensionCallbacks[d.ext];if(j){var k=j.callback.call(j.scope,d.src,d.type,d.id,d.data,f,this);if(k===!1)return null;k===!0||(null!=k.src&&(d.src=k.src),null!=k.id&&(d.id=k.id),null!=k.tag&&(d.tag=k.tag),null!=k.completeHandler&&(d.completeHandler=k.completeHandler),k.type&&(d.type=k.type),e=this._parseURI(d.src),null!=e&&null!=e[6]&&(d.ext=e[6].toLowerCase()))}return this._loadItemsById[d.id]=d,this._loadItemsBySrc[d.src]=d,d},b._createLoader=function(a){var b=this.useXHR;switch(a.type){case createjs.LoadQueue.JSON:case createjs.LoadQueue.MANIFEST:b=!a._loadAsJSONP;break;case createjs.LoadQueue.XML:case createjs.LoadQueue.TEXT:b=!0;break;case createjs.LoadQueue.SOUND:case createjs.LoadQueue.JSONP:b=!1;break;case null:return null}return b?new createjs.XHRLoader(a,this._crossOrigin):new createjs.TagLoader(a)},b._loadNext=function(){if(!this._paused){this._loadStartWasDispatched||(this._sendLoadStart(),this._loadStartWasDispatched=!0),this._numItems==this._numItemsLoaded?(this.loaded=!0,this._sendComplete(),this.next&&this.next.load&&this.next.load()):this.loaded=!1;for(var a=0;a<this._loadQueue.length&&!(this._currentLoads.length>=this._maxConnections);a++){var b=this._loadQueue[a];if(this.maintainScriptOrder&&b instanceof createjs.TagLoader&&b.getItem().type==createjs.LoadQueue.JAVASCRIPT){if(this._currentlyLoadingScript)continue;this._currentlyLoadingScript=!0}this._loadQueue.splice(a,1),a--,this._loadItem(b)}}},b._loadItem=function(a){a.on("progress",this._handleProgress,this),a.on("complete",this._handleFileComplete,this),a.on("error",this._handleFileError,this),this._currentLoads.push(a),this._sendFileStart(a.getItem()),a.load()},b._handleFileError=function(a){var b=a.target;this._numItemsLoaded++,this._updateProgress();var c=new createjs.Event("error");c.text="FILE_LOAD_ERROR",c.item=b.getItem(),this._sendError(c),this.stopOnError||(this._removeLoadItem(b),this._loadNext())},b._handleFileComplete=function(a){var b=a.target,c=b.getItem();if(this._loadedResults[c.id]=b.getResult(),b instanceof createjs.XHRLoader&&(this._loadedRawResults[c.id]=b.getResult(!0)),this._removeLoadItem(b),this.maintainScriptOrder&&c.type==createjs.LoadQueue.JAVASCRIPT){if(!(b instanceof createjs.TagLoader))return this._loadedScripts[createjs.indexOf(this._scriptOrder,c)]=c,this._checkScriptLoadOrder(b),void 0;this._currentlyLoadingScript=!1}if(delete c._loadAsJSONP,c.type==createjs.LoadQueue.MANIFEST){var d=b.getResult();null!=d&&void 0!==d.manifest&&this.loadManifest(d,!0)}this._processFinishedLoad(c,b)},b._processFinishedLoad=function(a,b){this._numItemsLoaded++,this._updateProgress(),this._sendFileComplete(a,b),this._loadNext()},b._checkScriptLoadOrder=function(){for(var a=this._loadedScripts.length,b=0;a>b;b++){var c=this._loadedScripts[b];if(null===c)break;if(c!==!0){var d=this._loadedResults[c.id];(document.body||document.getElementsByTagName("body")[0]).appendChild(d),this._processFinishedLoad(c),this._loadedScripts[b]=!0}}},b._removeLoadItem=function(a){for(var b=this._currentLoads.length,c=0;b>c;c++)if(this._currentLoads[c]==a){this._currentLoads.splice(c,1);break}},b._handleProgress=function(a){var b=a.target;this._sendFileProgress(b.getItem(),b.progress),this._updateProgress()},b._updateProgress=function(){var a=this._numItemsLoaded/this._numItems,b=this._numItems-this._numItemsLoaded;if(b>0){for(var c=0,d=0,e=this._currentLoads.length;e>d;d++)c+=this._currentLoads[d].progress;a+=c/b*(b/this._numItems)}this._sendProgress(a)},b._disposeItem=function(a){delete this._loadedResults[a.id],delete this._loadedRawResults[a.id],delete this._loadItemsById[a.id],delete this._loadItemsBySrc[a.src]},b._createTag=function(a){var b=null;switch(a.type){case createjs.LoadQueue.IMAGE:return b=document.createElement("img"),""==this._crossOrigin||this._isLocal(a)||(b.crossOrigin=this._crossOrigin),b;case createjs.LoadQueue.SOUND:return b=document.createElement("audio"),b.autoplay=!1,b;case createjs.LoadQueue.JSON:case createjs.LoadQueue.JSONP:case createjs.LoadQueue.JAVASCRIPT:case createjs.LoadQueue.MANIFEST:return b=document.createElement("script"),b.type="text/javascript",b;case createjs.LoadQueue.CSS:return b=this.useXHR?document.createElement("style"):document.createElement("link"),b.rel="stylesheet",b.type="text/css",b;case createjs.LoadQueue.SVG:return this.useXHR?b=document.createElement("svg"):(b=document.createElement("object"),b.type="image/svg+xml"),b}return null},b._getTypeByExtension=function(a){if(null==a)return createjs.LoadQueue.TEXT;switch(a.toLowerCase()){case"jpeg":case"jpg":case"gif":case"png":case"webp":case"bmp":return createjs.LoadQueue.IMAGE;case"ogg":case"mp3":case"wav":return createjs.LoadQueue.SOUND;case"json":return createjs.LoadQueue.JSON;case"xml":return createjs.LoadQueue.XML;case"css":return createjs.LoadQueue.CSS;case"js":return createjs.LoadQueue.JAVASCRIPT;case"svg":return createjs.LoadQueue.SVG;default:return createjs.LoadQueue.TEXT}},b._sendFileProgress=function(a,b){if(this._isCanceled())return this._cleanUp(),void 0;if(this.hasEventListener("fileprogress")){var c=new createjs.Event("fileprogress");c.progress=b,c.loaded=b,c.total=1,c.item=a,this.dispatchEvent(c)}},b._sendFileComplete=function(a,b){if(!this._isCanceled()){var c=new createjs.Event("fileload");c.loader=b,c.item=a,c.result=this._loadedResults[a.id],c.rawResult=this._loadedRawResults[a.id],a.completeHandler&&a.completeHandler(c),this.hasEventListener("fileload")&&this.dispatchEvent(c)}},b._sendFileStart=function(a){var b=new createjs.Event("filestart");b.item=a,this.hasEventListener("filestart")&&this.dispatchEvent(b)},b.toString=function(){return"[PreloadJS LoadQueue]"},createjs.LoadQueue=a;var d=function(){};d.init=function(){var a=navigator.userAgent;d.isFirefox=a.indexOf("Firefox")>-1,d.isOpera=null!=window.opera,d.isChrome=a.indexOf("Chrome")>-1,d.isIOS=a.indexOf("iPod")>-1||a.indexOf("iPhone")>-1||a.indexOf("iPad")>-1},d.init(),createjs.LoadQueue.BrowserDetect=d}(),this.createjs=this.createjs||{},function(){"use strict";var a=function(a){this.init(a)},b=a.prototype=new createjs.AbstractLoader;b._loadTimeout=null,b._tagCompleteProxy=null,b._isAudio=!1,b._tag=null,b._jsonResult=null,b.init=function(a){this._item=a,this._tag=a.tag,this._isAudio=window.HTMLAudioElement&&a.tag instanceof window.HTMLAudioElement,this._tagCompleteProxy=createjs.proxy(this._handleLoad,this)},b.getResult=function(){return this._item.type==createjs.LoadQueue.JSONP||this._item.type==createjs.LoadQueue.MANIFEST?this._jsonResult:this._tag},b.cancel=function(){this.canceled=!0,this._clean()},b.load=function(){var a=this._item,b=this._tag;clearTimeout(this._loadTimeout);var c=createjs.LoadQueue.LOAD_TIMEOUT;0==c&&(c=createjs.LoadQueue.loadTimeout),this._loadTimeout=setTimeout(createjs.proxy(this._handleTimeout,this),c),this._isAudio&&(b.src=null,b.preload="auto"),b.onerror=createjs.proxy(this._handleError,this),this._isAudio?(b.onstalled=createjs.proxy(this._handleStalled,this),b.addEventListener("canplaythrough",this._tagCompleteProxy,!1)):(b.onload=createjs.proxy(this._handleLoad,this),b.onreadystatechange=createjs.proxy(this._handleReadyStateChange,this));var d=this.buildPath(a.src,a.values);switch(a.type){case createjs.LoadQueue.CSS:b.href=d;break;case createjs.LoadQueue.SVG:b.data=d;break;default:b.src=d}if(a.type==createjs.LoadQueue.JSONP||a.type==createjs.LoadQueue.JSON||a.type==createjs.LoadQueue.MANIFEST){if(null==a.callback)throw new Error("callback is required for loading JSONP requests.");if(null!=window[a.callback])throw new Error('JSONP callback "'+a.callback+'" already exists on window. You need to specify a different callback. Or re-name the current one.');window[a.callback]=createjs.proxy(this._handleJSONPLoad,this)}(a.type==createjs.LoadQueue.SVG||a.type==createjs.LoadQueue.JSONP||a.type==createjs.LoadQueue.JSON||a.type==createjs.LoadQueue.MANIFEST||a.type==createjs.LoadQueue.JAVASCRIPT||a.type==createjs.LoadQueue.CSS)&&(this._startTagVisibility=b.style.visibility,b.style.visibility="hidden",(document.body||document.getElementsByTagName("body")[0]).appendChild(b)),null!=b.load&&b.load()},b._handleJSONPLoad=function(a){this._jsonResult=a},b._handleTimeout=function(){this._clean();var a=new createjs.Event("error");a.text="PRELOAD_TIMEOUT",this._sendError(a)},b._handleStalled=function(){},b._handleError=function(){this._clean();var a=new createjs.Event("error");this._sendError(a)},b._handleReadyStateChange=function(){clearTimeout(this._loadTimeout);var a=this.getItem().tag;("loaded"==a.readyState||"complete"==a.readyState)&&this._handleLoad()},b._handleLoad=function(){if(!this._isCanceled()){var a=this.getItem(),b=a.tag;if(!(this.loaded||this._isAudio&&4!==b.readyState)){switch(this.loaded=!0,a.type){case createjs.LoadQueue.SVG:case createjs.LoadQueue.JSON:case createjs.LoadQueue.JSONP:case createjs.LoadQueue.MANIFEST:case createjs.LoadQueue.CSS:b.style.visibility=this._startTagVisibility,(document.body||document.getElementsByTagName("body")[0]).removeChild(b)}this._clean(),this._sendComplete()}}},b._clean=function(){clearTimeout(this._loadTimeout);var a=this.getItem(),b=a.tag;null!=b&&(b.onload=null,b.removeEventListener&&b.removeEventListener("canplaythrough",this._tagCompleteProxy,!1),b.onstalled=null,b.onprogress=null,b.onerror=null,null!=b.parentNode&&a.type==createjs.LoadQueue.SVG&&a.type==createjs.LoadQueue.JSON&&a.type==createjs.LoadQueue.MANIFEST&&a.type==createjs.LoadQueue.CSS&&a.type==createjs.LoadQueue.JSONP&&b.parentNode.removeChild(b));var a=this.getItem();(a.type==createjs.LoadQueue.JSONP||a.type==createjs.LoadQueue.MANIFEST)&&(window[a.callback]=null)},b.toString=function(){return"[PreloadJS TagLoader]"},createjs.TagLoader=a}(),this.createjs=this.createjs||{},function(){"use strict";var a=function(a,b){this.init(a,b)},b=a.prototype=new createjs.AbstractLoader;b._request=null,b._loadTimeout=null,b._xhrLevel=1,b._response=null,b._rawResponse=null,b._crossOrigin="",b.init=function(a,b){this._item=a,this._crossOrigin=b,!this._createXHR(a)},b.getResult=function(a){return a&&this._rawResponse?this._rawResponse:this._response},b.cancel=function(){this.canceled=!0,this._clean(),this._request.abort()},b.load=function(){if(null==this._request)return this._handleError(),void 0;if(this._request.onloadstart=createjs.proxy(this._handleLoadStart,this),this._request.onprogress=createjs.proxy(this._handleProgress,this),this._request.onabort=createjs.proxy(this._handleAbort,this),this._request.onerror=createjs.proxy(this._handleError,this),this._request.ontimeout=createjs.proxy(this._handleTimeout,this),1==this._xhrLevel){var a=createjs.LoadQueue.LOAD_TIMEOUT;if(0==a)a=createjs.LoadQueue.loadTimeout;else try{console.warn("LoadQueue.LOAD_TIMEOUT has been deprecated in favor of LoadQueue.loadTimeout")}catch(b){}this._loadTimeout=setTimeout(createjs.proxy(this._handleTimeout,this),a)}this._request.onload=createjs.proxy(this._handleLoad,this),this._request.onreadystatechange=createjs.proxy(this._handleReadyStateChange,this);try{this._item.values&&this._item.method!=createjs.LoadQueue.GET?this._item.method==createjs.LoadQueue.POST&&this._request.send(this._formatQueryString(this._item.values)):this._request.send()}catch(c){var d=new createjs.Event("error");d.error=c,this._sendError(d)}},b.getAllResponseHeaders=function(){return this._request.getAllResponseHeaders instanceof Function?this._request.getAllResponseHeaders():null},b.getResponseHeader=function(a){return this._request.getResponseHeader instanceof Function?this._request.getResponseHeader(a):null},b._handleProgress=function(a){if(a&&!(a.loaded>0&&0==a.total)){var b=new createjs.Event("progress");b.loaded=a.loaded,b.total=a.total,this._sendProgress(b)}},b._handleLoadStart=function(){clearTimeout(this._loadTimeout),this._sendLoadStart()},b._handleAbort=function(){this._clean();var a=new createjs.Event("error");a.text="XHR_ABORTED",this._sendError(a)},b._handleError=function(){this._clean();var a=new createjs.Event("error");this._sendError(a)},b._handleReadyStateChange=function(){4==this._request.readyState&&this._handleLoad()},b._handleLoad=function(){if(!this.loaded){if(this.loaded=!0,!this._checkError())return this._handleError(),void 0;this._response=this._getResponse(),this._clean();var a=this._generateTag();a&&this._sendComplete()}},b._handleTimeout=function(a){this._clean();var b=new createjs.Event("error");b.text="PRELOAD_TIMEOUT",this._sendError(a)},b._checkError=function(){var a=parseInt(this._request.status);switch(a){case 404:case 0:return!1}return!0},b._getResponse=function(){if(null!=this._response)return this._response;if(null!=this._request.response)return this._request.response;try{if(null!=this._request.responseText)return this._request.responseText}catch(a){}try{if(null!=this._request.responseXML)return this._request.responseXML}catch(a){}return null},b._createXHR=function(a){var b=this._isCrossDomain(a),c=null;if(b&&window.XDomainRequest)c=new XDomainRequest;else if(window.XMLHttpRequest)c=new XMLHttpRequest;else try{c=new ActiveXObject("Msxml2.XMLHTTP.6.0")}catch(d){try{c=new ActiveXObject("Msxml2.XMLHTTP.3.0")}catch(d){try{c=new ActiveXObject("Msxml2.XMLHTTP")}catch(d){return!1}}}createjs.LoadQueue.isText(a.type)&&c.overrideMimeType&&c.overrideMimeType("text/plain; charset=utf-8"),this._xhrLevel="string"==typeof c.responseType?2:1;var e=null;return e=a.method==createjs.LoadQueue.GET?this.buildPath(a.src,a.values):a.src,c.open(a.method||createjs.LoadQueue.GET,e,!0),b&&c instanceof XMLHttpRequest&&1==this._xhrLevel&&c.setRequestHeader("Origin",location.origin),a.values&&a.method==createjs.LoadQueue.POST&&c.setRequestHeader("Content-Type","application/x-www-form-urlencoded"),createjs.LoadQueue.isBinary(a.type)&&(c.responseType="arraybuffer"),this._request=c,!0},b._clean=function(){clearTimeout(this._loadTimeout);var a=this._request;a.onloadstart=null,a.onprogress=null,a.onabort=null,a.onerror=null,a.onload=null,a.ontimeout=null,a.onloadend=null,a.onreadystatechange=null},b._generateTag=function(){var a=this._item.type,b=this._item.tag;switch(a){case createjs.LoadQueue.IMAGE:return b.onload=createjs.proxy(this._handleTagReady,this),""!=this._crossOrigin&&(b.crossOrigin="Anonymous"),b.src=this.buildPath(this._item.src,this._item.values),this._rawResponse=this._response,this._response=b,!1;case createjs.LoadQueue.JAVASCRIPT:return b=document.createElement("script"),b.text=this._response,this._rawResponse=this._response,this._response=b,!0;case createjs.LoadQueue.CSS:var c=document.getElementsByTagName("head")[0];if(c.appendChild(b),b.styleSheet)b.styleSheet.cssText=this._response;else{var d=document.createTextNode(this._response);b.appendChild(d)}return this._rawResponse=this._response,this._response=b,!0;case createjs.LoadQueue.XML:var e=this._parseXML(this._response,"text/xml");return this._rawResponse=this._response,this._response=e,!0;case createjs.LoadQueue.SVG:var e=this._parseXML(this._response,"image/svg+xml");return this._rawResponse=this._response,null!=e.documentElement?(b.appendChild(e.documentElement),this._response=b):this._response=e,!0;case createjs.LoadQueue.JSON:case createjs.LoadQueue.MANIFEST:var f={};try{f=JSON.parse(this._response)}catch(g){f=g}return this._rawResponse=this._response,this._response=f,!0}return!0},b._parseXML=function(a,b){var c=null;try{if(window.DOMParser){var d=new DOMParser;c=d.parseFromString(a,b)}else c=new ActiveXObject("Microsoft.XMLDOM"),c.async=!1,c.loadXML(a)}catch(e){}return c},b._handleTagReady=function(){this._sendComplete()},b.toString=function(){return"[PreloadJS XHRLoader]"},createjs.XHRLoader=a}(),"object"!=typeof JSON&&(JSON={}),function(){"use strict";function f(a){return 10>a?"0"+a:a}function quote(a){return escapable.lastIndex=0,escapable.test(a)?'"'+a.replace(escapable,function(a){var b=meta[a];return"string"==typeof b?b:"\\u"+("0000"+a.charCodeAt(0).toString(16)).slice(-4)})+'"':'"'+a+'"'}function str(a,b){var c,d,e,f,g,h=gap,i=b[a];switch(i&&"object"==typeof i&&"function"==typeof i.toJSON&&(i=i.toJSON(a)),"function"==typeof rep&&(i=rep.call(b,a,i)),typeof i){case"string":return quote(i);case"number":return isFinite(i)?String(i):"null";case"boolean":case"null":return String(i);case"object":if(!i)return"null";if(gap+=indent,g=[],"[object Array]"===Object.prototype.toString.apply(i)){for(f=i.length,c=0;f>c;c+=1)g[c]=str(c,i)||"null";return e=0===g.length?"[]":gap?"[\n"+gap+g.join(",\n"+gap)+"\n"+h+"]":"["+g.join(",")+"]",gap=h,e}if(rep&&"object"==typeof rep)for(f=rep.length,c=0;f>c;c+=1)"string"==typeof rep[c]&&(d=rep[c],e=str(d,i),e&&g.push(quote(d)+(gap?": ":":")+e));else for(d in i)Object.prototype.hasOwnProperty.call(i,d)&&(e=str(d,i),e&&g.push(quote(d)+(gap?": ":":")+e));return e=0===g.length?"{}":gap?"{\n"+gap+g.join(",\n"+gap)+"\n"+h+"}":"{"+g.join(",")+"}",gap=h,e}}"function"!=typeof Date.prototype.toJSON&&(Date.prototype.toJSON=function(){return isFinite(this.valueOf())?this.getUTCFullYear()+"-"+f(this.getUTCMonth()+1)+"-"+f(this.getUTCDate())+"T"+f(this.getUTCHours())+":"+f(this.getUTCMinutes())+":"+f(this.getUTCSeconds())+"Z":null},String.prototype.toJSON=Number.prototype.toJSON=Boolean.prototype.toJSON=function(){return this.valueOf()});var cx=/[\u0000\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,escapable=/[\\\"\x00-\x1f\x7f-\x9f\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,gap,indent,meta={"\b":"\\b","	":"\\t","\n":"\\n","\f":"\\f","\r":"\\r",'"':'\\"',"\\":"\\\\"},rep;"function"!=typeof JSON.stringify&&(JSON.stringify=function(a,b,c){var d;if(gap="",indent="","number"==typeof c)for(d=0;c>d;d+=1)indent+=" ";else"string"==typeof c&&(indent=c);if(rep=b,b&&"function"!=typeof b&&("object"!=typeof b||"number"!=typeof b.length))throw new Error("JSON.stringify");return str("",{"":a})}),"function"!=typeof JSON.parse&&(JSON.parse=function(text,reviver){function walk(a,b){var c,d,e=a[b];if(e&&"object"==typeof e)for(c in e)Object.prototype.hasOwnProperty.call(e,c)&&(d=walk(e,c),void 0!==d?e[c]=d:delete e[c]);return reviver.call(a,b,e)}var j;if(text=String(text),cx.lastIndex=0,cx.test(text)&&(text=text.replace(cx,function(a){return"\\u"+("0000"+a.charCodeAt(0).toString(16)).slice(-4)})),/^[\],:{}\s]*$/.test(text.replace(/\\(?:["\\\/bfnrt]|u[0-9a-fA-F]{4})/g,"@").replace(/"[^"\\\n\r]*"|true|false|null|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?/g,"]").replace(/(?:^|:|,)(?:\s*\[)+/g,"")))return j=eval("("+text+")"),"function"==typeof reviver?walk({"":j},""):j;throw new SyntaxError("JSON.parse")})}();
/*
 * $Id: rawinflate.js,v 0.3 2013/04/09 14:25:38 dankogai Exp dankogai $
 *
 * GNU General Public License, version 2 (GPL-2.0)
 *   http://opensource.org/licenses/GPL-2.0
 * original:
 *   http://www.onicos.com/staff/iz/amuse/javascript/expert/inflate.txt
 */
(function(ctx){

/* Copyright (C) 1999 Masanao Izumo <iz@onicos.co.jp>
 * Version: 1.0.0.1
 * LastModified: Dec 25 1999
 */

/* Interface:
 * data = zip_inflate(src);
 */

/* constant parameters */
var zip_WSIZE = 32768;		// Sliding Window size
var zip_STORED_BLOCK = 0;
var zip_STATIC_TREES = 1;
var zip_DYN_TREES    = 2;

/* for inflate */
var zip_lbits = 9; 		// bits in base literal/length lookup table
var zip_dbits = 6; 		// bits in base distance lookup table
var zip_INBUFSIZ = 32768;	// Input buffer size
var zip_INBUF_EXTRA = 64;	// Extra buffer

/* variables (inflate) */
var zip_slide;
var zip_wp;			// current position in slide
var zip_fixed_tl = null;	// inflate static
var zip_fixed_td;		// inflate static
var zip_fixed_bl, zip_fixed_bd;	// inflate static
var zip_bit_buf;		// bit buffer
var zip_bit_len;		// bits in bit buffer
var zip_method;
var zip_eof;
var zip_copy_leng;
var zip_copy_dist;
var zip_tl, zip_td;	// literal/length and distance decoder tables
var zip_bl, zip_bd;	// number of bits decoded by tl and td

var zip_inflate_data;
var zip_inflate_pos;


/* constant tables (inflate) */
var zip_MASK_BITS = new Uint16Array(
    [0x0000,
    0x0001, 0x0003, 0x0007, 0x000f, 0x001f, 0x003f, 0x007f, 0x00ff,
    0x01ff, 0x03ff, 0x07ff, 0x0fff, 0x1fff, 0x3fff, 0x7fff, 0xffff]);
// Tables for deflate from PKZIP's appnote.txt.
var zip_cplens = new Uint16Array( // Copy lengths for literal codes 257..285
    [3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
    35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0]);
/* note: see note #13 above about the 258 in this list. */
var zip_cplext = new Uint16Array( // Extra bits for literal codes 257..285
    [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2,
    3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0, 99, 99]); // 99==invalid
var zip_cpdist = new Uint16Array( // Copy offsets for distance codes 0..29
    [1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
    257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
    8193, 12289, 16385, 24577]);
var zip_cpdext = new Uint16Array( // Extra bits for distance codes
    [0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,
    7, 7, 8, 8, 9, 9, 10, 10, 11, 11,
    12, 12, 13, 13]);
var zip_border = new Uint16Array(  // Order of the bit length code lengths
    [16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15]);
/* objects (inflate) */

var zip_HuftList = function() {
    this.next = null;
    this.list = null;
};

var zip_HuftNode = function() {
    this.e = 0; // number of extra bits or operation
    this.b = 0; // number of bits in this code or subcode

    // union
    this.n = 0; // literal, length base, or distance base
    this.t = null; // (zip_HuftNode) pointer to next level of table
};

var zip_HuftBuild = function(b,	// code lengths in bits (all assumed <= BMAX)
		       n,	// number of codes (assumed <= N_MAX)
		       s,	// number of simple-valued codes (0..s-1)
		       d,	// list of base values for non-simple codes
		       e,	// list of extra bits for non-simple codes
		       mm	// maximum lookup bits
		   ) {
    this.BMAX = 16;   // maximum bit length of any code
    this.N_MAX = 288; // maximum number of codes in any set
    this.status = 0;	// 0: success, 1: incomplete table, 2: bad input
    this.root = null;	// (zip_HuftList) starting table
    this.m = 0;		// maximum lookup bits, returns actual

/* Given a list of code lengths and a maximum table size, make a set of
   tables to decode that set of codes.	Return zero on success, one if
   the given code set is incomplete (the tables are still built in this
   case), two if the input is invalid (all zero length codes or an
   oversubscribed set of lengths), and three if not enough memory.
   The code with value 256 is special, and the tables are constructed
   so that no bits beyond that code are fetched when that code is
   decoded. */
    {
	var a;			// counter for codes of length k
	var c = new Uint16Array(this.BMAX+1);	// bit length count table
	var el;			// length of EOB code (value 256)
	var f;			// i repeats in table every f entries
	var g;			// maximum code length
	var h;			// table level
	var i;			// counter, current code
	var j;			// counter
	var k;			// number of bits in current code
	var lx = new Uint16Array(this.BMAX+1);	// stack of bits per table
	var p;			// pointer into c[], b[], or v[]
	var pidx;		// index of p
	var q;			// (zip_HuftNode) points to current table
	var r = new zip_HuftNode(); // table entry for structure assignment
	var u = new Array(this.BMAX); // zip_HuftNode[BMAX][]  table stack
	var v = new Uint16Array(this.N_MAX); // values in order of bit length
	var w;
	var x = new Uint16Array(this.BMAX+1);// bit offsets, then code stack
	var xp;			// pointer into x or c
	var y;			// number of dummy codes added
	var z;			// number of entries in current table
	var o;
	var tail;		// (zip_HuftList)

	tail = this.root = null;
	for(i = 0; i < c.length; i++)
	    c[i] = 0;
	for(i = 0; i < lx.length; i++)
	    lx[i] = 0;
	for(i = 0; i < u.length; i++)
	    u[i] = null;
	for(i = 0; i < v.length; i++)
	    v[i] = 0;
	for(i = 0; i < x.length; i++)
	    x[i] = 0;

	// Generate counts for each bit length
	el = n > 256 ? b[256] : this.BMAX; // set length of EOB code, if any
	p = b; pidx = 0;
	i = n;
	do {
	    c[p[pidx]]++;	// assume all entries <= BMAX
	    pidx++;
	} while(--i > 0);
	if(c[0] == n) {	// null input--all zero length codes
	    this.root = null;
	    this.m = 0;
	    this.status = 0;
	    return;
	}

	// Find minimum and maximum length, bound *m by those
	for(j = 1; j <= this.BMAX; j++)
	    if(c[j] != 0)
		break;
	k = j;			// minimum code length
	if(mm < j)
	    mm = j;
	for(i = this.BMAX; i != 0; i--)
	    if(c[i] != 0)
		break;
	g = i;			// maximum code length
	if(mm > i)
	    mm = i;

	// Adjust last length count to fill out codes, if needed
	for(y = 1 << j; j < i; j++, y <<= 1)
	    if((y -= c[j]) < 0) {
		this.status = 2;	// bad input: more codes than bits
		this.m = mm;
		return;
	    }
	if((y -= c[i]) < 0) {
	    this.status = 2;
	    this.m = mm;
	    return;
	}
	c[i] += y;

	// Generate starting offsets into the value table for each length
	x[1] = j = 0;
	p = c;
	pidx = 1;
	xp = 2;
	while(--i > 0)		// note that i == g from above
	    x[xp++] = (j += p[pidx++]);

	// Make a table of values in order of bit lengths
	p = b; pidx = 0;
	i = 0;
	do {
	    if((j = p[pidx++]) != 0)
		v[x[j]++] = i;
	} while(++i < n);
	n = x[g];			// set n to length of v

	// Generate the Huffman codes and for each, make the table entries
	x[0] = i = 0;		// first Huffman code is zero
	p = v; pidx = 0;		// grab values in bit order
	h = -1;			// no tables yet--level -1
	w = lx[0] = 0;		// no bits decoded yet
	q = null;			// ditto
	z = 0;			// ditto

	// go through the bit lengths (k already is bits in shortest code)
	for(; k <= g; k++) {
	    a = c[k];
	    while(a-- > 0) {
		// here i is the Huffman code of length k bits for value p[pidx]
		// make tables up to required level
		while(k > w + lx[1 + h]) {
		    w += lx[1 + h]; // add bits already decoded
		    h++;

		    // compute minimum size table less than or equal to *m bits
		    z = (z = g - w) > mm ? mm : z; // upper limit
		    if((f = 1 << (j = k - w)) > a + 1) { // try a k-w bit table
			// too few codes for k-w bit table
			f -= a + 1;	// deduct codes from patterns left
			xp = k;
			while(++j < z) { // try smaller tables up to z bits
			    if((f <<= 1) <= c[++xp])
				break;	// enough codes to use up j bits
			    f -= c[xp];	// else deduct codes from patterns
			}
		    }
		    if(w + j > el && w < el)
			j = el - w;	// make EOB code end at table
		    z = 1 << j;	// table entries for j-bit table
		    lx[1 + h] = j; // set table size in stack

		    // allocate and link in new table
		    q = new Array(z);
		    for(o = 0; o < z; o++) {
			q[o] = new zip_HuftNode();
		    }

		    if(tail == null)
			tail = this.root = new zip_HuftList();
		    else
			tail = tail.next = new zip_HuftList();
		    tail.next = null;
		    tail.list = q;
		    u[h] = q;	// table starts after link

		    /* connect to last table, if there is one */
		    if(h > 0) {
			x[h] = i;		// save pattern for backing up
			r.b = lx[h];	// bits to dump before this table
			r.e = 16 + j;	// bits in this table
			r.t = q;		// pointer to this table
			j = (i & ((1 << w) - 1)) >> (w - lx[h]);
			u[h-1][j].e = r.e;
			u[h-1][j].b = r.b;
			u[h-1][j].n = r.n;
			u[h-1][j].t = r.t;
		    }
		}

		// set up table entry in r
		r.b = k - w;
		if(pidx >= n)
		    r.e = 99;		// out of values--invalid code
		else if(p[pidx] < s) {
		    r.e = (p[pidx] < 256 ? 16 : 15); // 256 is end-of-block code
		    r.n = p[pidx++];	// simple code is just the value
		} else {
		    r.e = e[p[pidx] - s];	// non-simple--look up in lists
		    r.n = d[p[pidx++] - s];
		}

		// fill code-like entries with r //
		f = 1 << (k - w);
		for(j = i >> w; j < z; j += f) {
		    q[j].e = r.e;
		    q[j].b = r.b;
		    q[j].n = r.n;
		    q[j].t = r.t;
		}

		// backwards increment the k-bit code i
		for(j = 1 << (k - 1); (i & j) != 0; j >>= 1)
		    i ^= j;
		i ^= j;

		// backup over finished tables
		while((i & ((1 << w) - 1)) != x[h]) {
		    w -= lx[h];		// don't need to update q
		    h--;
		}
	    }
	}

	/* return actual size of base table */
	this.m = lx[1];

	/* Return true (1) if we were given an incomplete table */
	this.status = ((y != 0 && g != 1) ? 1 : 0);
    } /* end of constructor */
};


/* routines (inflate) */

var zip_GET_BYTE;

var zip_GET_BYTE_String = function() {
    if(zip_inflate_data.length == zip_inflate_pos)
	return -1;
    return zip_inflate_data.charCodeAt(zip_inflate_pos++) & 0xff;
};

var zip_GET_BYTE_Array = function() {
	if(zip_inflate_data.length == zip_inflate_pos)
	return -1;
	return zip_inflate_data[zip_inflate_pos++] & 0xff;
};

var zip_NEEDBITS = function(n) {
    while(zip_bit_len < n) {
	zip_bit_buf |= zip_GET_BYTE() << zip_bit_len;
	zip_bit_len += 8;
    }
};

var zip_GETBITS = function(n) {
    return zip_bit_buf & zip_MASK_BITS[n];
};

var zip_DUMPBITS = function(n) {
    zip_bit_buf >>= n;
    zip_bit_len -= n;
};

var zip_inflate_codes = function(buff, off, size) {
    /* inflate (decompress) the codes in a deflated (compressed) block.
       Return an error code or zero if it all goes ok. */
    var e;		// table entry flag/number of extra bits
    var t;		// (zip_HuftNode) pointer to table entry
    var n;

    if(size == 0)
      return 0;

    // inflate the coded data
    n = 0;
    for(;;) {			// do until end of block
	zip_NEEDBITS(zip_bl);
	t = zip_tl.list[zip_GETBITS(zip_bl)];
	e = t.e;
	while(e > 16) {
	    if(e == 99)
		return -1;
	    zip_DUMPBITS(t.b);
	    e -= 16;
	    zip_NEEDBITS(e);
	    t = t.t[zip_GETBITS(e)];
	    e = t.e;
	}
	zip_DUMPBITS(t.b);

	if(e == 16) {		// then it's a literal
	    zip_wp &= zip_WSIZE - 1;
	    buff[off + n++] = zip_slide[zip_wp++] = t.n;
	    if(n == size)
		return size;
	    continue;
	}

	// exit if end of block
	if(e == 15)
	    break;

	// it's an EOB or a length

	// get length of block to copy
	zip_NEEDBITS(e);
	zip_copy_leng = t.n + zip_GETBITS(e);
	zip_DUMPBITS(e);

	// decode distance of block to copy
	zip_NEEDBITS(zip_bd);
	t = zip_td.list[zip_GETBITS(zip_bd)];
	e = t.e;

	while(e > 16) {
	    if(e == 99)
		return -1;
	    zip_DUMPBITS(t.b);
	    e -= 16;
	    zip_NEEDBITS(e);
	    t = t.t[zip_GETBITS(e)];
	    e = t.e;
	}
	zip_DUMPBITS(t.b);
	zip_NEEDBITS(e);
	zip_copy_dist = zip_wp - t.n - zip_GETBITS(e);
	zip_DUMPBITS(e);

	// do the copy
	while(zip_copy_leng > 0 && n < size) {
	    zip_copy_leng--;
	    zip_copy_dist &= zip_WSIZE - 1;
	    zip_wp &= zip_WSIZE - 1;
	    buff[off + n++] = zip_slide[zip_wp++]
		= zip_slide[zip_copy_dist++];
	}

	if(n == size)
	    return size;
    }

    zip_method = -1; // done
    return n;
};

var zip_inflate_stored = function(buff, off, size) {
    /* "decompress" an inflated type 0 (stored) block. */
    var n;

    // go to byte boundary
    n = zip_bit_len & 7;
    zip_DUMPBITS(n);

    // get the length and its complement
    zip_NEEDBITS(16);
    n = zip_GETBITS(16);
    zip_DUMPBITS(16);
    zip_NEEDBITS(16);
    if(n != ((~zip_bit_buf) & 0xffff))
	return -1;			// error in compressed data
    zip_DUMPBITS(16);

    // read and output the compressed data
    zip_copy_leng = n;

    n = 0;
    while(zip_copy_leng > 0 && n < size) {
	zip_copy_leng--;
	zip_wp &= zip_WSIZE - 1;
	zip_NEEDBITS(8);
	buff[off + n++] = zip_slide[zip_wp++] =
	    zip_GETBITS(8);
	zip_DUMPBITS(8);
    }

    if(zip_copy_leng == 0)
      zip_method = -1; // done
    return n;
};

var zip_inflate_fixed = function(buff, off, size) {
    /* decompress an inflated type 1 (fixed Huffman codes) block.  We should
       either replace this with a custom decoder, or at least precompute the
       Huffman tables. */

    // if first time, set up tables for fixed blocks
    if(zip_fixed_tl == null) {
	var i;			// temporary variable
	var l = new Uint16Array(288);	// length list for huft_build
	var h;	// zip_HuftBuild

	// literal table
	for(i = 0; i < 144; i++)
	    l[i] = 8;
	for(; i < 256; i++)
	    l[i] = 9;
	for(; i < 280; i++)
	    l[i] = 7;
	for(; i < 288; i++)	// make a complete, but wrong code set
	    l[i] = 8;
	zip_fixed_bl = 7;

	h = new zip_HuftBuild(l, 288, 257, zip_cplens, zip_cplext,
			      zip_fixed_bl);
	if(h.status != 0) {
	    //alert("HufBuild error: "+h.status);
	    return -1;
	}
	zip_fixed_tl = h.root;
	zip_fixed_bl = h.m;

	// distance table
	for(i = 0; i < 30; i++)	// make an incomplete code set
	    l[i] = 5;
	zip_fixed_bd = 5;

	h = new zip_HuftBuild(l, 30, 0, zip_cpdist, zip_cpdext, zip_fixed_bd);
	if(h.status > 1) {
	    zip_fixed_tl = null;
	    //alert("HufBuild error: "+h.status);
	    return -1;
	}
	zip_fixed_td = h.root;
	zip_fixed_bd = h.m;
    }

    zip_tl = zip_fixed_tl;
    zip_td = zip_fixed_td;
    zip_bl = zip_fixed_bl;
    zip_bd = zip_fixed_bd;
    return zip_inflate_codes(buff, off, size);
};

var zip_inflate_dynamic = function(buff, off, size) {
    // decompress an inflated type 2 (dynamic Huffman codes) block.
    var i;		// temporary variables
    var j;
    var l;		// last length
    var n;		// number of lengths to get
    var t;		// (zip_HuftNode) literal/length code table
    var nb;		// number of bit length codes
    var nl;		// number of literal/length codes
    var nd;		// number of distance codes
    var ll = new Uint16Array(286+30); // literal/length and distance code lengths
    var h;		// (zip_HuftBuild)

    for(i = 0; i < ll.length; i++)
	ll[i] = 0;

    // read in table lengths
    zip_NEEDBITS(5);
    nl = 257 + zip_GETBITS(5);	// number of literal/length codes
    zip_DUMPBITS(5);
    zip_NEEDBITS(5);
    nd = 1 + zip_GETBITS(5);	// number of distance codes
    zip_DUMPBITS(5);
    zip_NEEDBITS(4);
    nb = 4 + zip_GETBITS(4);	// number of bit length codes
    zip_DUMPBITS(4);
    if(nl > 286 || nd > 30)
      return -1;		// bad lengths

    // read in bit-length-code lengths
    for(j = 0; j < nb; j++)
    {
	zip_NEEDBITS(3);
	ll[zip_border[j]] = zip_GETBITS(3);
	zip_DUMPBITS(3);
    }
    for(; j < 19; j++)
	ll[zip_border[j]] = 0;

    // build decoding table for trees--single level, 7 bit lookup
    zip_bl = 7;
    h = new zip_HuftBuild(ll, 19, 19, null, null, zip_bl);
    if(h.status != 0)
	return -1;	// incomplete code set

    zip_tl = h.root;
    zip_bl = h.m;

    // read in literal and distance code lengths
    n = nl + nd;
    i = l = 0;
    while(i < n) {
	zip_NEEDBITS(zip_bl);
	t = zip_tl.list[zip_GETBITS(zip_bl)];
	j = t.b;
	zip_DUMPBITS(j);
	j = t.n;
	if(j < 16)		// length of code in bits (0..15)
	    ll[i++] = l = j;	// save last length in l
	else if(j == 16) {	// repeat last length 3 to 6 times
	    zip_NEEDBITS(2);
	    j = 3 + zip_GETBITS(2);
	    zip_DUMPBITS(2);
	    if(i + j > n)
		return -1;
	    while(j-- > 0)
		ll[i++] = l;
	} else if(j == 17) {	// 3 to 10 zero length codes
	    zip_NEEDBITS(3);
	    j = 3 + zip_GETBITS(3);
	    zip_DUMPBITS(3);
	    if(i + j > n)
		return -1;
	    while(j-- > 0)
		ll[i++] = 0;
	    l = 0;
	} else {		// j == 18: 11 to 138 zero length codes
	    zip_NEEDBITS(7);
	    j = 11 + zip_GETBITS(7);
	    zip_DUMPBITS(7);
	    if(i + j > n)
		return -1;
	    while(j-- > 0)
		ll[i++] = 0;
	    l = 0;
	}
    }

    // build the decoding tables for literal/length and distance codes
    zip_bl = zip_lbits;
    h = new zip_HuftBuild(ll, nl, 257, zip_cplens, zip_cplext, zip_bl);
    if(zip_bl == 0)	// no literals or lengths
	h.status = 1;
    if(h.status != 0) {
	if(h.status == 1)
	    ;// **incomplete literal tree**
	return -1;		// incomplete code set
    }
    zip_tl = h.root;
    zip_bl = h.m;

    for(i = 0; i < nd; i++)
	ll[i] = ll[i + nl];
    zip_bd = zip_dbits;
    h = new zip_HuftBuild(ll, nd, 0, zip_cpdist, zip_cpdext, zip_bd);
    zip_td = h.root;
    zip_bd = h.m;

    if(zip_bd == 0 && nl > 257) {   // lengths but no distances
	// **incomplete distance tree**
	return -1;
    }

    if(h.status == 1) {
	// **incomplete distance tree**
    }
    if(h.status != 0)
	return -1;

    // decompress until an end-of-block code
    return zip_inflate_codes(buff, off, size);
};

var zip_inflate_start = function() {
    if(zip_slide == null)
		zip_slide = new Uint16Array(2 * zip_WSIZE);
    zip_wp = 0;
    zip_bit_buf = 0;
    zip_bit_len = 0;
    zip_method = -1;
    zip_eof = false;
    zip_copy_leng = zip_copy_dist = 0;
    zip_tl = null;
};

var zip_inflate_internal = function(buff, off, size) {
    // decompress an inflated entry
    var n, i;

    n = 0;
    while(n < size) {
	if(zip_eof && zip_method == -1)
	    return n;

	if(zip_copy_leng > 0) {
	    if(zip_method != zip_STORED_BLOCK) {
		// STATIC_TREES or DYN_TREES
		while(zip_copy_leng > 0 && n < size) {
		    zip_copy_leng--;
		    zip_copy_dist &= zip_WSIZE - 1;
		    zip_wp &= zip_WSIZE - 1;
		    buff[off + n++] = zip_slide[zip_wp++] =
			zip_slide[zip_copy_dist++];
		}
	    } else {
		while(zip_copy_leng > 0 && n < size) {
		    zip_copy_leng--;
		    zip_wp &= zip_WSIZE - 1;
		    zip_NEEDBITS(8);
		    buff[off + n++] = zip_slide[zip_wp++] = zip_GETBITS(8);
		    zip_DUMPBITS(8);
		}
		if(zip_copy_leng == 0)
		    zip_method = -1; // done
	    }
	    if(n == size)
		return n;
	}

	if(zip_method == -1) {
	    if(zip_eof)
		break;

	    // read in last block bit
	    zip_NEEDBITS(1);
	    if(zip_GETBITS(1) != 0)
		zip_eof = true;
	    zip_DUMPBITS(1);

	    // read in block type
	    zip_NEEDBITS(2);
	    zip_method = zip_GETBITS(2);
	    zip_DUMPBITS(2);
	    zip_tl = null;
	    zip_copy_leng = 0;
	}

	switch(zip_method) {
	  case 0: // zip_STORED_BLOCK
	    i = zip_inflate_stored(buff, off + n, size - n);
	    break;

	  case 1: // zip_STATIC_TREES
	    if(zip_tl != null)
		i = zip_inflate_codes(buff, off + n, size - n);
	    else
		i = zip_inflate_fixed(buff, off + n, size - n);
	    break;

	  case 2: // zip_DYN_TREES
	    if(zip_tl != null)
		i = zip_inflate_codes(buff, off + n, size - n);
	    else
		i = zip_inflate_dynamic(buff, off + n, size - n);
	    break;

	  default: // error
	    i = -1;
	    break;
	}

	if(i == -1) {
	    if(zip_eof)
		return 0;
	    return -1;
	}
	n += i;
    }
    return n;
};

var zip_inflate_array = function(str) {
	var i, ii, j, n = 0, l;

	zip_GET_BYTE = zip_GET_BYTE_Array;
	
    zip_inflate_start();
    zip_inflate_data = str;
    zip_inflate_pos = 0;
    
	var buffLength = 1024;
	var buff = new Uint8Array(buffLength);
    var aBuff = [];
	
    while((i = zip_inflate_internal(buff, 0, buffLength)) > 0) {
		aBuff.push(buff.subarray(0,i));
		buff = new Uint8Array(buffLength);
		n += i;
    }
		
    zip_inflate_data = null; // G.C.
	
	l = aBuff.length;
	
	var a = new Uint8Array(n);

	for(i=0, ii=0; i<l; i++){
		buff = aBuff[i];
		n = buff.length;
		for(j=0; j<n; j++){
			a[ii++] = buff[j];
		}
		buff = null;
	}
	
	aBuff = null;
	
	return a;
};

var zip_inflate = function(str) {
	if(str instanceof Uint8Array){
		return zip_inflate_array(str);
	}
	
    var i, j;

    zip_inflate_start();
    zip_inflate_data = str;
    zip_inflate_pos = 0;

    var buff = new Array(1024);
    var aout = [];
    while((i = zip_inflate_internal(buff, 0, buff.length)) > 0) {
		var cbuf = new Array(i);
		for(j = 0; j < i; j++){
			cbuf[j] = String.fromCharCode(buff[j]);
		}
		aout[aout.length] = (str instanceof Uint8Array) ? cbuf : cbuf.join("");
    }
    zip_inflate_data = null; // G.C.
	
    return aout.join("");
};

if (! ctx.RawDeflate) ctx.RawDeflate = {};
ctx.RawDeflate.inflate = zip_inflate;

})(this);

//=======================================��� 6Bit-��������������, ��������� Base64=======================================
function StringDecoder6Bit(){
	
	const ALPHABET_CHAR_CODES = 
					new Uint8Array(
				[65,   66,  67,  68,  69,  70,  71,  72,
				73,   74,  75,  76,  77,  78,  79,  80,
				81,   82,  83,  84,  85,  86,  87,  88,
				89,   90,  97,  98,  99, 100, 101, 102,
				103, 104, 105, 106, 107, 108, 109, 110,
				111, 112, 113, 114, 115, 116, 117, 118,
				119, 120, 121, 122,  48,  49,  50,  51,
				52,   53,  54,  55,  56,  57,  43,  47]
								);
								
	const INVERSE_CHAR_CODES = new Uint8Array(256);

	const aMask6BitDecode = [new Uint8Array([0x3F, 0x30]), new Uint8Array([0x0F, 0x3C]), new Uint8Array([0x03, 0x3F])];
	const aShiftL6Bit = new Uint8Array([2, 4, 6]);
	const aShiftR6Bit = new Uint8Array([4, 2, 0]);
	
	//
	function stringDecoder6Bit(){
		initInverseCharCodes();
	}
	
	//
	function initInverseCharCodes(){
		var i, l = ALPHABET_CHAR_CODES.length;
		for(i=0; i<l; i++){
			INVERSE_CHAR_CODES[ALPHABET_CHAR_CODES[i]] = i;
		}
	}
	
	//
	function decode6Bit(bA){
		var l = bA.length;
		var bAOut = new Uint8Array(l);
		var i, q, qOut=0, inPos=0, outPos=0, j, nRest, ii;
		var aMask = aMask6BitDecode, aShiftL = aShiftL6Bit, aShiftR = aShiftR6Bit;
		if(l>0){
			q = INVERSE_CHAR_CODES[bA[inPos++]];
			for(j=0, ii=0; ; ii=1-ii){
				if(ii==0){
					qOut += (q & aMask[j][0]) << aShiftL[j];
					nRest = 0;
				}else{
					qOut += (q & aMask[j][1]) >> aShiftR[j];
					bAOut[outPos++] = qOut;
					qOut = 0;
					nRest = aShiftR[j];
					j = (j==2) ? 0 : (j+1);
				}
				if(nRest==0){
					if(inPos<l){
						q = INVERSE_CHAR_CODES[bA[inPos++]];
					}else{
						break;
					}
				}
			}
		}
		if(outPos!=l){
			var bAOutTmp = bAOut.subarray(0, outPos);
			bAOut = null;
			bAOut = bAOutTmp;
		}
		
		return bAOut;
	}
	
	//
	stringDecoder6Bit.prototype.decodeString6Bit = function(s){
		var i, l = s.length;
		var bA = new Uint8Array(l);
		for(i=0; i<l; i++){
			bA[i] = s.charCodeAt(i);
		}
		var bAOut = decode6Bit(bA);
		return bAOut;
	};
	
	return new stringDecoder6Bit();
}

const stringDecoder6Bit = new StringDecoder6Bit();
// Mersenne Twister,
/*
   A C-program for MT19937, with initialization improved 2002/1/26.
   Coded by Takuji Nishimura and Makoto Matsumoto.
   Before using, initialize the state by using init_genrand(seed)
   or init_by_array(init_key, key_length).
   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.
     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
     3. The names of its contributors may not be used to endorse or promote
        products derived from this software without specific prior written
        permission.
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   Any feedback is very welcome.
   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
*/

function RandomGenerator(seed) {

  var N = 624;
  var M = 397;
  var MATRIX_A = 0x9908b0df;   /* constant vector a */
  var UPPER_MASK = 0x80000000; /* most significant w-r bits */
  var LOWER_MASK = 0x7fffffff; /* least significant r bits */
  
  var mt = new Array(N);   /* the array for the state vector  */
  var mti = N+1;           /* mti==N+1 means mt[N] is not initialized */
  
  var cnt = 0;

  function unsigned32 (n1) // returns a 32-bits unsiged integer from an operand to which applied a bit operator.
  {
	return n1 < 0 ? (n1 ^ UPPER_MASK) + UPPER_MASK : n1;
  }

  function subtraction32 (n1, n2) // emulates lowerflow of a c 32-bits unsiged integer variable, instead of the operator -. these both arguments must be non-negative integers expressible using unsigned 32 bits.
  {
	return n1 < n2 ? unsigned32((0x100000000 - (n2 - n1)) & 0xffffffff) : n1 - n2;
  }

  function addition32 (n1, n2) // emulates overflow of a c 32-bits unsiged integer variable, instead of the operator +. these both arguments must be non-negative integers expressible using unsigned 32 bits.
  {
	return unsigned32((n1 + n2) & 0xffffffff);
  }

  function multiplication32 (n1, n2) // emulates overflow of a c 32-bits unsiged integer variable, instead of the operator *. these both arguments must be non-negative integers expressible using unsigned 32 bits.
  {
	var sum = 0;
	for (var i = 0; i < 32; ++i){
	  if ((n1 >>> i) & 0x1){
		sum = addition32(sum, unsigned32(n2 << i));
	  }
	}
	return sum;
  }

  /* initializes mt[N] with a seed */
  function init_genrand(s) {
	mt[0]= unsigned32(s & 0xffffffff);
	for (mti=1; mti<N; mti++) {
	  mt[mti] =
	  addition32(multiplication32(1812433253, unsigned32(mt[mti-1] ^ (mt[mti-1] >>> 30))), mti);
	  mt[mti] = unsigned32(mt[mti] & 0xffffffff);
	}
  }

  /* initialize by an array with array-length */
  /* init_key is the array for initializing keys */
  /* key_length is its length */
  function init_by_array(init_key, key_length) {
	var i, j, k;
	init_genrand(19650218);
	i=1; j=0;
	k = (N>key_length ? N : key_length);
	for (; k; k--) {
	  mt[i] = addition32(addition32(unsigned32(mt[i] ^ multiplication32(unsigned32(mt[i-1] ^ (mt[i-1] >>> 30)), 1664525)), init_key[j]), j);
	  mt[i] =
	  unsigned32(mt[i] & 0xffffffff);
	  i++; j++;
	  if (i>=N) { mt[0] = mt[N-1]; i=1; }
	  if (j>=key_length) {
		j=0;
	  }
	}
	for (k=N-1; k; k--) {
	  mt[i] = subtraction32(unsigned32((mt[i]) ^ multiplication32(unsigned32(mt[i-1] ^ (mt[i-1] >>> 30)), 1566083941)), i);
	  mt[i] = unsigned32(mt[i] & 0xffffffff);
	  i++;
	  if (i>=N) { mt[0] = mt[N-1]; i=1; }
	}
	mt[0] = 0x80000000; /* MSB is 1; assuring non-zero initial array */
  }

  /* generates a random number on [0,0xffffffff]-interval */
  function genrand_int32() {
	var y;
	var mag01 = [0x0, MATRIX_A];
	if (mti >= N) { /* generate N words at one time */
	  var kk;
	  if (mti == N+1) {   /* if init_genrand() has not been called, */
		init_genrand(Date.now()); /* a default initial seed is used */
	  }

	  for (kk=0;kk<N-M;kk++) {
		y = unsigned32((mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK));
		mt[kk] = unsigned32(mt[kk+M] ^ (y >>> 1) ^ mag01[y & 0x1]);
	  }
	  for (;kk<N-1;kk++) {
		y = unsigned32((mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK));
		mt[kk] = unsigned32(mt[kk+(M-N)] ^ (y >>> 1) ^ mag01[y & 0x1]);
	  }
	  y = unsigned32((mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK));
	  mt[N-1] = unsigned32(mt[M-1] ^ (y >>> 1) ^ mag01[y & 0x1]);
	  mti = 0;
	}

	y = mt[mti++];

	/* Tempering */
	y = unsigned32(y ^ (y >>> 11));
	y = unsigned32(y ^ ((y << 7) & 0x9d2c5680));
	y = unsigned32(y ^ ((y << 15) & 0xefc60000));
	y = unsigned32(y ^ (y >>> 18));

	cnt++;
	
	return y;
  }

  /* generates a random number on [0,0x7fffffff]-interval */
  function genrand_int31() {
	return (genrand_int32()>>>1);
  }

  /* generates a random number on [0,1]-real-interval */
  function genrand_real1() {
	return genrand_int32()*(1.0/4294967295.0);
  }

  /* generates a random number on [0,1)-real-interval */
  function genrand_real2() {
	return genrand_int32()*(1.0/4294967296.0);
	/* divided by 2^32 */
  }

  /* generates a random number on (0,1)-real-interval */
  //c//double genrand_real3(void)
  function genrand_real3() {
	return ((genrand_int32()) + 0.5)*(1.0/4294967296.0);
	/* divided by 2^32 */
  }

  /* generates a random number on [0,1) with 53-bit resolution*/
  function genrand_res53() {
	var a=genrand_int32()>>>5, b=genrand_int32()>>>6;
	return (a*67108864.0+b)*(1.0/9007199254740992.0);
  }
  /* These real versions are due to Isaku Wada, 2002/01/09 added */

  var random = genrand_real2;

  random.seed = function (seed) {
	if (! seed) {
	  seed = Date.now();
	}
	if (typeof seed != "number") {
	  seed = parseInt(seed, 10);
	}
	if ((seed !== 0 && ! seed) || isNaN(seed)) {
	  throw "Bad seed";
	}
	init_genrand(seed);
	cnt = 0;
  };

  random.count = function(){
	  return cnt;
  }
  
  random.seed(seed);

  random.int32 = genrand_int32;
  random.int31 = genrand_int31;
  random.real1 = genrand_real1;
  random.real2 = genrand_real2;
  random.real3 = genrand_real3;
  random.res53 = genrand_res53;

  return random;
}

																										
function MiscFunctions(){
	this.TWO_PI = 2*Math.PI;
	this.PI_OVER_180 = Math.PI/180;//0.017453292519943295769236907684886;
	this.PI_OVER_6 = Math.PI/6;
	this.PI_OVER_3 = Math.PI/3;
	this.PI_OVER_2 = Math.PI/2;
	this.THREE_PI_OVER_2 = 3*Math.PI/2;
	this.ONE_OVER_SQRT5 =  0.44721359549995793928183473374626;
	this.SQRT2 =  1.4142135623730950488016887242097;
	this.ONE_OVER_SQRT2 =  0.70710678118654752440084436210485;
	this.SQRT_3 = 1.7320508075688772935274463415059;
	this.ONE_OVER_THREE = 1/3;
	this.TWO_OVER_THREE = 2/3;
	this.ONE_OVER_SIX = 1/6;
	this.BITMAP_FILL_PREFIX = "bitmap::";
	//для p=0.050
	this.chi2CriticalValues = [
		3.84146, 5.99146, 7.81473, 9.48773, 11.0705, 12.59159, 14.06714, 15.50731, 16.91898, 18.30704,
		19.67514, 21.02607, 22.36203, 23.68479, 24.99579, 26.29623, 27.58711, 28.8693, 30.14353, 31.41043,
		32.67057, 33.92444, 35.17246, 36.41503, 37.65248, 38.88514, 40.11327, 41.33714, 42.55697, 43.77297
	];
	//Pixels Per MilliMeter, hPPMM, vPPMM
	var testDiv = document.createElement("DIV");
	testDiv.style.position = "absolute";
	testDiv.style.backgroundColor = "#0000FF";
	testDiv.style.left = "0px";
	testDiv.style.width = "5cm";
	testDiv.style.height = "5cm";
	document.body.appendChild(testDiv);
	this.hPPMM = testDiv.offsetWidth/50.0;
	this.vPPMM = testDiv.offsetHeight/50.0;
	document.body.removeChild(testDiv);
	//
	this.nRandomSequence = 100;
	this.aRandomSequence = null;
	this.randomSequenceLength = 0;
	this.randomGenerator = new RandomGenerator(Math.abs(new Date().getTime()));
	//
	this.aCustomPattern = [];
}

MiscFunctions.prototype.MouseEventName = {
	mousedown:	["pointerdown", 	"MSPointerDown",	"mousedown touchstart"],
	mousemove:	["pointermove",		"MSPointerMove",	"mousemove touchmove"],
	mousecancel:["pointercancel",	"MSPointerCancel",	"touchcancel"],
	mouseup:	["pointerup",		"MSPointerUp",		"mouseup touchend"],
	mouseover:	["pointerover",		"MSPointerHover",	"mouseover"],
	mouseout:	["pointerout",		"MSPointerOut",		"mouseout"],
	mouseleave: ["pointerleave",	"MSPointerLeave",	"mouseleave"]
};

MiscFunctions.prototype.aNonCloneable = [HTMLImageElement];

//Функция получения случайного числа
MiscFunctions.prototype.random = function(){
	return this.randomGenerator.real2();
};

//
MiscFunctions.prototype.randomize = function(_seed){
	this.randomGenerator.seed((typeof _seed == TYPE_NUMBER) ? _seed : Math.abs(new Date().getTime()));
};

//функция для сортировки массива по возрастанию
MiscFunctions.prototype.NUMERIC = function(v1, v2){
	return (v1 == v2) ? 0 : ((v1<v2) ? -1 : 1);
};

//milliseconds
MiscFunctions.prototype.getTime = function(){
	return new Date().getTime();
};

//
MiscFunctions.prototype.getElementByIdAndLocation = function(sName, sLocation){
	if(typeof this.modelMain == TYPE_UNDEFINED)
		return null;
	return this.modelMain.getElementByIdAndLocation.call(this.modelMain, sName, sLocation);
};

//
MiscFunctions.prototype.getElementBySetupNameAndLocation = function(sName, sLocation){
	if(typeof this.modelMain == TYPE_UNDEFINED)
		return null;
	return this.modelMain.getElementBySetupNameAndLocation.call(this.modelMain, sName, sLocation);
};

//хранилище текстур из модели
MiscFunctions.prototype.setCustomPatterns = function(a){
	var i, n, obj, objStore;
	this.aCustomPattern.splice(0);
	if(typeof a != TYPE_UNDEFINED && a != null){
		n = a.length;
		for(i=0; i<n; i++){
			obj = a[i];
			objStore = {id:obj.id, width:obj.width, height:obj.height, transparent:obj.transparent, keep:obj.keep, canvas:obj.canvas};
			this.aCustomPattern.push(objStore);
		}
	}
};

//является ли данный идентификатор идентификатором текстуры
MiscFunctions.prototype.isCustomPatternId = function(sId){
	return (this.getCustomPattern(sId) != null);
};

//
MiscFunctions.prototype.getCustomPattern = function(sId){
	var i, n = this.aCustomPattern.length, obj;
	for(i=0; i<n; i++){
		obj = this.aCustomPattern[i];
		if(obj.id == sId)
			break;
	}
	return (i<n) ? obj : null;
};

//
MiscFunctions.prototype.getCustomPatternIds = function(){
	var a = [], i, n = this.aCustomPattern.length;
	for(i=0; i<n; i++)
		a.push(this.aCustomPattern[i].id);
	return a;
};

//первую букву сделать заглавной
MiscFunctions.prototype.capitalize = function(s){
	return ((s==null) || (s.length==0)) ? s : s.charAt(0).toUpperCase().concat(s.slice(1));
};

//наследование
MiscFunctions.prototype.extend = function(Child, Parent){
	var F = function() { };
	F.prototype = Parent.prototype;
	Child.prototype = new F();
	Child.prototype.constructor = Child;
	Child.superclass = Parent.prototype;
};

//палитра для диалога выбора цвета
MiscFunctions.prototype.getPalette = function(){
	var line = [0, 51, 102, 153, 204, 255];
	var r = 0, g = 0, b = 0, i = 0;
	var palette = [];
	//
	while(i<4){
		b=0;
		while (b<line.length) {
			for (g=0; g<line.length; g++){
				palette.push((line[r] << 16) | (line[g] << 8) | line[b]);
			}
			r++;
			if(r>0){
				if (r%3==0) {
					r = i;
					b++;
				}
			}
		}
		i+=3;
		r=i;
	}
	//by Philip: я заменил 0x999999 на 0x808080, он соответствует (S=0, L=50)
	var grayColors = [0x000000, 0x333333, 0x666666, 0x999999-0x191919, 0xcccccc, 0xefefef, 0xffffff];
	for(i=0; i<grayColors.length; i++)
		palette.push(grayColors[i]);
	return palette;
};

//проверка, что это цвет
MiscFunctions.prototype.isRGBString = function(c){
	return (typeof c == TYPE_STRING && c.length == 7 && c.charAt(0) == "#" && !isNaN(parseInt(c.slice(1))));
};

//преобразование #RRGGBB в int
MiscFunctions.prototype.RGBString2Int = function(v){
	return parseInt((v!=null && v.length>0 && isNaN(v.charAt(0))) ? v.slice(1, v.length) : v, 16);
};

//преобразование int в #RRGGBB
MiscFunctions.prototype.int2RGBString = function(v){
	var s = v.toString(16).toUpperCase();
	while(s.length<6)
		s = "0".concat(s);
	return "#".concat(s);
};

//преобразование int и alpha в "вызов" rgba(..)
MiscFunctions.prototype.intAlpha2RGBAFunction = function(v, alpha){
	var r = (v>>16) & 0xFF, g = (v>>8) & 0xFF, b = v & 0xFF, a = (alpha<0) ? 0 : ( (alpha>1) ? 1 : alpha);
	return "rgba("+r+","+g+","+b+","+a+")";
};

//преобразование RGBString и alpha в "вызов" rgba(..)
MiscFunctions.prototype.RGBStringAlpha2RGBAFunction = function(s, alpha){
	var v = MiscFunctions.prototype.RGBString2Int(s);
	var r = (v>>16) & 0xFF, g = (v>>8) & 0xFF, b = v & 0xFF, a = (alpha<0) ? 0 : ( (alpha>1) ? 1 : alpha);
	return "rgba("+r+","+g+","+b+","+a+")";
};

//подбор css-значения для fontFace
MiscFunctions.prototype.matchWebSafeFontFamily = function(_sFF){
	var sFF = ((typeof _sFF == TYPE_UNDEFINED) ? "Arial" : _sFF).toLowerCase();
	switch(sFF){
		case "arial":
			sFF = "Arial, Helvetica, sans-serif";
			break;
		case "times new roman":
		case "times":
			sFF = "\"Times New Roman\", Times, serif";
			break;
		case "courier":
			sFF = "\"Courier New\", Courier, monospace";
			break;
		case "a_domino":
			sFF = "\"Comic Sans MS\", cursive, sans-serif";
			break;
		default:
	}
	return sFF;
};

//удаление тегов font и замена их на span с соотв. css
MiscFunctions.prototype.replaceFontTag = function(sHTML, baseFontSize){

	var s = String(sHTML), sLower, iStart, iStartClose, iEnd, sFont, i, a, sCSS, sSpan, sSpanClose = "</span>", sSize, v;
	for(i=0;;i++){
		sLower = s.toLowerCase();
		iEnd = sLower.indexOf("</font>");
		iStart = sLower.lastIndexOf("<font", iEnd);

		if(iStart==-1 || iEnd==-1)
			break;
		iStartClose = sLower.indexOf(">", iStart);
		if(iStartClose>=iEnd)
			break;
		sFont = s.slice(iStart+5, iStartClose);

		sCSS = "";
		a = sFont.match(/face=\"[A-z0-9]+(?=\")/i);
		if(a!=null && a.length>0){
			sCSS = sCSS.concat("font-family"+":"+a[0].slice(6)+";");
		}
		a = sFont.match(/size=([\"']){1}[+-]?[A-z0-9]+(?=\1)/i);

		if(a!=null && a.length>0){
			sSize = a[0].slice(6);
			if(!isNaN(sSize)){
				v = parseInt(sSize);
				if(v<0){
					v += ((typeof baseFontSize == TYPE_NUMBER) ? baseFontSize : 11);
					sSize = v+"px";
				}else{
					sSize = sSize+"px";
				}
			}
			sCSS = sCSS.concat("font-size"+":"+sSize+";");
		}
		a = sFont.match(/color=\"[#A-z0-9]+(?=\")/i);
		if(a!=null && a.length>0){
			sCSS = sCSS.concat("color"+":"+a[0].slice(7)+";");
		}
		//sSpan = (sCSS.length==0) ? "<span>" : "<span style=\"line-height:100%; display:block; "+sCSS+"\">";
		sSpan = (sCSS.length==0) ? "<span>" : "<span style=\"line-height:100%; "+sCSS+"\">";

		s = s.slice(0, iEnd)+sSpanClose+s.slice(iEnd+7);
		s = s.slice(0, iStart)+sSpan+s.slice(iStartClose+1);
	}
	return s;
};

//
MiscFunctions.prototype.stripSpaces = function(s){
	if(s == null || s.length == 0)
		return s;
	var a = s.split(""), i = 0, n = a.length;
	while(i<n && a[i] == ' ')
		i++;
	if(i>0)
		a.splice(0, i);
	i = a.length-1;
	while(i>0 && a[i] == ' ')
		i--;
	n = a.length;
	if(i<n-1)
		a.splice(i+1);
	return a.join("");
};

//замена идущих подряд пробелов соответствующим количеством Alt+0160
MiscFunctions.prototype.fReplaceSpaces = function(sFound, offset, s){
	var l = sFound.length, i, a = new Array(l);
	for(i=0; i<l; i++, a[i] = " "){}
	return a.join("");
};
MiscFunctions.prototype.replaceMultiSpacesWithNonBreakable = function(sHTML){
	return sHTML.replace(/\x20{2,}/g, this.fReplaceSpaces);
};

//пересечение прямоугольников
MiscFunctions.prototype.rectanglesIntersect = function(rA, rB){
	var dX = Math.abs(2*(rA.x-rB.x)+(rA.width-rB.width)), dY = Math.abs(2*(rA.y-rB.y)+(rA.height-rB.height));
	return (dX<=(rA.width+rB.width)) && (dY<=(rA.height+rB.height));
};

MiscFunctions.prototype.rectanglesIntersectX = function(rA, rB){
	var dX = Math.abs(2*(rA.x-rB.x)+(rA.width-rB.width));
	return dX<=(rA.width+rB.width);
};

MiscFunctions.prototype.rectanglesIntersectY = function(rA, rB){
	var dY = Math.abs(2*(rA.y-rB.y)+(rA.height-rB.height));
	return dY<=(rA.height+rB.height);
};

//пересечение отрезка и прямоугольника (objLine = {xFrom, yFrom, xTau, yTau})
MiscFunctions.prototype.lineIntersectsRectangle = function(objLine, objRect){
	var xFrom = objLine.xFrom, yFrom = objLine.yFrom, xTau = objLine.xTau, yTau = objLine.yTau;
	var xR = objRect.x, yR = objRect.y, w = objRect.width, h = objRect.h;
	var tMin = Number.POSITIVE_INFINITY, t, tRect;
	//верхняя сторона прямоугольника
	var xRFrom = xR, yRFrom = yR, xRTau = w, yRTau = 0;
	var D = xRTau*yTau-yRTau*xTau;
	if(D!=0){
		tRect = (yTau*(xFrom-xRFrom)-xTau*(yFrom-yRFrom))/D;
		if(tRect>=0 && tRect<=1){
			t = (yRTau*(xFrom-xRFrom)-xRTau*(yFrom-yRFrom))/D;
			if(t>0 && t<1 && t<tMin)
				tMin = t;
		}
		//нижняя сторона
		yRFrom = yR+h;
		tRect = (yTau*(xFrom-xRFrom)-xTau*(yFrom-yRFrom))/D;
		if(tRect>=0 && tRect<=1){
			t = (yRTau*(xFrom-xRFrom)-xRTau*(yFrom-yRFrom))/D;
			if(t>0 && t<1 && t<tMin)
				tMin = t;
		}
	}
	//левая сторона
	xRTau = 0;
	yRTau = h;
	xRFrom = xR;
	yRFrom = yR;
	D = xRTau*yTau-yRTau*xTau;
	if(D!=0){
		tRect = (yTau*(xFrom-xRFrom)-xTau*(yFrom-yRFrom))/D;
		if(tRect>=0 && tRect<=1){
			t = (yRTau*(xFrom-xRFrom)-xRTau*(yFrom-yRFrom))/D;
			if(t>0 && t<1 && t<tMin)
				tMin = t;
		}
		//правая сторона
		xRFrom = xR+w;
		tRect = (yTau*(xFrom-xRFrom)-xTau*(yFrom-yRFrom))/D;
		if(tRect>=0 && tRect<=1){
			t = (yRTau*(xFrom-xRFrom)-xRTau*(yFrom-yRFrom))/D;
			if(t>0 && t<1 && t<tMin)
				tMin = t;
		}
	}
	return tMin;
};

//вычисление массива точек-вершин, отстоящих от границы многоугольника (aPoint) на расстояние d
//точек может быть меньше для невыпуклого многоугольника
MiscFunctions.prototype.getPolygonOutline = function(aPoint, d){
	var aN = [], aStart = [];
	//tA - начало, tB - конец
	var i, n = aPoint.length, v, pStart;
	var aTA = new Array(n), aTB = new Array(n), t1, t2, n1, n2, dStart, nVProd;
	for(i=0;i<n;i++){
		v = (aPoint[(i+1)%n].subtract(aPoint[i]));
		v.normalize(1);
		aN.push(v);
		pStart = aPoint[i].add(new Point(-d*v.y, d * v.x));
		aStart.push(pStart);
	}
	//параметры t точек пересечения "отодвинутых" сторон
	for(i=0; i<n; i++){
		n1 = aN[i];
		n2 = aN[(i+1)%n];
		dStart = aStart[(i+1)%n].subtract(aStart[i]);
		nVProd = n1.x*n2.y-n1.y*n2.x;
		if(Math.abs(nVProd)>1E-10){
			t1 = (dStart.x*n2.y-dStart.y*n2.x)/nVProd;
			t2 = (dStart.x*n1.y-dStart.y*n1.x)/nVProd;
		}else{
			t1 = dStart.length();
			t2 = 0;
		}
		aTB[i] = t1;
		aTA[(i+1)%n] = t2;
	}
	//исключить "перехлёстывания", когда tA>=tB
	var aExclude = [];
	for(i=0; i<n; i++)
		if(aTA[i]>=aTB[i])
			aExclude.push(i);
	if(aExclude.length>0) {
		var aPointEx = aPoint.slice(0);
		while (aExclude.length > 0)
			aPointEx.splice(i, aExclude.pop());
		return MiscFunctions.prototype.getPolygonOutline(aPointEx, d);
	}

	var aOutline = [];
	//расчитать по aTA
	for(i=0; i<n; i++){
		n1 = aN[i];
		t1 = aTA[i];
		aOutline.push(aStart[i].add(new Point(n1.x*t1, n1.y*t1)));
	}
	return aOutline;
};

//принадлежность точки отрезку, ширина отрезка bW
MiscFunctions.prototype.isPointOnSegment = function(pA, pStart, pEnd, bW){
	var pSE = pEnd.subtract(pStart), pSA = pA.subtract(pStart), pEA = pA.subtract(pEnd);
	var bW2 = bW/2;
	var bW4 = bW2*bW2;
	var pSE2 = pSE.x*pSE.x+pSE.y*pSE.y;
	var pSA2 = pSA.x*pSA.x+pSA.y*pSA.y;
	var cProd1 = (pSE.x*pSA.x+pSE.y*pSA.y);
	var cProd2 = (pSE.x*pEA.x+pSE.y*pEA.y);
	return (pSE2*(pSA2-bW4)<=cProd1*cProd1) && (cProd2<=0 || cProd2*cProd2<=pSE2*bW4) && (cProd1>=0 || cProd1*cProd1<=pSE2*bW4);
};

//whether pA is on a segment of polyline with vertices aPoint and border width of bW
MiscFunctions.prototype.isPointOnPolygonBorder = function(pA, aPoint, bW){
	return MiscFunctions.prototype.getPolygonBorderSegmentNumberForPoint(pA, aPoint, bW) != -1;
};

//on which segment, if on any at all, pA is
MiscFunctions.prototype.getPolygonBorderSegmentNumberForPoint = function(pA, aPoint, bW){
	var i, n = aPoint.length;
	var j, n0 = n+1;
	var pStart, pEnd = aPoint[0];
	var pSE, pSA, pEA = pA.subtract(pEnd);
	var pSE2, pSA2, cProd1, cProd2;
	var bW2 = bW/2;
	var bW4 = bW2*bW2;
	//? hit border
	for(j=1; j<n0; j++){
		i = (j<n) ? j : (j % n);
		pStart = pEnd;
		pEnd = aPoint[i];
		if(pStart.equals(pEnd))
			continue;
		pSA = pEA;
		pEA = pA.subtract(pEnd);
		pSE = pEnd.subtract(pStart);
		pSE2 = pSE.x*pSE.x+pSE.y*pSE.y;
		pSA2 = pSA.x*pSA.x+pSA.y*pSA.y;
		cProd1 = (pSE.x*pSA.x+pSE.y*pSA.y);
		cProd2 = (pSE.x*pEA.x+pSE.y*pEA.y);
		if(
			(pSE2*(pSA2-bW4)<=cProd1*cProd1) &&
			(cProd2<=0 || cProd2*cProd2<=pSE2*bW4) &&
			(cProd1>=0 || cProd1*cProd1<=pSE2*bW4)
		)
			break;

	}
	return (j<n0) ? (j-1) : -1;
};


//принадлежность точки многоугольнику
MiscFunctions.prototype.isPointWithinPolygon = function(pA, aPoint){
	var aN = [];
	var nA = new Point(1, 0);
	var pStart, pEnd = aPoint[0], pSE, pSE2;
	var i, j;
	var n = aPoint.length, n0 = n+ 1;

	for(j=1; j<n0; j++) {
		i = (j < n) ? j : (j % n);
		pStart = pEnd;
		pEnd = aPoint[i];
		pSE = pEnd.subtract(pStart);
		pSE2 = pSE.x*pSE.x+pSE.y*pSE.y;
		aN.push({i:i, c2:pSE.x*pSE.x/pSE2, nX:pSE.x, nY:pSE.y, x0:pStart.x, y0:pStart.y});
	}
	aN.sort(function(a, b){
		return a.c2>b.c2;
	});
	//find a direction nonparallel to any side
	var dC2, dC2Max = 0;
	for(i=1, j=1; i<n; i++) {
		dC2 = aN[i].c2 - aN[i - 1].c2;
		if (dC2 > dC2Max) {
			j = i;
			dC2Max = dC2;
		}
	}
	nA.x = (Math.sqrt(aN[j].c2)+Math.sqrt((aN[j-1].c2)))/2;
	nA.y = Math.sqrt(1-nA.x*nA.x);
	//find all the intersections for nA with the sides (.nX .nY) of the polygon
	var tA, tP, nCross = 0, d, obj, nx, ny, px, py;
	for(i=0; i<n; i++){
		obj = aN[i];
		nx = obj.nX;
		ny = obj.nY;
		px = obj.x0-pA.x;
		py = obj.y0-pA.y;
		d = nx*nA.y-ny*nA.x;
		tP = (nA.x*py-nA.y*px)/d;
		tA = (nx*py-ny*px)/d;
		if(tA>0 && tP>=0 && tP<1)
			nCross++;
	}
	return ((nCross %2)==1);
};

MiscFunctions.prototype.splitNumericalArrayString = function(s) {
	var a = (s.length>0) ? s.split(",") : [], i, n = a.length, sN;
	var aResult = [];
	for (i = 0; i < n; i++) {
		sN = a[i];
		aResult.push((sN.indexOf(".")==-1) ? parseInt(sN) : parseFloat(sN));
	}
	return aResult;
};

MiscFunctions.prototype.getFirstElementChildNode = function(cNode){
	var node = null;
	if(cNode.nodeType == 1 && cNode.childNodes.length>0){
		node = cNode.childNodes[0];
		while(node!=null && node.nodeType != 1)
			node = node.nextSibling;
	}
	return node;
};

MiscFunctions.prototype.fillObjectWithXMLNodeAttributes = function(_obj, cNode){
	var obj = (typeof _obj == TYPE_UNDEFINED || _obj == null) ? {} : _obj, sName, sVal, sType;
	if(cNode.nodeType == 1){
		if(cNode.hasAttribute("name") && cNode.hasAttribute("value")){
			sName = cNode.getAttribute("name");
			sVal = cNode.getAttribute("value");
			sType = cNode.hasAttribute("type") ? cNode.getAttribute("type") : null;
			switch(sType){
				case "int":
				case "uint":
					obj[sName] = parseInt(sVal);
					break;
				case "Number":
					obj[sName] = parseFloat(sVal);
					break;
				case "Boolean":
					obj[sName] = (sVal==STRING_TRUE);
					break;
				case "String":
				default:
					obj[sName] = sVal;
			}
		}
	}
	return obj;
};


MiscFunctions.prototype.isXMLNode = function(obj){
    return (typeof obj == TYPE_OBJECT) && (obj != null) && (typeof obj.nodeType == TYPE_NUMBER) &&
        ((obj.nodeType == 1 && typeof obj.nodeName == TYPE_STRING) ||
        (obj.nodeType == 3 && typeof obj.nodeValue == TYPE_STRING));
};

//
MiscFunctions.prototype.unescapeEntities = function(s){
	var e = document.createElement('div');
	e.innerHTML = s;
	return e.childNodes.length === 0 ? "" : e.childNodes[0].nodeValue;
};

//
MiscFunctions.prototype.shallowCloneObject = function(objSrc){
	var obj = {};
	var s;
	for(s in objSrc)
		if(objSrc.hasOwnProperty(s))
			obj[s] = objSrc[s];
	return obj;
} ;

MiscFunctions.prototype.cloneObject = function(objSrc){
    if(Array.isArray(objSrc))
        return this.cloneArray(objSrc);
    else if(this.isXMLNode(objSrc))
        return this.cloneXMLNode(objSrc);

	var obj = {}, objProp;
	var s, flag, i, n = this.aNonCloneable.length;
    try {
        for (s in objSrc) {
            if (objSrc.hasOwnProperty(s)) {
                objProp = objSrc[s];
                flag = (typeof objProp == TYPE_OBJECT && objProp != null);
                if (flag) {
                    for (i = 0; i < n; i++)
                        if (objProp instanceof this.aNonCloneable[i])
                            break;
                    flag = (i == n);
                }
                obj[s] = flag ? this.cloneObject(objProp) : objProp;
            }
        }
    }catch(e){
        //possibly SECURITY_ERR 18
    }
    return obj;
};

MiscFunctions.prototype.cloneArray = function(aSrc){
    var a = [];
    var i, n = aSrc.length, obj;
    for(i=0; i<n; i++){
        obj = aSrc[i];
        a.push((typeof obj == TYPE_OBJECT && obj != null) ? this.cloneObject(obj) : obj);
    }
    return a;
};

MiscFunctions.prototype.cloneXMLNode = function(obj){
    return new DOMParser().parseFromString(new XMLSerializer().serializeToString(obj), "text/xml").childNodes[0];
};


MiscFunctions.prototype.fillObjectWithLabProperties = function(obj, objLabProps){
	var sProp, vProp;
	for(sProp in objLabProps){
		if(sProp.indexOf("trigger_")!=-1)
			continue;
		vProp = objLabProps[sProp];
		if(isNaN(vProp)){
			if(vProp==STRING_TRUE || vProp==STRING_FALSE)
				vProp = (vProp==STRING_TRUE);
		}else{
			vProp = (vProp.indexOf(".")==-1) ? parseInt(vProp) : parseFloat(vProp);
		}
		obj[sProp] = vProp;
	}
	return obj;
};

//===========================================================CANVAS OBJECT FROM 6BIT============================================================
MiscFunctions.prototype.canvasFrom6Bit = function(s6Bit, w, h, flagTransparent){
	var objCanvas = document.createElement('CANVAS');
	objCanvas.innerHTML = ModelElement.prototype.NO_CANVAS;
	objCanvas.width = w;
	objCanvas.height = h;

	try{
		var aARGB = RawDeflate.inflate(stringDecoder6Bit.decodeString6Bit(s6Bit));
		var context = objCanvas.getContext("2d");
		var imageData = context.createImageData(w, h);
		var aRGBA = imageData.data;
		var i, n;
		n = aRGBA.length;

		if(flagTransparent){
			for(i=0; i<n; i+=4){
				aRGBA[i+0] = aARGB[i+1];
				aRGBA[i+1] = aARGB[i+2];
				aRGBA[i+2] = aARGB[i+3];
				aRGBA[i+3] = aARGB[i];
			}
		}else{
			for(i=0; i<n; i+=4){
				aRGBA[i+0] = aARGB[i+1];
				aRGBA[i+1] = aARGB[i+2];
				aRGBA[i+2] = aARGB[i+3];
				aRGBA[i+3] = 255;
			}
		}
		aARGB = null;

		context.putImageData(imageData, 0, 0);
	}catch(e){
		//could not decode
		//alert(e.message);
	}

	return objCanvas;
};

//===========================================================CUSTOM PATTERN FROM CUSTOM PATTERN OBJECT============================================================
MiscFunctions.prototype.customPatternFromObject = function(_objCustomPattern){
	var objCustomPattern = {};
	objCustomPattern.id = _objCustomPattern.getAttribute("id");
	objCustomPattern.width = parseInt(_objCustomPattern.getAttribute("width"));
	objCustomPattern.height = parseInt(_objCustomPattern.getAttribute("height"));
	objCustomPattern.transparent = _objCustomPattern.getAttribute("transparent").toLowerCase() == STRING_TRUE;
	objCustomPattern.keep = _objCustomPattern.getAttribute("keep").toLowerCase() == STRING_TRUE;
	objCustomPattern.canvas = MiscFunctions.prototype.canvasFrom6Bit(_objCustomPattern.textContent || _objCustomPattern.innerText, objCustomPattern.width, objCustomPattern.height, objCustomPattern.transparent);
	return objCustomPattern;
};


MiscFunctions.prototype.getPageXY = function(evt) {
	var pageX, pageY;
	if(typeof evt.changedTouches != TYPE_UNDEFINED){
		var a = evt.changedTouches;
		if(a.length==0)
			return;
		pageX = a[0].pageX;
		pageY = a[0].pageY;
	}else if(typeof evt.pageX != TYPE_UNDEFINED && typeof evt.pageY != TYPE_UNDEFINED){
		pageX = evt.pageX;
		pageY = evt.pageY;
	}else
		return null;
	return new Point(pageX, pageY);
};

//=========================================================MouseDown event================================================================
MiscFunctions.prototype.disableDefaultTouchAction = function(objEl){
	objEl.style.touchAction = "none";
	//for Safari
	objEl.addEventListener("touchmove", MiscFunctions.prototype.preventDefaultForEvent);
};

MiscFunctions.prototype.enableDefaultTouchAction = function(objEl){
	objEl.style.touchAction = "auto";
	//for Safari
	objEl.removeEventListener("touchmove", MiscFunctions.prototype.preventDefaultForEvent);
};

MiscFunctions.prototype.preventDefaultForEvent = function(e){
	if(typeof e.touches != TYPE_UNDEFINED){
		if(e.touches.length>1)
			return;
	}
	e.preventDefault();
};

MiscFunctions.prototype.getClassName = function(obj){
	return (typeof obj == TYPE_UNDEFINED || typeof obj.constructor == TYPE_UNDEFINED) ? "Null" :
		((typeof obj.constructor.name != TYPE_UNDEFINED) ? obj.constructor.name : obj.constructor.toString().match(/function (.{1,})\(/)[1]);
};

//======make unselectable=====
MiscFunctions.prototype.isUnselectable = function(objDOMElement){
	return (objDOMElement.className.split(" ").indexOf(ModelElement.prototype.UNSELECTABLE_CLASS) != -1);
};

MiscFunctions.prototype.setUnselectable = function(objDOMElement, flag){
	var aClassName = objDOMElement.className.split(" "), iUnselectable = aClassName.indexOf(ModelElement.prototype.UNSELECTABLE_CLASS);
	while(aClassName.length>0 && aClassName[0] == "")
		aClassName.shift();
	if(flag){
		if(iUnselectable == -1){
			aClassName.push(ModelElement.prototype.UNSELECTABLE_CLASS);
			objDOMElement.className = aClassName.join(" ");
		}
	}else{
		if(iUnselectable != -1){
			aClassName.splice(iUnselectable, 1);
			objDOMElement.className = aClassName.join(" ");
		}
	}
};

//======random sequences======
MiscFunctions.prototype.randomizeArray = function(a, nSel, shuffle){
	var l = a.length;
	var nSelect = (typeof nSel == TYPE_UNDEFINED || nSel<0) ? l : nSel;
	var aRes = [], aNum = [], i, n = l-nSelect, j;
	for(i=0; i<l; i++)
		aNum.push(i);

	for(i=0; i<n; i++){
		j = Math.floor(Math.random()*l);
		aNum.splice(j, 1);
	}
	if(typeof shuffle == TYPE_UNDEFINED || shuffle){
		for (i=nSelect; i>1; i--) {
			j = Math.floor(i*Math.random());
			aNum.push(aNum[j]);
			aNum.splice(j, 1);
		}
	}
	for(i=0; i<nSelect; i++){
		aRes.push(a[aNum[i]]);
	}
	return aRes;
};

//
MiscFunctions.prototype.getRandomSequence = function(){
	return this.aRandomSequence[Math.floor(this.nRandomSequence*Math.random())];
};
//
MiscFunctions.prototype.initRandomSequences = function(count, n){
	if(this.aRandomSequence == null){
		this.aRandomSequence = [];
	}
	var aRandomSequence = this.aRandomSequence;
	var l = aRandomSequence.length, i;
	for(i=0;i<l;i++)
		aRandomSequence[i].splice(0);

	aRandomSequence.splice(0);
	for(i=0;i<count;i++)
		aRandomSequence[i] = this.randomSequence(n);

	this.nRandomSequence = count;
	this.randomSequenceLength = n;
};
//
MiscFunctions.prototype.getRandomSequenceLength = function(){
	return this.randomSequenceLength;
};
//
MiscFunctions.prototype.randomSequence = function(n, a){
	var aNum = (typeof a != TYPE_UNDEFINED && a != null && this.getClassName(a)=="Array") ? a :(new Array(n));
	var i, j, q;
	for(i=0; i<n; i++){
		aNum[i] = i;
	}
	if(aNum.length>n){
		aNum.splice(n);
	}
	for (i=0; i<n; i++) {
		j = Math.floor(n*Math.random());
		if(j==i)
			continue;
		q = aNum[i];
		aNum[i] = aNum[j];
		aNum[j] = q;
	}
	return aNum;
};

MiscFunctions.prototype.normalRandom = function(){
	var xx, yy, zz, s = 0;
	for(;(s==0)||(s>=1); xx=2*Math.random()-1, yy=2*Math.random()-1, s = xx*xx+yy*yy){
	}
	zz = xx*Math.sqrt(-2*Math.log(s)/s);
	return zz;
};

MiscFunctions.prototype.isInDOM = function(el){
	var elC = el, objBody = document.body;
	while(elC!=null && elC!=objBody){
		elC = elC.parentNode;
	}
	return (elC!=null);
};

MiscFunctions.prototype.getLeftAndTop = function(el){
	if(!this.isInDOM(el))
		return null;
	var p = this.getPosition(el), objStyle = el.style;
	objStyle.left = "0px";
	objStyle.top = "0px";
	var pZero = this.getPosition(el), obj = {left:(p.x-pZero.x), top:(p.y-pZero.y)};
	objStyle.left = obj.left+"px";
	objStyle.top = obj.top+"px";
	return obj;
};

MiscFunctions.prototype.getPosition = function(el) {
	if(el == null)
		return null;
	var rect = el.getBoundingClientRect();
	var docEl = document.documentElement;
	var xP = rect.left + (window.pageXOffset - docEl.clientLeft);
	var yP = rect.top + (window.pageYOffset - docEl.clientTop);
	return {x:xP, y:yP};
};

MiscFunctions.prototype.getElementBounds = function(_el, _onlyRootElement){
	var onlyRootElement  = (typeof _onlyRootElement == TYPE_UNDEFINED) ? false : _onlyRootElement;
	var el = (typeof _el.getBoundingClientRect == TYPE_FUNCTION) ? _el : _el.getEntity();
	var a = el.childNodes, i, n = a.length, aCh, j, l, elCurr;
	var docEl = document.documentElement, dX = window.pageXOffset - docEl.clientLeft, dY = window.pageYOffset - docEl.clientTop;
	var aElRet = [], aIRet = [], aNRet = [], aElAll = [el];
	var b = {left:Number.POSITIVE_INFINITY, top:Number.POSITIVE_INFINITY, right:Number.NEGATIVE_INFINITY, bottom:Number.NEGATIVE_INFINITY}, bCurr;
	//
	if(!onlyRootElement){
		for(i=0; i<n; i++){
			elCurr = a[i];
			aElAll.push(elCurr);
			aCh = elCurr.childNodes;
			l = aCh.length;
			if(l>0){
				aElRet.push(a);
				aIRet.push(i);
				aNRet.push(n);
				a = aCh;
				i = -1;
				n = a.length;
			}
			while((i == n-1) && (aElRet.length>0)){
				i = aIRet.pop();
				n = aNRet.pop();
				a = aElRet.pop();
			}
		}
	}
	//
	n = aElAll.length;
	//
	for(i=0; i<n; i++){
		elCurr = aElAll[i];
		if(typeof elCurr.getBoundingClientRect == TYPE_UNDEFINED)
			continue;
		bCurr = elCurr.getBoundingClientRect();
		if((bCurr.left == bCurr.right) || (bCurr.top == bCurr.bottom))
			continue;
		if(bCurr.left<b.left)
			b.left = bCurr.left;
		if(bCurr.right>b.right)
			b.right = bCurr.right;
		if(bCurr.top<b.top)
			b.top = bCurr.top;
		if(bCurr.bottom>b.bottom)
			b.bottom = bCurr.bottom;
	}
	//
	b.width = b.right-b.left;
	b.height = b.bottom-b.top;
	//
	return b;
};

MiscFunctions.prototype.getOffsetWidthHeight = function(domObj, _obj){
	var obj = (typeof _obj != TYPE_OBJECT) ? {offsetWidth:0, offsetHeight:0} : _obj;
	if(typeof domObj != TYPE_UNDEFINED && typeof domObj.parentNode != TYPE_UNDEFINED){
		if(domObj.offsetWidth == 0 && domObj.offsetHeight == 0){
			var p = domObj.parentNode;
			var d = (typeof domObj.style.display != TYPE_STRING) ? undefined : domObj.style.display;
			if(p!=null)
				p.removeChild(domObj);
			domObj.style.display = "block";
			domObj.style.width = "";
			domObj.style.height = "";
			document.body.appendChild(domObj);
			obj.offsetWidth = domObj.offsetWidth;
			obj.offsetHeight = domObj.offsetHeight;
			document.body.removeChild(domObj);
			if(p!=null)
				p.appendChild(domObj);
			domObj.style.display = d;
		}else{
			obj.offsetWidth = domObj.offsetWidth;
			obj.offsetHeight = domObj.offsetHeight;
		}
	}
	return obj;
};

MiscFunctions.prototype.getScrollLeft = function(){
	var doc = document.documentElement;
	return (window.pageXOffset || doc.scrollLeft) - (doc.clientLeft || 0);
};

MiscFunctions.prototype.getScrollTop = function(){
	var doc = document.documentElement;
	return (window.pageYOffset || doc.scrollTop)  - (doc.clientTop || 0);
};

MiscFunctions.prototype.getSWFSizeCorrection = function(el){
	var a = el.childNodes, i, n = a.length, aCh, j, l, elCurr, ii;
	var aElRet = [], aIRet = [], aNRet = [], aStyleAll = [], objProp;
	var sSize, sFace, objStyle, foundSize = false, foundFace = false, maxSize = -1;
	for(i=0; i<n; i++){
		elCurr = a[i];
		if(typeof elCurr.style != TYPE_UNDEFINED){
			objStyle = elCurr.style;
			foundSize = (typeof objStyle.fontSize != TYPE_UNDEFINED);
			foundFace = (typeof objStyle.fontFace != TYPE_UNDEFINED);
			if(foundSize || foundFace){
				objProp = {};
				if(foundSize){
					sSize = objStyle.fontSize.toLowerCase();
					ii = sSize.indexOf("px");
					if(ii>0)
						objProp.size = parseInt(sSize.slice(0, ii));
					else
						foundSize = false;
				}
				if(foundFace)
					objProp.face = objStyle.fontFace;
				if(foundFace || foundSize)
					aStyleAll.push(objProp);
			}
		}
		aCh = elCurr.childNodes;
		l = aCh.length;
		if(l>0){
			aElRet.push(a);
			aIRet.push(i);
			aNRet.push(n);
			a = aCh;
			i = -1;
			n = a.length;
		}
		while((i == n-1) && (aElRet.length>0)){
			i = aIRet.pop();
			n = aNRet.pop();
			a = aElRet.pop();
		}
	}
	//
	n = aStyleAll.length;
	for(i=0; i<n; i++){
		objProp = aStyleAll[i];
		if(typeof objProp.size != TYPE_UNDEFINED && maxSize < objProp.size)
			maxSize = objProp.size;
	}
	return {width:0, height:0.117*((maxSize<0) ? 0 : maxSize)+3.96};
};

MiscFunctions.prototype.removeChildSafe = function(el){
	while(el.childNodes.length > 0) {
		MiscFunctions.prototype.removeChildSafe(el.childNodes[el.childNodes.length-1]);
	}
	if(el.parentNode != null)
		el.parentNode.removeChild(el);
};

MiscFunctions.prototype.getArrayBounds = function(a){
	var i, n = a.length, b = {left:Number.POSITIVE_INFINITY, top:Number.POSITIVE_INFINITY, right:Number.NEGATIVE_INFINITY, bottom:Number.NEGATIVE_INFINITY}, bCurr;
	for(i=0; i<n; i++){
		bCurr = this.getElementBounds(a[i].getEntity());
		if(bCurr.left<b.left)
			b.left = bCurr.left;
		if(bCurr.right>b.right)
			b.right = bCurr.right;
		if(bCurr.top<b.top)
			b.top = bCurr.top;
		if(bCurr.bottom>b.bottom)
			b.bottom = bCurr.bottom;
	}
	b.width = b.right-b.left;
	b.height = b.bottom-b.top;
	return b;
};

//
MiscFunctions.prototype.removeArrayItemsWithPropertyValue = function(a, sProp, v){
	var c = 0, i, l = a.length, curr;
	for(i=0; i<l; i++){
		curr = a[i];
		if((typeof curr != TYPE_UNDEFINED) && (curr!=null) && typeof curr[sProp] != TYPE_UNDEFINED && (curr[sProp]==v)){
			a.splice(i--,1);
			l--;
			c++;
		}
	}
	return c;
};

//
MiscFunctions.prototype.RGB2HSV = function(vRGB){
	//r,g,b=[0..255]
	if(isNaN(vRGB) || (vRGB<0x000000) || (vRGB>0xFFFFFF)){
		return {h:0, s:0, v:0};
	}
	var r = ((vRGB & 0xFF0000)>>16)/255, g = ((vRGB & 0x00FF00)>>8)/255, b = (vRGB & 0xFF)/255, cMin, cMax, h, s;
	cMin = (b<g) ? ((b<r)?(b):(r)) : ((g<r)?(g):(r));
	cMax = (b>g) ? ((b>r)?(b):(r)) : ((g>r)?(g):(r));
	if(cMax == cMin){
		h = 0;
	}else if(cMax == r){
		h = 60*(g-b)/(cMax-cMin);
		if(g<b){
			h += 360;
		}
	}else if(cMax == g){
		h = 60*(b-r)/(cMax-cMin) + 120;
	}else{
		h = 60*(r-g)/(cMax-cMin) + 240;
	}
	s = (cMax == 0) ? (0) : (1-cMin/cMax);
	return {h:h, s:100*s, v:100*cMax};
};

//
MiscFunctions.prototype.HSV2RGB = function(objHSV){
	if(typeof objHSV == TYPE_UNDEFINED || objHSV == null || isNaN(objHSV.h))
		return 0x000000;
	//v=[0..100], s=[0..100], h=[0..360]
	var h1, f, p, q, t, vRGB, r, g, b;
	var xS = objHSV.s*0.01, xV = objHSV.v*0.01;
	f = objHSV.h/60;
	h1 = Math.floor(f);
	f -= h1;
	h1 %= 6;
	p = xV*(1-xS);
	q = xV*(1-f*xS);
	t = xV*(1-(1-f)*xS);
	switch(h1){
		case 0:
			r = xV;
			g = t;
			b = p;
			break;
		case 1:
			r = q;
			g = xV;
			b = p;
			break;
		case 2:
			r = p;
			g = xV;
			b = t;
			break;
		case 3:
			r = p;
			g = q;
			b = xV;
			break;
		case 4:
			r = t;
			g = p;
			b = xV;
			break;
		case 5:
			r = xV;
			g = p;
			b = q;
			break;
	}
	r = Math.floor(255*r);
	g = Math.floor(255*g);
	b = Math.floor(255*b);
	vRGB = (r<<16)+(g<<8)+b;
	return vRGB;
};

//
MiscFunctions.prototype.HSV2HSL = function(objHSV){
	//v=[0..100], l = [0..100], s=[0..100], h=[0..360]
	var h = objHSV.h, s = objHSV.s*0.01, v = objHSV.v*0.01;
	var hh = (2-s)*v, l = hh/2, sat = (v<1E-5) ? (s/(2-s)) : (s*v/((hh<1)?(hh):(2-hh)));
	sat = (hh<1) ? ((v<1E-5) ? (0) : (s*v/hh)) : ((2-hh<1E-5) ? (0) : (s*v/(2-hh)));
	return {h:h, s:Math.round(100*sat*(1-Math.abs(2*l-1))), l:Math.round(100*l)};
};

//
MiscFunctions.prototype.RGB2HSL = function(vRGB, _obj){
	var obj = (typeof _obj == TYPE_OBJECT && _obj != null) ? _obj : {};
	//l = [0..100], s=[0..100], h=[0..360], r,g,b=[0..255]
	if(isNaN(vRGB) || (vRGB<0x000000) || (vRGB>0xFFFFFF)){
		obj.h = obj.s = obj.l = 0;
		return obj;
	}
	var v = vRGB % 0xFFFFFF;
	return this.integerRGB2HSL(((vRGB & 0xFF0000)>>16), ((vRGB & 0x00FF00)>>8), (vRGB & 0xFF), obj);
};

//
MiscFunctions.prototype.integerRGB2HSL = function(r8, g8, b8, _obj){
	var obj = (typeof _obj == TYPE_OBJECT && _obj != null) ? _obj : {};
	//l = [0..100], s=[0..100], h=[0..360], r8,g8,b8=[0..255]
	var r = r8/255, g = g8/255, b = b8/255;
	var rgbMAX = Math.max(r, g, b), rgbMIN = Math.min(r, g, b);
	var cL = 0.5*(rgbMIN+rgbMAX), dMM = rgbMAX-rgbMIN, cH = 0, cS = 0;
	if((dMM>1E-5) && (cL>1E-5)){
		switch(rgbMAX){
			case r:
				if(g>=b){
					cH = 60*(g-b)/dMM;
				}else{
					cH = 60*(g-b)/dMM + 360;
				}
				break;
			case g:
				cH = 60*(b-r)/dMM + 120;
				break;
			case b:
				cH = 60*(r-g)/dMM + 240;
				break;
		}
		if(cL<0.5){
			cS = 0.5*dMM/cL;
		}else if(cL<1){
			cS = dMM/(2-2*cL);
		}else{
			cS = 1;
		}
	}
	obj.h = Math.floor(cH);
	obj.s = Math.round(100*cS);
	obj.l = Math.round(100*cL);
	return obj;
};

//
MiscFunctions.prototype.HSL2RGB = function(objHSL){
	return this.valuesHSL2RGB(objHSL.h, objHSL.s, objHSL.l);
};

//
MiscFunctions.prototype.valuesHSL2RGB = function(h, s, l){
	var cH = h, cS = s*0.01, cL = l*0.01;
	var q = (cL<0.5)?(cL*(1.0+cS)):(cL+cS-cL*cS), p = 2*cL-q, hk = (cH%360)/360;
	var tr = hk+this.ONE_OVER_THREE, tg = hk, tb = hk-this.ONE_OVER_THREE;
	var r, g, b;
	if(tr<0){
		tr += 1;
	}else if(tr>1){
		tr -= 1;
	}
	if(tg<0){
		tg += 1;
	}else if(tr>1){
		tg -= 1;
	}
	if(tb<0){
		tb += 1;
	}else if(tr>1){
		tb -= 1;
	}
	r = Math.round(255*((tr<this.ONE_OVER_SIX)?(p+6*(q-p)*tr):((tr<0.5)?(q):((tr<this.TWO_OVER_THREE)?(p+6*(q-p)*(this.TWO_OVER_THREE-tr)):(p)))));
	g = Math.round(255*((tg<this.ONE_OVER_SIX)?(p+6*(q-p)*tg):((tg<0.5)?(q):((tg<this.TWO_OVER_THREE)?(p+6*(q-p)*(this.TWO_OVER_THREE-tg)):(p)))));
	b = Math.round(255*((tb<this.ONE_OVER_SIX)?(p+6*(q-p)*tb):((tb<0.5)?(q):((tb<this.TWO_OVER_THREE)?(p+6*(q-p)*(this.TWO_OVER_THREE-tb)):(p)))));
	return ((r<<16)+(g<<8)+b);
};

//
MiscFunctions.prototype.valuesHSL2RGBString = function(h, s, l){
	return this.int2RGBString(this.valuesHSL2RGB(h, s, l));
};

//
MiscFunctions.prototype.valuesHSL2Lab = function(h, s, l, _obj){
	var cH = h, cS = s*0.01, cL = l*0.01;
	var q = (cL<0.5)?(cL*(1.0+cS)):(cL+cS-cL*cS), p = 2*cL-q, hk = (cH%360)/360;
	var tr = hk+this.ONE_OVER_THREE, tg = hk, tb = hk-this.ONE_OVER_THREE;
	var r, g, b;
	if(tr<0){
		tr += 1;
	}else if(tr>1){
		tr -= 1;
	}
	if(tg<0){
		tg += 1;
	}else if(tr>1){
		tg -= 1;
	}
	if(tb<0){
		tb += 1;
	}else if(tr>1){
		tb -= 1;
	}
	r = Math.round(255*((tr<this.ONE_OVER_SIX)?(p+6*(q-p)*tr):((tr<0.5)?(q):((tr<this.TWO_OVER_THREE)?(p+6*(q-p)*(this.TWO_OVER_THREE-tr)):(p)))));
	g = Math.round(255*((tg<this.ONE_OVER_SIX)?(p+6*(q-p)*tg):((tg<0.5)?(q):((tg<this.TWO_OVER_THREE)?(p+6*(q-p)*(this.TWO_OVER_THREE-tg)):(p)))));
	b = Math.round(255*((tb<this.ONE_OVER_SIX)?(p+6*(q-p)*tb):((tb<0.5)?(q):((tb<this.TWO_OVER_THREE)?(p+6*(q-p)*(this.TWO_OVER_THREE-tb)):(p)))));
	//
	var n_R = r/255, n_G = g/255, n_B = b/255, cX, cY, cZ;
	n_R = 100*((n_R>0.04045) ? Math.exp(2.4*Math.log((n_R+0.055)/1.055)) : (n_R/12.92));
	n_G = 100*((n_G>0.04045) ? Math.exp(2.4*Math.log((n_G+0.055)/1.055)) : (n_G/12.92));
	n_B = 100*((n_B>0.04045) ? Math.exp(2.4*Math.log((n_B+0.055)/1.055)) : (n_B/12.92));
	//Observer. = 2°, Illuminant = D65
	cX = n_R * 0.4124 + n_G * 0.3576 + n_B * 0.1805;
	cY = n_R * 0.2126 + n_G * 0.7152 + n_B * 0.0722;
	cZ = n_R * 0.0193 + n_G * 0.1192 + n_B * 0.9505;
	//
	var n_X = cX/95.047, n_Y = cY/100.000, n_Z = cZ/108.883;
	n_X = (n_X>0.008856) ? Math.exp(Math.log(n_X)/3) : (7.787*n_X+0.1379);
	n_Y = (n_Y>0.008856) ? Math.exp(Math.log(n_Y)/3) : (7.787*n_Y+0.1379);
	n_Z = (n_Z>0.008856) ? Math.exp(Math.log(n_Z)/3) : (7.787*n_Z+0.1379);
	//
	var nL = 116*n_Y - 16, na = 500*(n_X-n_Y), nb = 200*(n_Y-n_Z);
	if(nL<0)
		nL = 0;
	//
	var obj = (typeof _obj == TYPE_OBJECT) ? _obj : {};
	obj.L = nL;
	obj.a = na;
	obj.b = nb;
	return obj;
};

//
MiscFunctions.prototype.deltaE_CIE94 = function(L1, a1, b1, L2, a2, b2){
	var deltaL = L1-L2, a1sqr = a1*a1, a2sqr = a2*a2, b1sqr = b1*b1, b2sqr = b2*b2;
	var C1 = Math.sqrt(a1sqr+b1sqr), C2 = Math.sqrt(a2sqr+b2sqr), deltaC = C1-C2, delta_a = a1-a2, delta_b = b1-b2;
	var Hsqr = delta_a*delta_a+delta_b*delta_b-deltaC*deltaC;
	var deltaH = (Hsqr>0) ? Math.sqrt(delta_a*delta_a+delta_b*delta_b-deltaC*deltaC) : 0;
	var SL = 1, SC = 1+0.045*C1, SH = 1+0.015*C1;
	var ndL = deltaL, ndC = deltaC/SC, ndH = deltaH/SH, ndLsqr = ndL*ndL, ndCsqr = ndC*ndC, ndHsqr = ndH*ndH;
	var dE94 = Math.sqrt(ndLsqr+ndCsqr+ndHsqr);
	return (dE94>100) ? 100 : dE94;
};

//
MiscFunctions.prototype.colorSimilarityHSL_CIE94 = function(c1, c2){
	var objLab1 = this.valuesHSL2Lab(c1.h, c1.s, c1.l),  objLab2 = this.valuesHSL2Lab(c2.h, c2.s, c2.l);
	return 1.0-0.01*this.deltaE_CIE94(objLab1.L, objLab1.a, objLab1.b, objLab2.L, objLab2.a, objLab2.b);
};

//
MiscFunctions.prototype.colorSimilarityValuesHSL_CIE94 = function(h1, s1, l1, h2, s2, l2){
	var objLab1 = this.valuesHSL2Lab(h1, s1, l1),  objLab2 = this.valuesHSL2Lab(h2, s2, l2);
	return 1.0-0.01*this.deltaE_CIE94(objLab1.L, objLab1.a, objLab1.b, objLab2.L, objLab2.a, objLab2.b);
};

//
MiscFunctions.prototype.getColorString = function(v){
	var s = "", attr;
	for(attr in v){
		if(s.length>0){
			s += ","+attr+":"+v[attr];
		}else{
			s += attr+":"+v[attr];
		}
	}
	return s.slice(0, s.length);
};

//
MiscFunctions.prototype.getColorObject = function(v){
	var obj = {};
	var a = (v.length>0) ? v.split(",") : [], l = a.length, i, j, s;
	for(i=0; i<l; i++){
		s = a[i];
		j = s.indexOf(":");
		if(j!=-1)
			obj[s.slice(0, j)] = Number(s.slice(j+1));
	}
	return obj;
};

//
MiscFunctions.prototype.getColorAlphaString = function(c, a){
	return this.HSV2RGB(c) + "," + (isNaN(a) ? 1.0 : a);
};

//
MiscFunctions.prototype.getCA = function(v){
	var obj = {c:0x000000, a:1};
	if(v != null){
		var a = (v.length>0) ? v.split(",") : [];
		if(a.length > 0){
			var bitmapFlag = (a[0].indexOf(this.BITMAP_FILL_PREFIX) == 0);
			if(bitmapFlag){
				obj.pattern = a[0].slice(8);
			}else{
				if(isNaN(a[0])){
					if(CustomPatterns.getCustomPatternById(a[0]) != null){
						obj.pattern = a[0];
					}
				}else{
					obj.c = parseInt(a[0]);
				}
			}
			if(a.length>1){
				obj.a = Number(a[1]);
			}
		}
	}
	return obj;
};

//
MiscFunctions.prototype.colorAlphaString2HSVA = function(v){
	var objCA = this.getCA(v);
	var obj = this.RGB2HSV(objCA.c);
	obj.a = isNaN(objCA.a) ? 1.0 : objCA.a;
	return obj;
};

//
MiscFunctions.prototype.colorAlphaString2HSLA = function(v){
	var objCA = this.getCA(v);
	var obj = this.RGB2HSL(objCA.c);
	obj.a = isNaN(objCA.a) ? 1.0 : objCA.a;
	return obj;
};

//
MiscFunctions.prototype.colorSimilarityRGB = function(c1, c2){
	//от 0 до 1 (диагональ цветового куба)
	//1.7320508075688772935274463415059 * 256 = 443.4050067376325871430262634255    1/443.405=0.00225527  0.00225527^0.5 = 0.0474897
	var c1r = (c1 & 0xFF0000)>>16, c1g = (c1 & 0x00FF00)>>8, c1b = c1 & 0x0000FF;
	var c2r = (c2 & 0xFF0000)>>16, c2g = (c2 & 0x00FF00)>>8, c2b = c2 & 0x0000FF;
	var dr, dg, db, d;
	//
	dr = c2r-c1r;
	dg = c2g-c1g;
	db = c2b-c1b;
	//
	//корень четвёртой степени, чтобы непохожесть начиналась уже на серединах противоположных граней
	d = Math.sqrt(Math.sqrt(dr*dr+dg*dg+db*db))*0.0474897;
	d = 1 - d;
	return (d<0) ? 0 : d;
};

//
MiscFunctions.prototype.colorSimilarityHSV = function(c1, c2){
	//через HSL
	var objHSL1 = this.HSV2HSL(c1), objHSL2 = this.HSV2HSL(c2);
	var l1 = objHSL1.l*0.01, l2 = objHSL2.l*0.01, h1 = objHSL1.h*PI_OVER_180, h2 = objHSL2.h*this.PI_OVER_180;
	var s1 = objHSL1.s/200, s2 = objHSL2.s/200, dl = l1-l2;
	var r = 1-Math.sqrt(dl*dl+s1*s1+s2*s2-2*s1*s2*Math.cos(h1-h2));
	//
	return (r<0) ? 0 : r;
};

//
MiscFunctions.prototype.colorSimilarityHSL = function(c1, c2, _cosConeTheta, _sinConeTheta, _cosConePhi, _sinConePhi){
	return this.colorSimilarityValuesHSL(c1.h, c1.s, c1.l, c2.h, c2.s, c2.l, _cosConeTheta, _sinConeTheta, _cosConePhi, _sinConePhi);
};

//
MiscFunctions.prototype.colorSimilarityValuesHSL = function(c1h, c1s, c1l, c2h, c2s, c2l, _cosConeTheta, _sinConeTheta, _cosConePhi, _sinConePhi){
	//используется HSL-пространство в виде конуса, скошенного и повёрнутого так, чтобы были близки (жёлтый и белый) и (синий и чёрный)
	//для эволюции coneTheta = 0.256 рад, conePhi = PI/3 рад
	var cosConeTheta = (typeof _cosConeTheta == TYPE_UNDEFINED || _cosConeTheta == null) ? 0 : _cosConeTheta;
	var sinConeTheta = (typeof _sinConeTheta == TYPE_UNDEFINED || _sinConeTheta == null) ? 1 : _sinConeTheta;
	var cosConePhi = (typeof _cosConePhi == TYPE_UNDEFINED || _cosConePhi == null) ? 1 : _cosConePhi;
	var sinConePhi = (typeof _sinConePhi == TYPE_UNDEFINED || _sinConePhi == null) ? 0 : _sinConePhi;
	var h1 = c1h*this.PI_OVER_180, h2 = c2h*this.PI_OVER_180;
	var l1 = c1l*0.01, l2 = c2l*0.01;
	var s1 = 0.01*c1s * ((l1<0.5)?(l1):(1-l1));
	var s2 = 0.01*c2s * ((l2<0.5)?(l2):(1-l2));
	l1 -= 0.5;
	l2 -= 0.5;
	var x1 = l1*cosConeTheta*cosConePhi + s1*Math.cos(h1);
	var y1 = l1*cosConeTheta*sinConePhi + s1*Math.sin(h1);
	var z1 = l1*sinConeTheta;
	var x2 = l2*cosConeTheta*cosConePhi + s2*Math.cos(h2);
	var y2 = l2*cosConeTheta*sinConePhi + s2*Math.sin(h2);
	var z2 = l2*sinConeTheta;
	var dx = x1-x2, dy = y1-y2, dz = z1-z2;
	var r = 1 - Math.sqrt(dx*dx+dy*dy+dz*dz);
	//
	return (r<0) ? 0 : r;
};

//
MiscFunctions.prototype.colorSimilarityHSL_plain = function(c1, c2){
	//l = [0..100], s=[0..100], h=[0..360]
	var h1 = c1.h*this.PI_OVER_180, h2 = c2.h*this.PI_OVER_180;
	var l1 = c1.l*0.01, l2 = c2.l*0.01;
	var s1 = 0.01*c1.s * ((l1<0.5)?(l1):(1-l1));
	var s2 = 0.01*c2.s * ((l2<0.5)?(l2):(1-l2));
	var dsh2 = s1*s1+s2*s2-2*s1*s2*Math.cos(h1-h2);
	var dl = l1-l2;
	var r = 1-Math.sqrt(dl*dl+dsh2);
	//
	return (r<0)?(0):(r);
};

//
MiscFunctions.prototype.RGB2Gray = function(vRGB){
	var v = vRGB % 0xFFFFFF;
	var r = ((vRGB & 0xFF0000)>>16), g = ((vRGB & 0x00FF00)>>8), b = (vRGB & 0xFF);
	var gr = Math.round(0.299*r+0.587*g+0.114*b);
	if(gr>255)
		gr = 255;
	return gr;
};

//
MiscFunctions.prototype.RGB2Grayscale = function(vRGB){
	return 65793*this.RGB2Gray(vRGB);
};

//
MiscFunctions.prototype.chi2Critical = function(df, _p){
	var p = (typeof _p == TYPE_UNDEFINED || _p == null) ? 0.05 : _p;
	if((p != 0.050) || (df<1) || (df>30)){
		return NaN;
	}
	return this.chi2CriticalValues[df];
};

//
MiscFunctions.prototype.escapeRegexChars = function(s){
	return s.replace(new RegExp("([{}\(\)\^$&.\*\?\/\+\|\[\\\\]|\]|\-)","g"), "\\$1");
};

//
MiscFunctions.prototype.drawRoundRect = function(ctx, xL, yT, w, h, wE, hE){
	var rW = wE/2, rH = hE/2;
	ctx.beginPath();
	ctx.moveTo(xL+rW, yT);
	ctx.lineTo(xL+w-rW, yT);
	ctx.quadraticCurveTo(xL+w, yT, xL+w, yT+rH);
	ctx.lineTo(xL+w, yT+h-rH);
	ctx.quadraticCurveTo(xL+w, yT+h, xL+w-rW, yT+h);
	ctx.lineTo(xL+rW, yT+h);
	ctx.quadraticCurveTo(xL, yT+h, xL, yT+h-rH);
	ctx.lineTo(xL, yT+rH);
	ctx.quadraticCurveTo(xL, yT, xL+rW, yT);
	ctx.closePath();
};

//
MiscFunctions.prototype.drawEllipse = function(ctx, xC, yC, rX, rY, flagFill){
	ctx.save();
	ctx.beginPath();
	ctx.translate(xC-rX, yC-rY);
	ctx.scale(rX, rY);
	ctx.arc(1, 1, 1, 0, 2 * Math.PI, false);
	ctx.restore();
	if(flagFill)
		ctx.fill();
	ctx.stroke();
};

//
MiscFunctions.prototype.isPointWithinEllipse = function(x, y, xC, yC, a, b){
	var dx = (x - xC)/a;
	var dy = (y - yC)/b;
	return dx*dx+dy*dy<=1;
};

const miscFunctions = new MiscFunctions();

const extend = miscFunctions.extend;

//************************* класс Point, наподобие класса в AS3****************
/**
 * @constructor
 */
function Point(x, y){
	this.x = x || 0;
	this.y = y || 0;
}
extend(Point, Object);

Point.prototype.x = null;
Point.prototype.y = null;
Point.prototype.C180_OVER_PI = 180 / Math.PI;

Point.prototype.getX = function(){
    return this.x;
};

Point.prototype.getY = function(){
    return this.y;
};

Point.prototype.add = function(v){
	return new Point(this.x + v.x, this.y + v.y);
};

Point.prototype.clone = function(){
    return new Point(this.x, this.y);
};

Point.prototype.degreesTo = function(v){
    var dx = this.x - v.x;
    var dy = this.y - v.y;
    var angle = Math.atan2(dy, dx); // radians
    return angle * Point.prototype.C180_OVER_PI; // degrees
};

Point.prototype.distance = function(v){
    var x = this.x - v.x;
    var y = this.y - v.y;
    return Math.sqrt(x*x + y*y);
};

Point.prototype.distanceSquared = function(v){
    var x = this.x - v.x;
    var y = this.y - v.y;
    return x*x + y*y;
};

Point.prototype.equals = function(toCompare){
    return (this.x == toCompare.x) && (this.y == toCompare.y);
};

Point.prototype.interpolate = function(v, f){
    return new Point((this.x + v.x)*f, (this.y + v.y)*f);
};

Point.prototype.length = function(){
    return Math.sqrt(this.x*this.x + this.y*this.y);
};

Point.prototype.lengthSquared = function(){
    return this.x*this.x + this.y*this.y;
};

Point.prototype.normalize = function(thickness){
	try{
		var l = thickness/this.length();
		this.x = this.x * l;
		this.y = this.y * l;
	}catch(e){
		//деление на ноль
	}
};

Point.prototype.multiply = function(v){
    this.x = this.x*v;
    this.y = this.y*v;
};

Point.prototype.scalarProduct = function(v){
    return this.x*v.x+this.y*v.y;
};

Point.prototype.orbit = function(origin, arcWidth, arcHeight, degrees){
    var radians = degrees * Point.prototype.C180_OVER_PI;
    this.x = origin.x + arcWidth * Math.cos(radians);
    this.y = origin.y + arcHeight * Math.sin(radians);
};

Point.prototype.offset = function(dx, dy){
    this.x += dx;
    this.y += dy;
};

Point.prototype.subtract = function(v){
    return new Point(this.x - v.x, this.y - v.y);
};

Point.prototype.toString = function(){
    return "(x=" + this.x + ", y=" + this.y + ")";
};
 
Point.interpolate = function(pt1, pt2, f){
    return new Point((pt1.x + pt2.x)*f, (pt1.y + pt2.y)*f);
};

Point.polar = function(len, angle){
    return new Point(len * Math.cos(angle), len * Math.sin(angle));
};

Point.distance = function(pt1, pt2){
    var x = pt1.x - pt2.x;
    var y = pt1.y - pt2.y;
    return Math.sqrt(x*x + y*y);
};

Point.distanceSquared = function(pt1, pt2){
    var x = pt1.x - pt2.x;
    var y = pt1.y - pt2.y;
    return x*x + y*y;
};

//MAPKIT VERSION OF ModelEvent class
//
//if MOUSE_OR_TOUCH > 0 it is a touch-screen, if  ==0 – a non-touch screen, == -1 – unknown
var MOUSE_OR_TOUCH = -1;
//
/**
 * @constructor
 */
function ModelEvent(sName, _flagBubbles, _flagCancelable){
	var e = null, flagBubbles = (typeof _flagBubbles == TYPE_BOOLEAN) ? _flagBubbles : true, flagCancelable = (typeof _flagCancelable == TYPE_BOOLEAN) ? _flagCancelable : true;
	if(typeof Event == TYPE_FUNCTION)
		e = new Event(sName, {cancelable:flagCancelable, bubbles:flagBubbles});
	else if(typeof document.createEvent == TYPE_FUNCTION){
		e = document.createEvent("Event");
		e.initEvent(sName, flagBubbles, flagCancelable);
	}
	return e;
}
extend(ModelEvent, Object);

//
ModelEvent.prototype.isProperEvent = function(e){
	var i, a = ModelEvent.prototype.TOUCH_EVENTS_ARRAY, n = a.length;
	for(i=0; i<n; i++)
		if(a[i].indexOf(e.type) != -1)
			break;
	if(i == n)
		return true;
	if(MOUSE_OR_TOUCH == -1 || (MOUSE_OR_TOUCH == 0 && i>0)){
		MOUSE_OR_TOUCH = i;	
		if(MOUSE_OR_TOUCH > 0){
			window.addEventListener("mousedown", ModelEvent.prototype.preventMouseEvent, true);
		}
	}
	
	if(MOUSE_OR_TOUCH != i && (e.type != "mousedown" || typeof e.target == TYPE_UNDEFINED || typeof e.target.tagName == TYPE_UNDEFINED || e.target.tagName != "INPUT")){
		e.stopPropagation();
		e.preventDefault();
		return false;
	}
	return true;
};
//
ModelEvent.prototype.preventMouseEvent = function(e){
	if(typeof e.target != TYPE_UNDEFINED && typeof e.target.tagName != TYPE_UNDEFINED && e.target.tagName == "INPUT")
		return;
	document.activeElement.blur();
	e.stopPropagation();
	e.preventDefault();
};
//

ModelEvent.prototype.aHoldDown = [];
ModelEvent.prototype.holdDownX = undefined;
ModelEvent.prototype.holdDownY = undefined;
ModelEvent.prototype.startHoldDown = function(fCall, argThis, aArg, delay){
	var prototypeE = ModelEvent.prototype, aHD = prototypeE.aHoldDown;	
	var obj = {func:fCall, argThis:argThis, args:aArg};
	var holdDownDelay = (typeof delay == TYPE_NUMBER && delay>=0) ? delay : prototypeE.HOLD_DOWN_TIME;
	var id = obj.id = window.setTimeout(prototypeE.fMouseTimeoutHoldDown, holdDownDelay, obj);
	aHD.push(id);
	if(aHD.length==1){
		prototypeE.radius2 = prototypeE.HOLD_DOWN_JITTER_RADIUS*prototypeE.HOLD_DOWN_JITTER_RADIUS;
		window.addEventListener((MOUSE_OR_TOUCH == 0) ? "mousemove" : "touchmove", prototypeE.fMouseMoveHoldDown, true);
		window.addEventListener((MOUSE_OR_TOUCH == 0) ? "mouseup" : "touchend", prototypeE.fMouseUpHoldDown, true);
	}
};

ModelEvent.prototype.fMouseTimeoutHoldDown = function(obj){
	var prototypeE = ModelEvent.prototype, aHD = prototypeE.aHoldDown, i;
	for(i=0; i<aHD.length; i++)
		if(aHD[i] == obj.id)
			break;
	if(i<aHD.length){
		aHD.splice(i, 1);
		window.clearTimeout(obj.id);
		if(aHD.length == 0){
			prototypeE.holdDownX = undefined;
			prototypeE.holdDownY = undefined;
			window.removeEventListener((MOUSE_OR_TOUCH == 0) ? "mousemove" : "touchmove", prototypeE.fMouseMoveHoldDown, true);
			window.removeEventListener((MOUSE_OR_TOUCH == 0) ? "mouseup" : "touchend", prototypeE.fMouseUpHoldDown, true);
		}
		obj.func.apply((typeof obj.argThis != TYPE_UNDEFINED) ? obj.argThis : window, obj.args);
	}
};

ModelEvent.prototype.fMouseMoveHoldDown = function(e){
	var prototypeE = ModelEvent.prototype, currX, currY;
	if(typeof prototypeE.holdDownX != TYPE_NUMBER){
		if(typeof e.changedTouches != TYPE_UNDEFINED && e.changedTouches.length>0){
			prototypeE.holdDownX = e.changedTouches[0].pageX;
			prototypeE.holdDownY = e.changedTouches[0].pageY;
		}else if(typeof e.pageX != TYPE_UNDEFINED){
			prototypeE.holdDownX = e.pageX;
			prototypeE.holdDownY = e.pageY;
		}
	}else{
		if(typeof e.changedTouches != TYPE_UNDEFINED && e.changedTouches.length>0){
			currX = e.changedTouches[0].pageX;
			currY = e.changedTouches[0].pageY;
		}else if(typeof e.pageX != TYPE_UNDEFINED){
			currX = e.pageX;
			currY = e.pageY;
		}
	}
	if(typeof currX == TYPE_NUMBER){
		var dx = currX-prototypeE.holdDownX, dy = currY-prototypeE.holdDownY, r2 = dx*dx+dy*dy;
		if(r2>prototypeE.radius2){
			//cancel all due to motion
			prototypeE.fMouseUpHoldDown(e);
		}
	}

};

ModelEvent.prototype.fMouseUpHoldDown = function(e){
	var prototypeE = ModelEvent.prototype, aHD = prototypeE.aHoldDown, i;
	prototypeE.holdDownX = undefined;
	prototypeE.holdDownY = undefined;
	for(i=0; i<aHD.length; i++)
		window.clearTimeout(aHD[i]);
	aHD.splice(0);
	window.removeEventListener((MOUSE_OR_TOUCH == 0) ? "mousemove" : "touchmove", prototypeE.fMouseMoveHoldDown, true);
	window.removeEventListener((MOUSE_OR_TOUCH == 0) ? "mouseup" : "touchend", prototypeE.fMouseUpHoldDown, true);
};

//
ModelEvent.prototype.TOUCH_EVENTS_ARRAY = [
			["mousedown", "mouseup", "mousemove", "mouseout", "mouseover"],
			["touchstart", "touchend", "touchmove", "touchout", "touchover"],
			["MSPointerDown", "MSPointerUp", "MSPointerMove", "MSPointerOut", "MSPointerOver"],
			["pointerdown", "pointerup", "pointermove", "pointerout", "pointerover"]
	];
//
ModelEvent.prototype.VALIDATE_SCENE = "model_event_validate_scene";
ModelEvent.prototype.RESET_SCENE = "model_event_reset_scene";
ModelEvent.prototype.VERIFY_ELEMENTS = "model_event_verify_elements";
ModelEvent.prototype.SHOW_HINT = "model_event_show_hint";
ModelEvent.prototype.SHOW_RESPONSE = "model_event_show_response";
//
ModelEvent.prototype.CONFIRM_ANSWER = "model_event_confirm_answer";
ModelEvent.prototype.CLICK = "model_event_click";
ModelEvent.prototype.BUTTON_CLICK = "model_event_button_click";
ModelEvent.prototype.DOUBLE_CLICK = "model_event_double_click";
//
ModelEvent.prototype.LEFT_BUTTON = 0;
ModelEvent.prototype.MIDDLE_BUTTON = 1;
ModelEvent.prototype.RIGHT_BUTTON = 2;
//
ModelEvent.prototype.ITEM_SELECTED = "model_event_item_selected";
ModelEvent.prototype.FORM_CLOSED = "model_event_form_closed";
ModelEvent.prototype.MODEL_ELEMENT_SELECTED = "model_event_model_element_selected";
ModelEvent.prototype.MODEL_ELEMENT_DESELECTED = "model_event_model_element_deselected";
//
ModelEvent.prototype.REDRAW_WORKSPACE = "model_event_redraw_workspace";
ModelEvent.prototype.POSITION_WORKSPACE = "model_event_position_workspace";
//
ModelEvent.prototype.VALIDATE_INTERFACE = "model_event_validate_interface";
ModelEvent.prototype.INTERFACE_AREAS_SET = "model_event_interface_areas_set";
//
ModelEvent.prototype.ADD_DIALOG = "model_event_add_dialog";
ModelEvent.prototype.REMOVE_DIALOG = "model_event_remove_dialog";
//
ModelEvent.prototype.TAB_NO_SELECTED = "model_event_tab_selected";
//
ModelEvent.prototype.SEND_VALUE = "sendValue";
ModelEvent.prototype.GET_VALUE = "getValue";
//
ModelEvent.prototype.VALUE_CHANGED = "model_event_value_changed";
//
ModelEvent.prototype.ADDED = "model_event_added";
ModelEvent.prototype.REMOVED = "model_event_removed";
ModelEvent.prototype.REMOVED_ALL = "model_event_removed_all";
//
ModelEvent.prototype.SET_LOCATION = "model_event_set_location";
//
ModelEvent.prototype.TOOLBAR_BUTTON_EVENT = "model_event_toolbar_button_event";
ModelEvent.prototype.TOGGLE_DROPLIST = "model_event_toggle_droplist";
ModelEvent.prototype.OVER_DROPLIST = "model_event_over_droplist";
ModelEvent.prototype.OUT_DROPLIST = "model_event_out_droplist";
ModelEvent.prototype.SYMBOL_HIGHLIGHTED = "model_event_symbol_highlighted";
ModelEvent.prototype.SYMBOL_SELECTED = "model_event_symbol_selected";
ModelEvent.prototype.PALETTE_COLOR_SET = "model_event_palette_color_set";
ModelEvent.prototype.TOOL_EVENT = "model_event_tool_event";
ModelEvent.prototype.SET_TOOL = "model_event_set_tool";
ModelEvent.prototype.START_TOOL = "model_event_start_tool";
ModelEvent.prototype.USE_TOOL = "model_event_use_tool";
ModelEvent.prototype.TRACK_TOOL = "model_event_track_tool";
ModelEvent.prototype.END_TOOL = "model_event_end_tool";
ModelEvent.prototype.CANCEL_TOOL = "model_event_cancel_tool";
ModelEvent.prototype.ENABLE_TOOL = "model_event_enable_tool";
ModelEvent.prototype.DISABLE_TOOL = "model_event_disable_tool";
ModelEvent.prototype.CONFIGURE_TOOL = "model_event_configure_tool";
ModelEvent.prototype.APPLY_TOOL = "model_event_apply_tool";
//
ModelEvent.prototype.RELEASE = "model_event_release";
//
ModelEvent.prototype.CANVAS_TO_SCENE_RATIO_CHANGED = "model_event_canvas_to_scene_ratio_changed";
//
ModelEvent.prototype.SAVING_MODEL = "model_event_saving_model";
//
ModelEvent.prototype.MOUSE_EVENT = "model_event_mouse_event";
ModelEvent.prototype.KEYBOARD_EVENT = "model_event_keyboard_event";
//
ModelEvent.prototype.OVER_SCENE = "model_event_over_scene";
ModelEvent.prototype.OUT_SCENE = "model_event_out_scene";
//
ModelEvent.prototype.OVER_INTERFACE = "model_event_over_interface";
ModelEvent.prototype.OUT_INTERFACE = "model_event_out_interface";
//
ModelEvent.prototype.OVER_SELECTABLE_TEXT = "model_event_over_selectable_text";
ModelEvent.prototype.OUT_SELECTABLE_TEXT = "model_event_out_selectable_text";
//
ModelEvent.prototype.BLUR = "model_event_blur";
//
ModelEvent.prototype.SELECTED_ELEMENTS_MOVED = "model_event_selected_elements_moved";
//
ModelEvent.prototype.SCALE_FRAME_PUT = "model_event_scale_frame_put";
ModelEvent.prototype.MOVE_SCALE_HANDLE = "model_event_move_scale_handle";
ModelEvent.prototype.UP_SCALE_HANDLE = "model_event_up_scale_handle";
ModelEvent.prototype.DOWN_SCALE_HANDLE = "model_event_down_scale_handle";
ModelEvent.prototype.OVER_SCALE_HANDLE = "model_event_over_scale_handle";
ModelEvent.prototype.OUT_SCALE_HANDLE = "model_event_out_scale_handle";
ModelEvent.prototype.OVER_SCALE_FRAME = "model_event_over_scale_frame";
ModelEvent.prototype.OUT_SCALE_FRAME = "model_event_out_scale_frame";
//
ModelEvent.prototype.MOUSE_DOWN = "model_event_mouse_down";
ModelEvent.prototype.MOUSE_UP = "model_event_mouse_up";
ModelEvent.prototype.MOUSE_MOVE = "model_event_mouse_move";
ModelEvent.prototype.MOUSE_DRAG = "model_event_mouse_drag";
ModelEvent.prototype.MOUSE_DOWN_ON_ELEMENT = "model_event_mouse_down_on_element";
ModelEvent.prototype.MOUSE_UP_ON_ELEMENT = "model_event_mouse_up_on_element";
ModelEvent.prototype.MOUSE_MOVE_ON_ELEMENT = "model_event_mouse_move_on_element";
//
ModelEvent.prototype.OVER_SELECTABLE_TEXT = "model_event_over_selectable_text";
ModelEvent.prototype.OUT_SELECTABLE_TEXT = "model_event_out_selectable_text";
//
ModelEvent.prototype.INPUT = "model_event_input";
ModelEvent.prototype.TEXT_CHANGED = "model_event_text_changed";
ModelEvent.prototype.TOGGLE_SCREEN_KEYBOARD = "model_event_toggle_screen_keyboard";
//
ModelEvent.prototype.CHECKBOX_TOGGLED = "model_event_checkbox_toggled";
//
ModelEvent.prototype.IMPORT_SELECTED = "model_event_import_selected";
ModelEvent.prototype.IMPORT_CANCELLED = "model_event_import_cancelled";
ModelEvent.prototype.IMPORT_COMPLETE = "model_event_import_complete";
//
ModelEvent.prototype.ELEMENT_CREATED = "model_event_element_created";
ModelEvent.prototype.ELEMENT_REMOVE = "model_event_element_remove";
ModelEvent.prototype.ELEMENT_REMOVED = "model_event_element_removed";
//
ModelEvent.prototype.ELEMENT_SIZE_CHANGED = "model_event_element_size_changed";
ModelEvent.prototype.ELEMENT_POSITION_CHANGED = "model_event_element_position_changed";
ModelEvent.prototype.ELEMENT_SIZE_SET = "model_event_element_size_set";
//
ModelEvent.prototype.ELEMENT_UPDATE = "model_event_element_update";
//
ModelEvent.prototype.PAN_END = "model_event_pan_end";
ModelEvent.prototype.PAN_START = "model_event_pan_start";
ModelEvent.prototype.PAN_RESET = "model_event_pan_reset";
//
ModelEvent.prototype.ON_DATA = "model_event_on_data";
//
ModelEvent.prototype.SCROLL_X = "model_event_scroll_x";
ModelEvent.prototype.SCROLL_Y = "model_event_scroll_y";
//
ModelEvent.prototype.SCROLL_END = "model_event_scroll_end";
//
ModelEvent.prototype.SCROLL_TOGGLE = "model_event_scroll_toggle";
//
ModelEvent.prototype.SPINBOX_PREVIEW_START = "model_event_spinbox_preview_start";
ModelEvent.prototype.SPINBOX_PREVIEW_END = "model_event_spinbox_preview_end";
//
ModelEvent.prototype.BUTTON_EVENT = "model_event_button_event";
//
ModelEvent.prototype.WINDOW_RESIZE = "window_resize_event";
//
ModelEvent.prototype.DOUBLE_CLICK_INTERVAL = 350;
ModelEvent.prototype.HOLD_DOWN_TIME = 500;
ModelEvent.prototype.HOLD_DOWN_JITTER_RADIUS = 5;
ModelEvent.prototype.DOUBLE_TOUCH_JITTER_RADIUS = 10;
//

function ResizeHandler(_window){
	
	//delay, in ms, before the assigned redraw procedure is called
	const DEFAULT_UPDATE_DELAY = 500;
	
	var updateDelay;
	var timeoutId = -1;
	var me = null;
    var wnd = _window;
	
	function resizeHandler(_updateDelay){
		resizeHandler.superclass.constructor.call(this);
		me = this;
		updateDelay = (typeof _updateDelay == TYPE_NUMBER) ? _updateDelay :DEFAULT_UPDATE_DELAY;
		wnd.addEventListener("resize", eventHandler);
        wnd.addEventListener("unload", fOnUnload);
        //wnd.addEventListener("pagehide", fOnUnload);
	}
	
	function fDispatch(){
		//window.innerWidth, window.innerHeight
		var evt = new ModelEvent(ModelEvent.prototype.WINDOW_RESIZE, true, true);
		if(evt!=null && me!=null){
			evt.data = {width:window.innerWidth, height:window.innerHeight};
			me.dispatchEvent(evt);
		}
		if(timeoutId != -1){
			window.clearTimeout(timeoutId);
			timeoutId = -1;
		}
	}
	
	function eventHandler(e){
		if(timeoutId != -1)
			window.clearTimeout(timeoutId);
		timeoutId = window.setTimeout(fDispatch, updateDelay);
	}

    function fOnUnload(e){
        me.deactivate();
        fRemoveListeners();
    }

    function fRemoveListeners(){
        wnd.removeEventListener("resize", eventHandler);
        wnd.removeEventListener("unload", fOnUnload);
        //wnd.removeEventListener("pagehide", fOnUnload);
    }
	
	extend(resizeHandler, NonDOMEventDispatcher);
	
	return new resizeHandler();
}

/**
 * @constructor
 */
function HitTestInfo(_element, _flagHit, _flagInside, _details){
    var me = this;
    me.element = _element;
    me.flagHit = _flagHit;
    me.flagInside = _flagInside;
    //for complex objects like bound text that have more than two hit areas
    me.details = (typeof _details != TYPE_UNDEFINED) ? _details : null;
}
extend(HitTestInfo, Object);

HitTestInfo.prototype.getElement = function(){
    return this.element;
};

HitTestInfo.prototype.isHit = function(){
    return this.flagHit;
};

HitTestInfo.prototype.isInside = function(){
    return this.flagInside;
};

HitTestInfo.prototype.hasDetails = function(){
    return this.details != null;
};

HitTestInfo.prototype.getDetails = function(){
    return this.details;
};

//override
HitTestInfo.prototype.toString = function(){
    var me = this;
    var a = [];
    a.push(me.flagHit);
    a.push(me.flagInside);
    var attr;
    if(me.details != null) {
        for (attr in me.details)
            a.push(me.details[attr]);
    }
    return a.join(",");
};
/**
 * @constructor
 */
function TextFunctions(){
    //calculation of font-related offsets and other such parameters
    /*var me = this;
    var cnv = me.measuringCanvas;
    var ctx = cnv.getContext("2d");
    var testString = "АБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдеёжзийклмнопрстуфхцчшщъыьэюя01234567890";
    var w, wExpected, r;
    var i, j, q, qS;
    var a = me.aScalingFactor;
    var aFontModifier = ["", "bold", "italic", "bold italic"];
    for(i=0; i<a.length; i++){
        q = a[i].size;
        qS = ""+q;
        for(j=0; j<aFontModifier.length; j++){
            ctx.font = aFontModifier[j].concat(" ").concat(qS).concat("px Arial");
            w = ctx.measureText(testString).width;
            wExpected = a[i].aExpected[j];
            r = wExpected/w;
            a[i].aXFactor[j] = r;
        }
    }*/
}
extend(TextFunctions, Object);
//
TextFunctions.prototype.measuringCanvas = document.createElement("CANVAS");

//second symbol is Alt+0160
TextFunctions.prototype.LINE_BREAK_SYMBOLS =
    [" ", " ", ".", ",", ";", "-", "–", "—", "(", ")", "[", "]", "/", "+", "=", "&", "#"];
//second symbol is Alt+0160
TextFunctions.prototype.SPACE_SYMBOLS =
    [" ", " "];
//
TextFunctions.prototype.aLineHeightAndVerticalShift = [
    [10, 11, 10, 2.075],
    [11, 12, 11, 2.125],
    [12, 14, 12, 1.675],
    [14, 16, 14, 1.800],
    [16, 18, 16, 1.900],
    [18, 20, 18, 2.025],
    [20, 22, 19, 2.150],
    [22, 25, 21, 1.775],
    [24, 27, 23, 1.875],
    [26, 29, 25, 2.000],
    [28, 31, 26, 2.100],
    [36, 40, 35, 2.075],
    [48, 54, 45, 1.800],
    [72, 81, 66, 1.700],
    [90, 101, 83, 1.750],
    [120, 134, 110, 2.000]
];

/*TextFunctions.prototype.aScalingFactor = [
    {size:4, aXFactor:[], aExpected:[195, 199, 196, 203]},
    {size:5, aXFactor:[], aExpected:[253, 265, 253, 269]},
    {size:6, aXFactor:[], aExpected:[289, 302, 290, 308]},
    {size:7, aXFactor:[], aExpected:[349, 354, 352, 366]},
    {size:8, aXFactor:[], aExpected:[381, 407, 385, 413]},
    {size:9, aXFactor:[], aExpected:[444, 467, 448, 478]},
    {size:10, aXFactor:[], aExpected:[501,519,506,525]},
    {size:11, aXFactor:[], aExpected:[529,563,543,582]},
    {size:12, aXFactor:[], aExpected:[592,609,603,631]},
    {size:14, aXFactor:[], aExpected:[679,725,698,724]},
    {size:16, aXFactor:[], aExpected:[782,824,792,835]},
    {size:18, aXFactor:[], aExpected:[880,927,891,939]},
    {size:20, aXFactor:[], aExpected:[977,1030,990,1043]},
    {size:22, aXFactor:[], aExpected:[1075,1133,1089,1148]},
    {size:24, aXFactor:[], aExpected:[1173,1236,1188,1252]},
    {size:26, aXFactor:[], aExpected:[1271,1339,1287,1356]},
    {size:28, aXFactor:[], aExpected:[1368,1442,1386,1461]},
    {size:36, aXFactor:[], aExpected:[1759,1854,1782,1878]},
    {size:48, aXFactor:[], aExpected:[2346,2472,2376,2504]},
    {size:72, aXFactor:[], aExpected:[3519,3708,3564,3756]},
    {size:90, aXFactor:[], aExpected:[4398,4635,4455,4695]},
    {size:120, aXFactor:[], aExpected:[5864,6180,5939,6260]}
];

//
TextFunctions.prototype.getXFactorByFontSize = function(fontSize, flagBold, flagItalic){
    var a = TextFunctions.prototype.aScalingFactor, n = a.length, i, v, j;
    j = flagBold ? (flagItalic ? 3 : 1) : (flagItalic ? 2 : 0);
    for (i = 0; i < n; i++) {
        v = a[i].size;
        if (v >= fontSize)
            break;
    }
    if(v == fontSize)
        return a[i].aXFactor[j];
    var v1, v2, r1, r2;
    if (i == n) {
        //extrapolate
        v1 = a[n-2].aXFactor[j];
        v2 = a[n-1].aXFactor[j];
        r1 = a[n-2].size;
        r2 = a[n-1].size;
    }else if(i==0){
        //extrapolate
        v1 = a[0].aXFactor[j];
        v2 = a[1].aXFactor[j];
        r1 = a[0].size;
        r2 = a[1].size;
    }else{
        //interpolate
        v1 = a[i-1].aXFactor[j];
        v2 = a[i].aXFactor[j];
        r1 = a[i-1].size;
        r2 = a[i].size;
    }
    return (v2-v1)/(r2-r1)*(fontSize-r2)+v2;
};*/

//
TextFunctions.prototype.getTextParameterByFontSize = function(fontSize, parameterNo) {
    var a = TextFunctions.prototype.aLineHeightAndVerticalShift, n = a.length, i, v;
    for (i = 0; i < n; i++) {
        v = a[i][0];
        if (v >= fontSize)
            break;
    }
    if (v == fontSize)
        return a[i][parameterNo];

    var v1, v2, fontSize1, fontSize2;
    if (i == n) {
        //extrapolate
        v1 = a[n-2][parameterNo];
        v2 = a[n-1][parameterNo];
        fontSize1 = a[n-2][0];
        fontSize2 = a[n-1][0];
    }else if(i==0){
        //extrapolate
        v1 = a[0][parameterNo];
        v2 = a[1][parameterNo];
        fontSize1 = a[0][0];
        fontSize2 = a[1][0];
    }else{
        //interpolate
        v1 = a[i-1][parameterNo];
        v2 = a[i][parameterNo];
        fontSize1 = a[i-1][0];
        fontSize2 = a[i][0];
    }
    return (v2-v1)/(fontSize2-fontSize1)*(fontSize-fontSize2)+v2;
};



//
TextFunctions.prototype.getLineHeightByFontSize = function (fontSize) {
    return TextFunctions.prototype.getTextParameterByFontSize(fontSize, 1);
};

//
TextFunctions.prototype.getVerticalShiftByFontSize = function (fontSize) {
    return TextFunctions.prototype.getTextParameterByFontSize(fontSize, 2);
};

//
TextFunctions.prototype.getVerticalPaddingByFontSize = function (fontSize) {
    return TextFunctions.prototype.getTextParameterByFontSize(fontSize, 3);
};

//
//
TextFunctions.prototype.splitLineForContextAndWidth = function(s, ctx, w0, w1, reduceSpaces){
    //w1 is the available width in the first line, might be less or equal to w0
    var flagReduceSpaces = (typeof reduceSpaces != TYPE_BOOLEAN) || reduceSpaces;
    var aSplit = [];
    var i, s1 = s, s2 = null;
    var w = ctx.measureText(s).width;
    var wFirst = (typeof w1 == TYPE_NUMBER) ? w1 : w0;
    var a, n, c, j, flagSpace;

    a = TextFunctions.prototype.LINE_BREAK_SYMBOLS;
    n = a.length;
    for(j=0; j<n; j++){
        c = a[j];
        flagSpace = flagReduceSpaces && (TextFunctions.prototype.SPACE_SYMBOLS.indexOf(c) != -1);
        i = 0x7FFFFFFF;
        while(w>wFirst){
            i = s.lastIndexOf(c, i-1);
            if(i != -1){
                s1 =flagSpace ? s.slice(0, i) : s.slice(0, i+1);
                s2 = s.slice(i + 1);
            }else
                break;
            w = ctx.measureText(s1).width;
        }
        if(w<=wFirst)
            break;
    }

    i = s.length;
    if(i>1) {
        //if wFirst is shorter than the common line width, the word should be carried to the next line, not split
        if(w>wFirst) {
            if (wFirst == w0) {
                while (w > wFirst) {
                    i--;
                    s1 = s.slice(0, i);
                    s2 = (s.charAt(i) == ' ') ? s.slice(i + 1) : s.slice(i);
                    w = ctx.measureText(s1).width;
                    if (s1.length == 1)
                        break;
                }
            } else {
                s1 = "";
                s2 = s;
            }
        }
        aSplit.push(s1);
        if (s2 != null)
            aSplit = aSplit.concat(TextFunctions.prototype.splitLineForContextAndWidth(s2, ctx, w0, w0, reduceSpaces));
    }else if(i>0)
        aSplit.push(s);
    return aSplit;
};


//
TextFunctions.prototype.getTextBlocksFromNode = function(node){
    //returns an array of {isTag, fontStyleObject, text}
    var a = [];
    var nodeCurr = node;
    var i, n = nodeCurr.hasChildNodes() ? nodeCurr.childNodes.length : 0;
    var aElRet = [], aIRet = [], aNRet = [], aFontRet = [], flagContinue;
    var nd, obj, aFace;
    var objFont, fontSize, fontFace, flagBold, flagItalic, fontColor;
    for(i=0; i<n; i++){
        nd = nodeCurr.childNodes[i];
        obj = {};
        flagContinue = false;
        switch(nd.nodeType){
            case Node.ELEMENT_NODE:
                obj.isTag = true;
                obj.tagName = nd.nodeName;

                switch(obj.tagName.toLowerCase()){
                    case "font":
                        aFace = (nd.hasAttribute("face") ? nd.getAttribute("face") : "Arial").split("_");
                        fontFace = aFace[0];
                        flagBold = aFace.indexOf("bold") != -1;
                        flagItalic = aFace.indexOf("italic") != -1;
                        fontSize = (nd.hasAttribute("size")) ? parseInt(nd.getAttribute("size")) : 14;
                        fontColor = nd.hasAttribute("color") ? nd.getAttribute("color") : "#000000";
                        objFont = {face:fontFace, size:fontSize, bold:flagBold, italic:flagItalic, color:fontColor};
                        break;
                    case "br":
                        obj.font = miscFunctions.shallowCloneObject(objFont);
                        a.push(obj);
                        break;
                    case "tag_s":
                        //a space
                        obj.isTag = false;
                        obj.text = " ";
                        obj.font = miscFunctions.shallowCloneObject(objFont);
                        a.push(obj);
                        break;
                    default:
                }

                if(nd.hasChildNodes() && nd.childNodes.length>0){
                    flagContinue = true;
                    aFontRet.push(objFont);
                    aIRet.push(i);
                    aNRet.push(n);
                    aElRet.push(nodeCurr);
                    i = -1;
                    n = nd.childNodes.length;
                    nodeCurr = nd;
                }

                break;
            case Node.TEXT_NODE:
                obj.isTag = false;
                obj.text = nd.textContent || nd.innerText;
                obj.font = miscFunctions.shallowCloneObject(objFont);
                a.push(obj);
                break;
            default:
        }

        if(flagContinue)
            continue;

        if((i == n-1) && aIRet.length>0){
            objFont = aFontRet.pop();
            i = aIRet.pop();
            n = aNRet.pop();
            nodeCurr = aElRet.pop();
        }
    }
    return a;
};


TextFunctions.prototype.getTextBlocksForWidthAndScaleFactor = function(aText, width, scaleFactor){
    //returns an array of objects with texts, coordinates, colors and font for canvas context "fillText"
    //actual text width is width*scaleFactor, font sizes are multiplied by scaleFactor
    var rSize = (typeof scaleFactor == TYPE_NUMBER) ? scaleFactor : 1.0;
    var w = width*rSize;
    var wCnv = Math.ceil(w);
    var cnv = TextFunctions.prototype.measuringCanvas;
    var ctx = cnv.getContext("2d");
    if(cnv.width != wCnv)
        cnv.width = wCnv;
    if(cnv.height<100)
        cnv.height = 100;
    var xCurr = 0;
    var yCurr = 0;
    var lineNumber = 0;
    var lineHeight;
    var a = [];
    var aTmp;
    var i, j, n = aText.length, obj, font;
    var textBlock;
    var fontSize;
    var sText, sFont;
    for(i=0; i<n; i++){
        obj = aText[i];
        fontSize = obj.font.size*rSize;
        lineHeight = TextFunctions.prototype.getLineHeightByFontSize(fontSize);
        if(i==0) {
            yCurr += TextFunctions.prototype.getVerticalShiftByFontSize(fontSize);
        }
        if(obj.isTag){
            switch(obj.tagName){
                case "br":
                    lineNumber++;
                    yCurr += lineHeight;
                    xCurr = 0;
                    break;
            }
        }else{
            font = obj.font;
            ctx.font = sFont =
                (font.italic ? "italic" : "") + " " + (font.bold ? "bold" : "") + " " +
                fontSize + "px " + font.face;
            aTmp = TextFunctions.prototype.splitLineForContextAndWidth(obj.text, ctx, w, w-xCurr);
            for(j=0; j<aTmp.length; j++){
                sText = aTmp[j];
                if(sText.length>0) {
                    textBlock = {
                        x:xCurr, y:yCurr, width:ctx.measureText(sText).width,
                        lineNumber:lineNumber, lineHeight:lineHeight
                    };
                    textBlock.text = sText;
                    textBlock.font = sFont;
                    textBlock.color = font.color;
                    a.push(textBlock);
                }
                if(j != aTmp.length-1) {
                    lineNumber++;
                    yCurr += lineHeight;
                    xCurr = 0;
                }
            }
            if(aTmp.length>0) {
                xCurr += ctx.measureText(aTmp[aTmp.length - 1]).width;
                if(xCurr>w){
                    lineNumber++;
                    yCurr += lineHeight;
                    xCurr = 0;
                }
            }
        }
    }
    return a;
};
/**
 * @constructor
 */
function ModelLocalization(_localizationXML){
    var me = this;
    me.objLoc = {strings:{}, parameters:{}};
    me.setSourceXML(_localizationXML);
}
extend(ModelLocalization, Object);

//
ModelLocalization.prototype.setSourceXML = function(_localizationXML){
    var me = this;
    me.localizationXML = _localizationXML;
    var localizationXML = me.localizationXML;
    var cNodes = localizationXML.childNodes, node;
    var i, n = cNodes.length, nn;

    for(i=0; i<n; i++) {
        node = cNodes[i];
        if (node.nodeType == 1 && node.nodeName.toLowerCase() == "localization") {
            cNodes = node.childNodes;
            nn = cNodes.length;
            break;
        }
    }
    if(i==n)
        return;

    n = nn;

    var objLoc = me.objLoc, obj;
    var objLocStr = objLoc.strings;
    var objLocPar = objLoc.parameters;
    var j, l;
    var aSId, s, sType, sValue;
    me.stringsNode = null;
    me.parametersNode = null;
    for(i=0; i<n; i++){
        node = cNodes[i];
        if(node.nodeType == 1) {
            switch(node.nodeName.toLowerCase()){
                case "strings":
                    me.stringsNode = node;
                    break;
                case "parameters":
                    me.parametersNode = node;
                    break;
                default:
            }
        }
    }
    if(me.stringsNode == null || me.parametersNode == null)
        return;

    cNodes = me.stringsNode.childNodes;
    n = cNodes.length;
    for(i=0; i<n; i++){
        node = cNodes[i];
        me.parseNode(objLocStr, node, true);

        /*if(node.nodeType != 1 || node.nodeName.toLowerCase() != "item")
            continue;
        if(typeof node.attributes.id == TYPE_UNDEFINED || typeof node.attributes.value == TYPE_UNDEFINED)
            continue;
        aSId = node.attributes.id.value.split(".");
        l = aSId.length-1;
        obj = objLocStr;
        for(j=0; j<l; j++) {
            s = aSId[j];
            if (typeof obj[s] != TYPE_OBJECT)
                obj[s] = {};
            obj = obj[s];
        }
        obj[aSId[l]] = node.attributes.value.value;*/
    }
    /////////////parameters///////////////
    cNodes = me.parametersNode.childNodes;
    n = cNodes.length;
    for(i=0; i<n; i++){
        node = cNodes[i];
        me.parseNode(objLocPar, node, false);
        /*
        if(node.nodeType != 1 || node.nodeName.toLowerCase() != "item")
            continue;
        if(
            typeof node.attributes.id == TYPE_UNDEFINED ||
            typeof node.attributes.value == TYPE_UNDEFINED ||
            typeof node.attributes.type == TYPE_UNDEFINED
        )
            continue;
        aSId = node.attributes.id.value.split(".");
        l = aSId.length-1;
        obj = objLocPar;
        for(j=0; j<l; j++) {
            s = aSId[j];
            if (typeof obj[s] != TYPE_OBJECT)
                obj[s] = {};
            obj = obj[s];
        }
        sType = node.attributes.type.value.toLowerCase();
        sValue = node.attributes.value.value;
        switch(sType){
            case "int":
            case "integer":
                obj[aSId[l]] = parseInt(sValue);
                break;
            case "float":
            case "number":
                obj[aSId[l]] = parseFloat(sValue);
                break;
            case "bool":
            case "boolean":
                obj[aSId[l]] = (sValue.toLowerCase() == "true");
                break;
            case "string":
            default:
                obj[aSId[l]] = sValue;
        }
        */
    }
};

//
ModelLocalization.prototype.parseNode = function(objLoc, node, _flagString){
    var me = this;
    var flagString = (typeof _flagString == TYPE_UNDEFINED) ? true : _flagString;
    if(node.nodeType != 1)
        return;
    if(node.nodeName.toLowerCase() != "item" && node.nodeName.toLowerCase() != "set")
        return;
    if(typeof node.attributes.id == TYPE_UNDEFINED)
        return;

    var sType, sValue;
    if(node.nodeName.toLowerCase() == "item") {
        if(typeof node.attributes.value == TYPE_UNDEFINED)
            return;
        if(flagString)
            sType = "string";
        else if(typeof node.attributes.type != TYPE_UNDEFINED)
            sType = node.attributes.type.value.toLowerCase();
        else
            return;
        sValue = node.attributes.value.value;
    }else {
        sType = null;
        sValue = null;
    }

    var aSId, s, obj;
    var flagSet = (node.nodeName.toLowerCase() == "set");
    aSId = node.attributes.id.value.split(".");
    var j, l = aSId.length - (flagSet ? 0 : 1);
    obj = objLoc;
    for(j=0; j<l; j++) {
        s = aSId[j];
        if (typeof obj[s] != TYPE_OBJECT)
            obj[s] = {};
        obj = obj[s];
    }
    if(flagSet){
        var cNodes = node.childNodes, cNode;
        var i, n = cNodes.length;
        for(i=0; i<n; i++){
            cNode = cNodes[i];
            me.parseNode(obj, cNode, flagString);
        }
    }else{
        switch(sType){
            case "int":
            case "integer":
                obj[aSId[l]] = parseInt(sValue);
                break;
            case "float":
            case "number":
                obj[aSId[l]] = parseFloat(sValue);
                break;
            case "bool":
            case "boolean":
                obj[aSId[l]] = (sValue.toLowerCase() == "true");
                break;
            case "string":
            default:
                obj[aSId[l]] = sValue;
        }
    }
};

//
ModelLocalization.prototype.fGetValue = function(rootObj, sId, defaultValue){
    var sDefault = (typeof defaultValue == TYPE_UNDEFINED) ? null : defaultValue;
    var aSId = sId.split(".");
    var i, s, objItem, q = aSId.length-1;
    objItem = rootObj;
    for(i=0; i<q; i++){
        s = aSId[i];
        if(typeof objItem[s] != TYPE_UNDEFINED)
            objItem = objItem[s];
        else
            break;
    }
    if(i<q)
        return sDefault;
    s = aSId[q];
    return (typeof objItem[s] == TYPE_UNDEFINED) ? sDefault : objItem[s];
};

//
ModelLocalization.prototype.getString = function(sId, defaultValue){
    return this.fGetValue(this.objLoc.strings, sId, defaultValue);
};

//
ModelLocalization.prototype.getParameter = function(sId, defaultValue){
    return this.fGetValue(this.objLoc.parameters, sId, defaultValue);
};

/**
 * @constructor
 */
function ModelResource(_modelXML){
	var me = this;
	
	this.nodeObjects = null;
	this.nodeCustomizePlayer = null;

    this.objectsAttributes = {};
	this.objectsData = {};

    this.customizePlayerAttributes = {};
    this.customizePlayerData = {};
	
	me.setSourceXML(_modelXML);
}
extend(ModelResource, Object);

//префиксы для разделов и отдельных объектов могут не совпадать, например <IMG_objects/> и <IM0/>
ModelResource.prototype.TYPE_IMAGES = "IMG";
ModelResource.prototype.TYPE_IMAGE = "IM";
//see %%LOCAL_IMAGES_DIRECTORY%% in mapkitModelStart. This "variable" is set by converter
ModelResource.prototype.LOCAL_IMAGES_DIRECTORY = LOCAL_IMAGES_DIRECTORY;//"files/";
ModelResource.prototype.CLASSID_IMAGE = "ImageModelElement";

ModelResource.prototype.TYPE_POINTS = "PO";
ModelResource.prototype.TYPE_POINT = "PO";
ModelResource.prototype.CLASSID_POINT = "PointModelElement";

ModelResource.prototype.TYPE_RECTANGLES = "RO";
ModelResource.prototype.TYPE_RECTANGLE = "RO";
ModelResource.prototype.TYPE_ELLIPSES = "EO";
ModelResource.prototype.TYPE_ELLIPSE = "EO";
ModelResource.prototype.CLASSID_RECTANGLE = "RectangleModelElement";

ModelResource.prototype.TYPE_PLOS = "PLO";
ModelResource.prototype.TYPE_PLO = "PLO";
ModelResource.prototype.CLASSID_PLO = "PolylineModelElement";

ModelResource.prototype.TYPE_TEXTS = "TO";
ModelResource.prototype.TYPE_TEXT = "TO";
ModelResource.prototype.CLASSID_TEXT = "TextModelElement";

ModelResource.prototype.TYPE_BTEXTS = "BTO";
ModelResource.prototype.TYPE_BTEXT = "BTO";
ModelResource.prototype.CLASSID_BTEXT = "BtextModelElement";

ModelResource.prototype.TYPE_ARROWS = "AO";
ModelResource.prototype.TYPE_ARROW = "AO";
ModelResource.prototype.CLASSID_ARROW = "ArrowModelElement";

ModelResource.prototype.TYPE_DIAGRAMS = "DGO";
ModelResource.prototype.TYPE_DIAGRAM = "DGO";
ModelResource.prototype.CLASSID_DIAGRAM = "DiagramModelElement";

ModelResource.prototype.TYPE_CHECKBOXSETS = "CBO";
ModelResource.prototype.TYPE_CHECKBOXSET = "CBO";
ModelResource.prototype.CLASSID_CHECKBOXSET = "CheckboxSetModelElement";

ModelResource.prototype.TYPES = [
		{
            prefix:ModelResource.prototype.TYPE_IMAGE,
            family:ModelResource.prototype.TYPE_IMAGES, classId:ModelResource.prototype.CLASSID_IMAGE
        },
		{
            prefix:ModelResource.prototype.TYPE_POINT,
            family:ModelResource.prototype.TYPE_POINTS, classId:ModelResource.prototype.CLASSID_POINT
        },
		{
            prefix:ModelResource.prototype.TYPE_RECTANGLE, prefixData:ModelResource.prototype.TYPE_RECTANGLE,
            family:ModelResource.prototype.TYPE_RECTANGLES, classId:ModelResource.prototype.CLASSID_RECTANGLE
        },
	    //prefixData property is for subtyped elements (rectangle and ellipse) so that they are stored in the same array "elements" of MapSceneModel
        {
            prefix:ModelResource.prototype.TYPE_ELLIPSE, prefixData:ModelResource.prototype.TYPE_RECTANGLE,
            family:ModelResource.prototype.TYPE_ELLIPSES, classId:ModelResource.prototype.CLASSID_RECTANGLE
        },
        {
            prefix:ModelResource.prototype.TYPE_PLO,
            family:ModelResource.prototype.TYPE_PLOS, classId:ModelResource.prototype.CLASSID_PLO
        },
		{
            prefix:ModelResource.prototype.TYPE_TEXT,
            family:ModelResource.prototype.TYPE_TEXTS, classId:ModelResource.prototype.CLASSID_TEXT
        },
		{
            prefix:ModelResource.prototype.TYPE_BTEXT,
            family:ModelResource.prototype.TYPE_BTEXTS, classId:ModelResource.prototype.CLASSID_BTEXT
        },
		{
            prefix:ModelResource.prototype.TYPE_ARROW,
            family:ModelResource.prototype.TYPE_ARROWS, classId:ModelResource.prototype.CLASSID_ARROW
        },
		{
            prefix:ModelResource.prototype.TYPE_DIAGRAM,
            family:ModelResource.prototype.TYPE_DIAGRAMS, classId:ModelResource.prototype.CLASSID_DIAGRAM
        },
        {
            prefix:ModelResource.prototype.TYPE_CHECKBOXSET,
            family:ModelResource.prototype.TYPE_CHECKBOXSETS, classId:ModelResource.prototype.CLASSID_CHECKBOXSET
        }
];

ModelResource.prototype.OBJECTS_ATTRIBUTES_PROPERTY = "objectsAttributes";
ModelResource.prototype.OBJECTS_PROPERTY = "objectsData";

ModelResource.prototype.CUSTOMIZE_PLAYER_ATTRIBUTES_PROPERTY = "customizePlayerAttributes";
ModelResource.prototype.CUSTOMIZE_PLAYER_PROPERTY = "customizePlayerData";

ModelResource.prototype.SUFFIX_OBJECTS = "_objects";

ModelResource.prototype.TAG_MAP = "Map";
ModelResource.prototype.TAG_TASK = "Task";
ModelResource.prototype.TAG_COMMENT = "Comment";
ModelResource.prototype.TAG_HINT_SOLUTION = "HintSolution";
ModelResource.prototype.TAG_GRID = "Setka";
ModelResource.prototype.TAG_TOOLBAR = "ToolBar";
ModelResource.prototype.TAG_COLOR_DEFAULTS = "ColorDefaults";
ModelResource.prototype.TAG_OBJECTS = "Objects";
ModelResource.prototype.TAG_CUSTOMIZE_PLAYER = "CustomizePlayer";

ModelResource.prototype.setSourceXML = function(_modelXML){
	this.modelXML = _modelXML;
	var modelXML = this.modelXML;
	var cNodes = modelXML.childNodes, node;
	var i, l = cNodes.length;
	for(i=0; i<l; i++){
		node = cNodes[i];
		if(node.nodeType == 1)
			break;
	}
	var rootNode = node;
	this.rootNode = rootNode;
	//miscFunctions.getFirstElementChildNode(modelXML);
	
	cNodes = rootNode.childNodes;
	l = cNodes.length;
	for(i=0; i<l; i++){
		node = cNodes[i];
		if(node.nodeType != 1)
			continue;
		switch(node.nodeName){
            case ModelResource.prototype.TAG_MAP:
                this.nodeMap = node;
                break;
            case ModelResource.prototype.TAG_TASK:
                this.nodeTask = node;
                break;
            case ModelResource.prototype.TAG_COMMENT:
                this.nodeComment = node;
                break;
            case ModelResource.prototype.TAG_HINT_SOLUTION:
                this.nodeHintSolution = node;
                break;
            case ModelResource.prototype.TAG_GRID:
                this.nodeSetka = node;
                break;
            case ModelResource.prototype.TAG_TOOLBAR:
                this.nodeToolbar = node;
                break;
            case ModelResource.prototype.TAG_COLOR_DEFAULTS:
                this.nodeColorDefaults = node;
                break;
			case ModelResource.prototype.TAG_OBJECTS:
				this.nodeObjects = node;
				break;
			case ModelResource.prototype.TAG_CUSTOMIZE_PLAYER:
				this.nodeCustomizePlayer = node;
				break;
			default:
		}
	}

    this.retrievePlayerCustomization(this.nodeCustomizePlayer);
	this.retrieveObjects(this.nodeObjects);
};

ModelResource.prototype.retrieveObjects = function(_nodeObjects){
	if(_nodeObjects == null)
		return;
    var me = this;
    var attribs = _nodeObjects.attributes, attrib;
    var ii, nn = attribs.length;
    for(ii=0; ii<nn; ii++){
        attrib = attribs[ii];
        me.objectsAttributes[attrib.name] = ""+attrib.value;
    }
	var cNodes = _nodeObjects.childNodes, node;
    var aType = ModelResource.prototype.TYPES;
	var i, l = cNodes.length, j, k, n = aType.length;
	var sNodeName, suffix = ModelResource.prototype.SUFFIX_OBJECTS;
	for(i=0; i<l; i++){
		node = cNodes[i];
		if(node.nodeType != 1)
			continue;
		sNodeName = node.nodeName;
		j = sNodeName.indexOf(suffix);
		if(j == -1)
			continue;
		var sType = sNodeName.slice(0,j), sTypeObj = null;
        for(k=0; k<n; k++)
            if(aType[k].family == sType)
                break;
        sTypeObj = (k<n) ? aType[k].prefix : null;
		if(sTypeObj != null){
			if(typeof me.objectsData[sType] == TYPE_UNDEFINED){
				me.objectsData[sType] = me.retrieveObjectsWithinTypeNode(sTypeObj, node);
			}
		}
	}
};

ModelResource.prototype.retrievePlayerCustomization = function(_nodeCustomizePlayer){
    if(_nodeCustomizePlayer == null)
        return;
    var me = this;
    var cNodes = _nodeCustomizePlayer.childNodes, node;
    var attribs = _nodeCustomizePlayer.attributes, attrib;
    var i, n = attribs.length, sNodeName, ii, nn;
    for(i=0; i<n; i++){
        attrib = attribs[i];
        me.customizePlayerAttributes[attrib.name] = ""+attrib.value;
    }
    n = cNodes.length;
    for(i=0; i<n; i++) {
        node = cNodes[i];
        if (node.nodeType != 1)
            continue;
        sNodeName = node.nodeName;
        if (typeof me.customizePlayerData[sNodeName] == TYPE_UNDEFINED) {
            var obj = me.customizePlayerData[sNodeName] = {};
            attribs = node.attributes;
            nn = attribs.length;
            for(ii = 0; ii<nn; ii++) {
                attrib = attribs[ii];
                obj[attrib.name] =  ""+attrib.value;
            }
        }
    }
};

ModelResource.prototype.retrieveObjectsWithinTypeNode = function(_sType, _nodeSameTypeObjects){
	var a = [];
	if(_nodeSameTypeObjects == null)
		return a;
	var cNodes = _nodeSameTypeObjects.childNodes, node;
	var i, l = cNodes.length, j, n;
	var attribs, attrib, obj;
    var cNodes2, node2, ii, nn;
	for(i=0; i<l; i++){
		node = cNodes[i];
		if(node.nodeType != 1 || !node.hasAttributes())
			continue;
		j = node.nodeName.indexOf(_sType);
		if(j != 0)
			continue;
		attribs = node.attributes;
		n = attribs.length;
		obj = {};
		for(j=0; j<n; j++){
			attrib = attribs[j];
			obj[attrib.name] = attrib.value;
		}
        if(node.hasChildNodes()){
            obj.aNode = [];
            cNodes2 = node.childNodes;
            nn = cNodes2.length;
            for(ii=0; ii<nn; ii++){
                node2 = cNodes2[ii];
                if(node2.nodeType != 1)
                    continue;
                obj.aNode.push(node2);
            }
        }
		a.push(obj);
	}
	return a;
};


ModelResource.prototype.getNode = function(sName){
	
};

ModelResource.prototype.getImageFileList = function(_flagNamesOnly){
	var aIMG = this.objectsData.IMG;
	var a = [];
	if(!Array.isArray(aIMG))
		return a;
	var obj, i, j, l = aIMG.length, s;
	var flagNamesOnly = (typeof _flagNamesOnly != TYPE_UNDEFINED) ? _flagNamesOnly : false;
	for(i=0; i<l; i++){
		obj = aIMG[i];
		if(typeof obj["File"] == TYPE_STRING && typeof obj["Id"] == TYPE_STRING){
			s = obj["File"].replace(/\\/g,"/");
			if(flagNamesOnly){
				j = s.lastIndexOf("/");
				if(j!=-1)
					s = s.slice(j+1);
			}
			a.push({id:obj["Id"].slice(0), src:s});
		}
	}
	return a;
};

ModelResource.prototype.makeImageURLsLocal = function(){
    var aIMG = this.objectsData["IMG"];
    if(!Array.isArray(aIMG))
        return;
    var obj, i, j, l = aIMG.length, s;
    for(i=0; i<l; i++){
        obj = aIMG[i];
        if(typeof obj["File"] == TYPE_STRING){
            s = obj["File"].replace(/\\/g,"/");
            j = s.lastIndexOf("/");
            if(j!=-1)
                s = s.slice(j+1);
            obj["File"] = ModelResource.prototype.LOCAL_IMAGES_DIRECTORY.concat(s);
        }
    }
};

ModelResource.prototype.addImagesDataFromQueue = function(queue){
    //add property img to existing objects
    var aIMG = this.objectsData.IMG;
    if(!Array.isArray(aIMG))
        return;
    var obj, i, l = aIMG.length;
    for(i=0; i<l; i++){
        obj = aIMG[i];
        if(typeof obj["Id"] == TYPE_STRING)
            obj["img"] = queue.getResult(obj["Id"]);
    }
};
/**
 * @constructor
 */
function AbstractSVGLibrary(aSVGStorage){
    var me = this;
    Object.getPrototypeOf(me).aSVGStorage = aSVGStorage;
    me.initSVGImages();
}
extend(AbstractSVGLibrary, Object);

AbstractSVGLibrary.prototype.srcSVGPrefix = "data:image/svg+xml;base64,";

//
AbstractSVGLibrary.prototype.makeSVG = function(data){
    var me = this;
    var img = document.createElement("IMG");
    img.src = me.srcSVGPrefix + window.btoa(data);
    return img;
};

//
AbstractSVGLibrary.prototype.initSVGImages = function(){
    var aIMG = this.aSVGStorage;
    if(!Array.isArray(aIMG))
        return;
    var obj, i, l = aIMG.length, s, j, w, h;
    for(i=0; i<l; i++){
        obj = aIMG[i];
        if(typeof obj["data"] == TYPE_STRING && typeof obj["id"] == TYPE_STRING){
            s = obj["data"];
            j = s.indexOf("width=\"");
            if(j!=-1) {
                w = parseInt(s.slice(j + 7));
                j = s.indexOf("height=\"");
                if(j!=-1) {
                    h = parseInt(s.slice(j + 8));
                    //we need to store these properties explicitly because of IE security block
                    // preventing us from requesting .width and .height directly from image objects
                    obj.width = w;
                    obj.height = h;
                }
            }
            obj.svg = this.makeSVG(s);
        }
    }
};

//"static" method
AbstractSVGLibrary.prototype.getSizeById = function(sId){
    var me = this;
    var i, n;
    var aIconStorage = me.aSVGStorage;
    n = aIconStorage.length;
    for(i=0; i<n; i++)
        if(aIconStorage[i].id == sId)
            break;
    if(i==n)
        return null;
    var obj = aIconStorage[i];
    var objSize = {width: obj.width, height: obj.height};
    objSize.xShift = (typeof obj.xShift == TYPE_NUMBER) ? obj.xShift : 0;
    objSize.yShift = (typeof obj.yShift == TYPE_NUMBER) ? obj.yShift : 0;
    return objSize;
};

//"static" method
AbstractSVGLibrary.prototype.getCachedSVG = function(sId){
    var a = this.aSVGStorage;
    if(!Array.isArray(a))
        return;
    var i, n = a.length;
    for (i = 0; i < n; i++)
        if (a[i].id == sId)
            return a[i].svg;
    return null;
};
/**
 * @constructor
 */function
ObjectWithMethodizedProperties(){
    //base class with properties matched to methods by means of (get/set)MethodNameForProperty
}
extend(ObjectWithMethodizedProperties, Object);

ObjectWithMethodizedProperties.prototype.aMethodNameForProperty = [];
ObjectWithMethodizedProperties.prototype.setMethodNameForProperty = function(aDest, sProperty, flagSetMethodName, sMethod){
    if(!Array.isArray(aDest))
        return;
    var i, n = aDest.length;
    var obj;
    for(i=0; i<n; i++){
        obj = aDest[i];
        if(obj.property == sProperty){
            if(flagSetMethodName)
                obj.set = sMethod;
            else
                obj.get = sMethod;
            if(
                (typeof obj.get == TYPE_UNDEFINED || obj.get == null) &&
                (typeof obj.set == TYPE_UNDEFINED || obj.set == null)
            )
                aDest.splice(i,1);
            return;
        }
    }

    if(typeof sMethod == TYPE_UNDEFINED || sMethod == null)
        return;

    obj = {property:sProperty};
    if(flagSetMethodName)
        obj.set = sMethod;
    else
        obj.get = sMethod;
    aDest.push(obj);
};

ObjectWithMethodizedProperties.prototype.getMethodNameForProperty = function(sProperty, flagSetMethodName){
    var me = this;
    var a = me.aMethodNameForProperty;
    var i, n = a.length;
    var obj;
    for(i=0; i<n; i++){
        obj = a[i];
        if(obj.property == sProperty){
            if(flagSetMethodName)
                return (typeof obj.set == TYPE_STRING && typeof me[obj.set] == TYPE_FUNCTION) ? obj.set : null;
            else
                return (typeof obj.get == TYPE_STRING && typeof me[obj.get] == TYPE_FUNCTION) ? obj.get : null;
        }
    }
    return null;
};
///////////////////////////////////////базовый класс для элементов, которые не отображаются но создают события//////////////////////
/**
 * @constructor
 */
function NonDOMEventDispatcher(){
	this.entityDiv = document.createElement("DIV");
	//слушатели событий должны вызываться для this, равного данному элементу, а не entityDiv. Соответствующие функции-посредники хранятся в objListeners
	var objListenersCapture = this.objListenersCapture = {};
	var objListenersNoCapture = this.objListenersNoCapture = {};
	var objTargetsCapture = this.objTargetsCapture = {};
	var objTargetsNoCapture = this.objTargetsNoCapture = {};
	//
	var fListenerWrapper = function(e, objListeners, objTargets){
		var sType = e.type;
		if(typeof objListeners[sType] != TYPE_UNDEFINED && typeof objListeners[sType].length == TYPE_NUMBER){
			var a = objListeners[sType].slice(0), aT = objTargets[sType].slice(0), i, l = a.length;
			for(i=0; i<l; i++) {
                a[i].call(aT[i], e);
			}
			a.splice(0);
			aT.splice(0);
		}
	};
	this.listenerWrapperCapture = function(e){
		fListenerWrapper(e, objListenersCapture, objTargetsCapture);
	};
	this.listenerWrapperNoCapture = function(e){
		fListenerWrapper(e, objListenersNoCapture, objTargetsNoCapture);
	};
}
extend(NonDOMEventDispatcher, ObjectWithMethodizedProperties);

//
NonDOMEventDispatcher.prototype.dispatchEvent = function(e){
	this.entityDiv.dispatchEvent(e);
};

//
NonDOMEventDispatcher.prototype.conveyEvent = function(e, objAddData){
    var evt = new ModelEvent(e.type);
    evt.data = miscFunctions.shallowCloneObject(e.data);
    if(typeof objAddData != TYPE_UNDEFINED && objAddData != null){
        for(var prop in objAddData)
            evt.data[prop] = objAddData[prop];
    }
    this.dispatchEvent(evt);
};

//
NonDOMEventDispatcher.prototype.addEventListener = function(sType, _fListener, _flagUseCapture){
	var flagUseCapture = (typeof _flagUseCapture == TYPE_UNDEFINED) ? false : _flagUseCapture;
	var fListener, targetListener;
	
	if(typeof _fListener == TYPE_FUNCTION){
		fListener = _fListener;
		targetListener = this;
	}else{
		try{
			fListener = _fListener.listener;
		}catch(e){
		}
		targetListener = _fListener.target;
	}
	
	var objListeners = flagUseCapture ? this.objListenersCapture : this.objListenersNoCapture;
	var objTargets = flagUseCapture ? this.objTargetsCapture : this.objTargetsNoCapture;
	if(typeof objListeners[sType] == TYPE_UNDEFINED){
		objListeners[sType] = [];
		objTargets[sType] = [];
	}
	var a = objListeners[sType], aT = objTargets[sType];
	if(a.length == 0)
		this.entityDiv.addEventListener(sType, flagUseCapture ? this.listenerWrapperCapture : this.listenerWrapperNoCapture, flagUseCapture);
	var i, n = a.length;
	for(i=0; i<n; i++)
		if(a[i] == fListener && aT[i] == targetListener)
			break;
	if(i == n){
		a.push(fListener);
		aT.push(targetListener);
	}
};
//
NonDOMEventDispatcher.prototype.removeEventListener = function(sType, _fListener, _flagUseCapture){
	var flagUseCapture = (typeof _flagUseCapture == TYPE_UNDEFINED) ? false : _flagUseCapture;
	var fListener, targetListener;
	
	if(typeof _fListener == TYPE_FUNCTION){
		fListener = _fListener;
		targetListener = this;
	}else{
		fListener = _fListener.listener;
		targetListener = _fListener.target;
	}
	
	var objListeners = flagUseCapture ? this.objListenersCapture : this.objListenersNoCapture;
	var objTargets = flagUseCapture ? this.objTargetsCapture : this.objTargetsNoCapture;
	if(typeof objListeners[sType] == TYPE_UNDEFINED)
		return;
	var a = objListeners[sType], aT = objTargets[sType];
	var i, n = a.length;
	for(i=0; i<n; i++)
		if(a[i] == fListener && aT[i] == targetListener)
			break;
	if(i < n){
		a.splice(i, 1);
		aT.splice(i, 1);
		if(a.length == 0)
			this.entityDiv.removeEventListener(sType, flagUseCapture ? this.listenerWrapperCapture : this.listenerWrapperNoCapture, flagUseCapture);
	}
};
//
NonDOMEventDispatcher.prototype.deactivate = function(){
	var sType, j, aOL = [this.objListenersCapture, this.objListenersNoCapture], objL;
	for(j=0; j<aOL.length; j++){
		objL = aOL[j];
		for(sType in objL){
			this.entityDiv.removeEventListener(sType, (j==0) ? this.listenerWrapperCapture : this.listenerWrapperNoCapture, (j==0));
			objL[sType].splice(0);
		}
	}
};
//
NonDOMEventDispatcher.prototype.willTrigger = function(sType){
	return (typeof this.objListenersCapture[sType]!=TYPE_UNDEFINED && this.objListenersCapture[sType].length>0) || (typeof this.objListenersNoCapture[sType]!=TYPE_UNDEFINED && this.objListenersNoCapture[sType].length>0);
};
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////